del H Na4Al4Si4O16 (J)                            0
del S Na4Al4Si4O16 (J/K)                          0
del V Na4Al4Si4O16 (J/bar)                        0
del H K4Al4Si4O16 (J)                      -3572.76
del S K4Al4Si4O16 (J/K)                           0
del V K4Al4Si4O16 (J/bar)                         0
del H []Na3Al5Si4O16 (J)                          0
del S []Na3Al5Si4O16 (J/K)                        0
del V []Na3Al5Si4O16 (J/bar)                      0
del H []CaNa2Al4Si4O16 (J)                    23096
del S []CaNa2Al4Si4O16 (J/K)                      0
del V []CaNa2Al4Si4O16 (J/bar)                    0
                    /* kals - neph */
#define DH1                  32145.67 /* Na4Al4Si4O16 joules */
#define DS1                     10.46 /* joules/T */
#define DV1                         0 /* joules/bar */
#define DH2                   -5648.4 /* K4Al4Si4O16 joules */
#define DS2                         0 /* joules/T */
#define DV2                     -0.04 /* joules/bar */
#define DH3                     14644 /* []Na3Al3Si5O16 joules */
#define DS3                     -18.7 /* joules/T */
#define DV3                         0 /* joules/bar */
#define DH4                     35184 /* []CaNa2Al2Si2O16 joules */
#define DS4                     -7.17 /* joules/T */
#define DV4                         0 /* joules/bar */

#define HEX                 -31744.63 /* joules */
#define SEX                    -20.92 /* joules/T */
#define VEX                     -1.05 /* joules/bar */
#define HX                  -13893.18 /* joules */
#define SX                      12.55 /* joules/T */
#define VX                          0 /* joules/bar */

#define H23                     73520 /* joules */
#define H24                     42560 /* joules */
                    /* nepheline structure */
#define WHNAKLS               6861.76 /* W(LS)Na-K joules */
#define WVNAKLS                  0.33 /* joules/bar */
#define WHNAKSS              51002.96 /* W(SS)Na-K joules */
#define WVNAKSS                  0.54 /* joules/bar */
#define WVN                     14644 /* W[]Si-NaAl joules */
#define WVK                      8368 /* W[]Si-KAl joules */
#define WCANA                       0 /* W[]Ca-Na2 joules */
#define WCAK                        0 /* W[]Ca-K2 joules */
#define WPLAG                       0 /* WCaAl-NaSi joules */
                    /* kalsilite structure */
#define WKALS                   29288 /* WNa4-K4 joules */
#define WKVKALS                 30334 /* WK4-[]Na3 joules */
#define WNVKALS                 14646 /* WNa4-[]Na3 joules */
#define WNCKALS                     0 /* WNa4-[]CaNa2 joules */
#define WKCKALS                 14646 /* WK4-[]CaNa2 joules */
#define WVVKALS                     0 /* W[]Na3-[]CaNa2 joules */

#define DWKNALS                     0 /* joules */
#define DWKNASS                     0 /* joules */
#define DWVNASS                     0 /* joules */
#define DWVNALS                     0 /* joules */
#define DWVKLS                      0 /* joules */
#define AWLS                        0 /* joules */
#define AWSS                        0 /* joules */
#define AWVNA                       0 /* joules */
#define AWVK                        0 /* joules */
Choice                              0 /* joules */
del H KAlSi2O6 (J)                          21866.4
del S KAlSi2O6 (J/K)                          21.75
del H NaAlSi2O5(OH)2 (J)                     -17400
del S NaAlSi2O5(OH)2 (J/K)                        0
del H NaAlSi2O6 (J)                       -21875.55
del S NaAlSi2O6 (J/K)                         -7.94
W Na-K (leucite)                 7000 /* joules */
W Na-H (leucite)                 7000 /* joules */
del GR (leucite)                53000 /* joules */
#define S23                         0 /* joules/K */
#define S24                         0 /* joules/K */
