/* Generated by Cython 0.29.26 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/usr/local/lib/python3.7/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/usr/local/lib/python3.7/site-packages/numpy/core/include/numpy/arrayscalars.h",
            "/usr/local/lib/python3.7/site-packages/numpy/core/include/numpy/ndarrayobject.h",
            "/usr/local/lib/python3.7/site-packages/numpy/core/include/numpy/ndarraytypes.h",
            "/usr/local/lib/python3.7/site-packages/numpy/core/include/numpy/ufuncobject.h",
            "src/melts/cframework.h"
        ],
        "extra_compile_args": [
            "-O3",
            "-DNPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION",
            "-DBATCH_VERSION",
            "-DRHYOLITE_ADJUSTMENTS",
            "-Wno-unused-but-set-variable",
            "-Wno-misleading-indentation"
        ],
        "extra_objects": [
            "/usr/lib/x86_64-linux-gnu/libxml2.so.2"
        ],
        "include_dirs": [
            "src/melts",
            "./inc",
            "./src/melts",
            "/usr/include/libxml2",
            "/usr/local/lib/python3.7/site-packages/numpy/core/include"
        ],
        "library_dirs": [
            "/usr/local/lib",
            "/usr/lib/x86_64-linux-gnu"
        ],
        "name": "melts.rMELTSframework",
        "runtime_library_dirs": [
            "/usr/local/lib",
            "/usr/lib/x86_64-linux-gnu"
        ],
        "sources": [
            "src/melts/rMELTSframework.pyx",
            "src/melts/cframework.c",
            "src/albite.c",
            "src/alloy-liquid.c",
            "src/alloy-solid.c",
            "src/amphibole.c",
            "src/biotite.c",
            "src/biotiteTaj.c",
            "src/check_coexisting_liquids.c",
            "src/check_coexisting_solids.c",
            "src/clinopyroxene.c",
            "src/cummingtonite.c",
            "src/equality_constraints.c",
            "src/est_saturation_state.c",
            "src/est_satState_revised.c",
            "src/evaluate_saturation.c",
            "src/feldspar.c",
            "src/fluid.c",
            "src/fluidPhase.c",
            "src/frac.c",
            "src/garnet.c",
            "src/gibbs.c",
            "src/gradient_hessian.c",
            "src/hornblende.c",
            "src/ilmenite.c",
            "src/lawson_hanson.c",
            "src/leucite.c",
            "src/linear_search.c",
            "src/liquid.c",
            "src/liquid_v34.c",
            "src/liquid_CO2.c",
            "src/liquid_CO2_H2O.c",
            "src/liquidus.c",
            "src/majorite.c",
            "src/melilite.c",
            "src/melts_support.c",
            "src/melts_threads.c",
            "src/mthread.c",
            "src/nash.c",
            "src/nepheline.c",
            "src/kalsilite.c",
            "src/olivine.c",
            "src/olivine-sx.c",
            "src/orthopyroxene.c",
            "src/ortho-oxide.c",
            "src/perovskite.c",
            "src/read_write.c",
            "src/recipes.c",
            "src/rhombohedral.c",
            "src/rhomsghiorso.c",
            "src/ringwoodite.c",
            "src/silmin.c",
            "src/silmin_support.c",
            "src/spinel.c",
            "src/subSolidusMuO2.c",
            "src/wadsleyite.c",
            "src/water.c",
            "src/wustite.c"
        ]
    },
    "module_name": "melts.rMELTSframework"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_26"
#define CYTHON_HEX_VERSION 0x001D1AF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030B00A1)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if defined(PyUnicode_IS_READY)
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #else
  #define __Pyx_PyUnicode_READY(op)       (0)
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #endif
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__melts__rMELTSframework
#define __PYX_HAVE_API__melts__rMELTSframework
/* Early includes */
#include "cframework.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "pythread.h"
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"

    /* NumPy API declarations from "numpy/__init__.pxd" */
    
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "src/melts/rMELTSframework.pyx",
  "__init__.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
};

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":690
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":691
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":692
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":693
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":697
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":698
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":699
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":700
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":704
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":705
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":714
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":715
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":716
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":718
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":719
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":720
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":722
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":723
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":725
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":726
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":727
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":729
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":730
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":731
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":733
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_bytes(
         PyObject* string, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    return __Pyx_decode_c_bytes(
        PyBytes_AS_STRING(string), PyBytes_GET_SIZE(string),
        start, stop, encoding, errors, decode_func);
}

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* IncludeStringH.proto */
#include <string.h>

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cpython.version' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_generic = 0;
static PyTypeObject *__pyx_ptype_5numpy_number = 0;
static PyTypeObject *__pyx_ptype_5numpy_integer = 0;
static PyTypeObject *__pyx_ptype_5numpy_signedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_unsignedinteger = 0;
static PyTypeObject *__pyx_ptype_5numpy_inexact = 0;
static PyTypeObject *__pyx_ptype_5numpy_floating = 0;
static PyTypeObject *__pyx_ptype_5numpy_complexfloating = 0;
static PyTypeObject *__pyx_ptype_5numpy_flexible = 0;
static PyTypeObject *__pyx_ptype_5numpy_character = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE int __pyx_f_5numpy_import_array(void); /*proto*/

/* Module declarations from 'melts.rMELTSframework' */
#define __Pyx_MODULE_NAME "melts.rMELTSframework"
extern int __pyx_module_is_main_melts__rMELTSframework;
int __pyx_module_is_main_melts__rMELTSframework = 0;

/* Implementation of 'melts.rMELTSframework' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_i[] = "i";
static const char __pyx_k_j[] = "j";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_bc[] = "bc";
static const char __pyx_k_lc[] = "lc";
static const char __pyx_k_nc[] = "nc";
static const char __pyx_k_nl[] = "nl";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_ns[] = "ns";
static const char __pyx_k_sc[] = "sc";
static const char __pyx_k_inc[] = "inc";
static const char __pyx_k_npc[] = "npc";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_entry[] = "entry";
static const char __pyx_k_names[] = "names";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_ctypes[] = "ctypes";
static const char __pyx_k_cy_fo2[] = "cy_fo2";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_hValue[] = "hValue";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_pValue[] = "pValue";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_sValue[] = "sValue";
static const char __pyx_k_tValue[] = "tValue";
static const char __pyx_k_vValue[] = "vValue";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_cy_init[] = "cy_init";
static const char __pyx_k_result_b[] = "result_b";
static const char __pyx_k_cy_setfo2[] = "cy_setfo2";
static const char __pyx_k_sessionID[] = "sessionID";
static const char __pyx_k_cy_fo2Path[] = "cy_fo2Path";
static const char __pyx_k_typeOfCalc[] = "typeOfCalc";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_cy_fo2Delta[] = "cy_fo2Delta";
static const char __pyx_k_cy_fracMass[] = "cy_fracMass";
static const char __pyx_k_setFracMass[] = "setFracMass";
static const char __pyx_k_cy_solidMass[] = "cy_solidMass";
static const char __pyx_k_cy_liquidMass[] = "cy_liquidMass";
static const char __pyx_k_cy_num_oxides[] = "cy_num_oxides";
static const char __pyx_k_cy_setfo2Path[] = "cy_setfo2Path";
static const char __pyx_k_cy_finalVolume[] = "cy_finalVolume";
static const char __pyx_k_cy_isIsochoric[] = "cy_isIsochoric";
static const char __pyx_k_cy_isIsotropic[] = "cy_isIsotropic";
static const char __pyx_k_cy_setFracMass[] = "cy_setFracMass";
static const char __pyx_k_cy_setfo2Delta[] = "cy_setfo2Delta";
static const char __pyx_k_inputXMLstring[] = "inputXMLstring";
static const char __pyx_k_modelSelection[] = "modelSelection";
static const char __pyx_k_cy_finalEntropy[] = "cy_finalEntropy";
static const char __pyx_k_cy_multipleLiqs[] = "cy_multipleLiqs";
static const char __pyx_k_cy_setIsochoric[] = "cy_setIsochoric";
static const char __pyx_k_cy_setIsotropic[] = "cy_setIsotropic";
static const char __pyx_k_cy_setSolidMass[] = "cy_setSolidMass";
static const char __pyx_k_cy_finalEnthalpy[] = "cy_finalEnthalpy";
static const char __pyx_k_cy_finalPressure[] = "cy_finalPressure";
static const char __pyx_k_cy_initialVolume[] = "cy_initialVolume";
static const char __pyx_k_cy_isIsenthalpic[] = "cy_isIsenthalpic";
static const char __pyx_k_cy_oxygenContent[] = "cy_oxygenContent";
static const char __pyx_k_cy_setLiquidMass[] = "cy_setLiquidMass";
static const char __pyx_k_cy_bulkCompAsList[] = "cy_bulkCompAsList";
static const char __pyx_k_cy_fractionateFlu[] = "cy_fractionateFlu";
static const char __pyx_k_cy_fractionateLiq[] = "cy_fractionateLiq";
static const char __pyx_k_cy_fractionateSol[] = "cy_fractionateSol";
static const char __pyx_k_cy_initialEntropy[] = "cy_initialEntropy";
static const char __pyx_k_cy_nLiquidCoexist[] = "cy_nLiquidCoexist";
static const char __pyx_k_cy_setFinalVolume[] = "cy_setFinalVolume";
static const char __pyx_k_cy_setIsenthalpic[] = "cy_setIsenthalpic";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_cy_fracLCompAsList[] = "cy_fracLCompAsList";
static const char __pyx_k_cy_incSolidsAsList[] = "cy_incSolidsAsList";
static const char __pyx_k_cy_incrementVolume[] = "cy_incrementVolume";
static const char __pyx_k_cy_initialEnthalpy[] = "cy_initialEnthalpy";
static const char __pyx_k_cy_initialPressure[] = "cy_initialPressure";
static const char __pyx_k_cy_setFinalEntropy[] = "cy_setFinalEntropy";
static const char __pyx_k_cy_setMultipleLiqs[] = "cy_setMultipleLiqs";
static const char __pyx_k_cy_finalTemperature[] = "cy_finalTemperature";
static const char __pyx_k_cy_incrementEntropy[] = "cy_incrementEntropy";
static const char __pyx_k_cy_setFinalEnthalpy[] = "cy_setFinalEnthalpy";
static const char __pyx_k_cy_setFinalPressure[] = "cy_setFinalPressure";
static const char __pyx_k_cy_setInitialVolume[] = "cy_setInitialVolume";
static const char __pyx_k_cy_setOxygenContent[] = "cy_setOxygenContent";
static const char __pyx_k_cy_incrementEnthalpy[] = "cy_incrementEnthalpy";
static const char __pyx_k_cy_incrementPressure[] = "cy_incrementPressure";
static const char __pyx_k_cy_setBulkCompAsList[] = "cy_setBulkCompAsList";
static const char __pyx_k_cy_setFractionateFlu[] = "cy_setFractionateFlu";
static const char __pyx_k_cy_setFractionateLiq[] = "cy_setFractionateLiq";
static const char __pyx_k_cy_setFractionateSol[] = "cy_setFractionateSol";
static const char __pyx_k_cy_setInitialEntropy[] = "cy_setInitialEntropy";
static const char __pyx_k_cy_setnLiquidCoexist[] = "cy_setnLiquidCoexist";
static const char __pyx_k_cy_initialTemperature[] = "cy_initialTemperature";
static const char __pyx_k_cy_nFracCoexistAsList[] = "cy_nFracCoexistAsList";
static const char __pyx_k_cy_oxideListAsStrings[] = "cy_oxideListAsStrings";
static const char __pyx_k_cy_phaseListAsStrings[] = "cy_phaseListAsStrings";
static const char __pyx_k_cy_setFracLCompAsList[] = "cy_setFracLCompAsList";
static const char __pyx_k_cy_setIncSolidsAsList[] = "cy_setIncSolidsAsList";
static const char __pyx_k_cy_setIncrementVolume[] = "cy_setIncrementVolume";
static const char __pyx_k_cy_setInitialEnthalpy[] = "cy_setInitialEnthalpy";
static const char __pyx_k_cy_setInitialPressure[] = "cy_setInitialPressure";
static const char __pyx_k_melts_rMELTSframework[] = "melts.rMELTSframework";
static const char __pyx_k_cy_nSolidCoexistAsList[] = "cy_nSolidCoexistAsList";
static const char __pyx_k_cy_setFinalTemperature[] = "cy_setFinalTemperature";
static const char __pyx_k_cy_setIncrementEntropy[] = "cy_setIncrementEntropy";
static const char __pyx_k_cy_incrementTemperature[] = "cy_incrementTemperature";
static const char __pyx_k_cy_num_phase_components[] = "cy_num_phase_components";
static const char __pyx_k_cy_setIncrementEnthalpy[] = "cy_setIncrementEnthalpy";
static const char __pyx_k_cy_setIncrementPressure[] = "cy_setIncrementPressure";
static const char __pyx_k_cy_setInitialTemperature[] = "cy_setInitialTemperature";
static const char __pyx_k_cy_setnFracCoexistAsList[] = "cy_setnFracCoexistAsList";
static const char __pyx_k_cy_fracSCompAsListOfLists[] = "cy_fracSCompAsListOfLists";
static const char __pyx_k_cy_setnSolidCoexistAsList[] = "cy_setnSolidCoexistAsList";
static const char __pyx_k_cy_solidCompAsListOfLists[] = "cy_solidCompAsListOfLists";
static const char __pyx_k_cy_liquidCompAsListOfLists[] = "cy_liquidCompAsListOfLists";
static const char __pyx_k_cy_performMELTScalculation[] = "cy_performMELTScalculation";
static const char __pyx_k_cy_setIncrementTemperature[] = "cy_setIncrementTemperature";
static const char __pyx_k_cy_setFracSCompAsListOfLists[] = "cy_setFracSCompAsListOfLists";
static const char __pyx_k_cy_setSolidCompAsListOfLists[] = "cy_setSolidCompAsListOfLists";
static const char __pyx_k_cy_setLiquidCompAsListOfLists[] = "cy_setLiquidCompAsListOfLists";
static const char __pyx_k_src_melts_rMELTSframework_pyx[] = "src/melts/rMELTSframework.pyx";
static const char __pyx_k_cy_MELTScalculationModeConstant[] = "cy_MELTScalculationModeConstant";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_cy_MELTSandCO2_H2OcalculationMod[] = "cy_MELTSandCO2_H2OcalculationModeConstant";
static const char __pyx_k_cy_MELTSandCO2calculationModeCon[] = "cy_MELTSandCO2calculationModeConstant";
static const char __pyx_k_cy_pMELTScalculationModeConstant[] = "cy_pMELTScalculationModeConstant";
static const char __pyx_k_cy_parseAndLoadDataStructuresFro[] = "cy_parseAndLoadDataStructuresFromXMLstring";
static const char __pyx_k_cy_writeDataStructuresToXMLstrin[] = "cy_writeDataStructuresToXMLstring";
static const char __pyx_k_cy_xMELTScalculationModeConstant[] = "cy_xMELTScalculationModeConstant";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_s_bc;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_ctypes;
static PyObject *__pyx_n_s_cy_MELTSandCO2_H2OcalculationMod;
static PyObject *__pyx_n_s_cy_MELTSandCO2calculationModeCon;
static PyObject *__pyx_n_s_cy_MELTScalculationModeConstant;
static PyObject *__pyx_n_s_cy_bulkCompAsList;
static PyObject *__pyx_n_s_cy_finalEnthalpy;
static PyObject *__pyx_n_s_cy_finalEntropy;
static PyObject *__pyx_n_s_cy_finalPressure;
static PyObject *__pyx_n_s_cy_finalTemperature;
static PyObject *__pyx_n_s_cy_finalVolume;
static PyObject *__pyx_n_s_cy_fo2;
static PyObject *__pyx_n_s_cy_fo2Delta;
static PyObject *__pyx_n_s_cy_fo2Path;
static PyObject *__pyx_n_s_cy_fracLCompAsList;
static PyObject *__pyx_n_s_cy_fracMass;
static PyObject *__pyx_n_s_cy_fracSCompAsListOfLists;
static PyObject *__pyx_n_s_cy_fractionateFlu;
static PyObject *__pyx_n_s_cy_fractionateLiq;
static PyObject *__pyx_n_s_cy_fractionateSol;
static PyObject *__pyx_n_s_cy_incSolidsAsList;
static PyObject *__pyx_n_s_cy_incrementEnthalpy;
static PyObject *__pyx_n_s_cy_incrementEntropy;
static PyObject *__pyx_n_s_cy_incrementPressure;
static PyObject *__pyx_n_s_cy_incrementTemperature;
static PyObject *__pyx_n_s_cy_incrementVolume;
static PyObject *__pyx_n_s_cy_init;
static PyObject *__pyx_n_s_cy_initialEnthalpy;
static PyObject *__pyx_n_s_cy_initialEntropy;
static PyObject *__pyx_n_s_cy_initialPressure;
static PyObject *__pyx_n_s_cy_initialTemperature;
static PyObject *__pyx_n_s_cy_initialVolume;
static PyObject *__pyx_n_s_cy_isIsenthalpic;
static PyObject *__pyx_n_s_cy_isIsochoric;
static PyObject *__pyx_n_s_cy_isIsotropic;
static PyObject *__pyx_n_s_cy_liquidCompAsListOfLists;
static PyObject *__pyx_n_s_cy_liquidMass;
static PyObject *__pyx_n_s_cy_multipleLiqs;
static PyObject *__pyx_n_s_cy_nFracCoexistAsList;
static PyObject *__pyx_n_s_cy_nLiquidCoexist;
static PyObject *__pyx_n_s_cy_nSolidCoexistAsList;
static PyObject *__pyx_n_s_cy_num_oxides;
static PyObject *__pyx_n_s_cy_num_phase_components;
static PyObject *__pyx_n_s_cy_oxideListAsStrings;
static PyObject *__pyx_n_s_cy_oxygenContent;
static PyObject *__pyx_n_s_cy_pMELTScalculationModeConstant;
static PyObject *__pyx_n_s_cy_parseAndLoadDataStructuresFro;
static PyObject *__pyx_n_s_cy_performMELTScalculation;
static PyObject *__pyx_n_s_cy_phaseListAsStrings;
static PyObject *__pyx_n_s_cy_setBulkCompAsList;
static PyObject *__pyx_n_s_cy_setFinalEnthalpy;
static PyObject *__pyx_n_s_cy_setFinalEntropy;
static PyObject *__pyx_n_s_cy_setFinalPressure;
static PyObject *__pyx_n_s_cy_setFinalTemperature;
static PyObject *__pyx_n_s_cy_setFinalVolume;
static PyObject *__pyx_n_s_cy_setFracLCompAsList;
static PyObject *__pyx_n_s_cy_setFracMass;
static PyObject *__pyx_n_s_cy_setFracSCompAsListOfLists;
static PyObject *__pyx_n_s_cy_setFractionateFlu;
static PyObject *__pyx_n_s_cy_setFractionateLiq;
static PyObject *__pyx_n_s_cy_setFractionateSol;
static PyObject *__pyx_n_s_cy_setIncSolidsAsList;
static PyObject *__pyx_n_s_cy_setIncrementEnthalpy;
static PyObject *__pyx_n_s_cy_setIncrementEntropy;
static PyObject *__pyx_n_s_cy_setIncrementPressure;
static PyObject *__pyx_n_s_cy_setIncrementTemperature;
static PyObject *__pyx_n_s_cy_setIncrementVolume;
static PyObject *__pyx_n_s_cy_setInitialEnthalpy;
static PyObject *__pyx_n_s_cy_setInitialEntropy;
static PyObject *__pyx_n_s_cy_setInitialPressure;
static PyObject *__pyx_n_s_cy_setInitialTemperature;
static PyObject *__pyx_n_s_cy_setInitialVolume;
static PyObject *__pyx_n_s_cy_setIsenthalpic;
static PyObject *__pyx_n_s_cy_setIsochoric;
static PyObject *__pyx_n_s_cy_setIsotropic;
static PyObject *__pyx_n_s_cy_setLiquidCompAsListOfLists;
static PyObject *__pyx_n_s_cy_setLiquidMass;
static PyObject *__pyx_n_s_cy_setMultipleLiqs;
static PyObject *__pyx_n_s_cy_setOxygenContent;
static PyObject *__pyx_n_s_cy_setSolidCompAsListOfLists;
static PyObject *__pyx_n_s_cy_setSolidMass;
static PyObject *__pyx_n_s_cy_setfo2;
static PyObject *__pyx_n_s_cy_setfo2Delta;
static PyObject *__pyx_n_s_cy_setfo2Path;
static PyObject *__pyx_n_s_cy_setnFracCoexistAsList;
static PyObject *__pyx_n_s_cy_setnLiquidCoexist;
static PyObject *__pyx_n_s_cy_setnSolidCoexistAsList;
static PyObject *__pyx_n_s_cy_solidCompAsListOfLists;
static PyObject *__pyx_n_s_cy_solidMass;
static PyObject *__pyx_n_s_cy_writeDataStructuresToXMLstrin;
static PyObject *__pyx_n_s_cy_xMELTScalculationModeConstant;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_entry;
static PyObject *__pyx_n_s_hValue;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_inc;
static PyObject *__pyx_n_s_inputXMLstring;
static PyObject *__pyx_n_s_j;
static PyObject *__pyx_n_s_lc;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_melts_rMELTSframework;
static PyObject *__pyx_n_s_modelSelection;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_names;
static PyObject *__pyx_n_s_nc;
static PyObject *__pyx_n_s_nl;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_npc;
static PyObject *__pyx_n_s_ns;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_u_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_u_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_pValue;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_n_s_result_b;
static PyObject *__pyx_n_s_sValue;
static PyObject *__pyx_n_s_sc;
static PyObject *__pyx_n_s_sessionID;
static PyObject *__pyx_n_s_setFracMass;
static PyObject *__pyx_kp_s_src_melts_rMELTSframework_pyx;
static PyObject *__pyx_n_s_tValue;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_typeOfCalc;
static PyObject *__pyx_kp_u_utf_8;
static PyObject *__pyx_n_s_vValue;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_pf_5melts_15rMELTSframework_cy_MELTScalculationModeConstant(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_2cy_pMELTScalculationModeConstant(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_4cy_xMELTScalculationModeConstant(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_6cy_MELTSandCO2calculationModeConstant(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_8cy_MELTSandCO2_H2OcalculationModeConstant(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_10cy_num_oxides(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_12cy_num_phase_components(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_14cy_oxideListAsStrings(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_16cy_phaseListAsStrings(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_modelSelection); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_18cy_init(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_modelSelection); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_20cy_parseAndLoadDataStructuresFromXMLstring(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputXMLstring); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_22cy_writeDataStructuresToXMLstring(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sessionID); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_24cy_performMELTScalculation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_typeOfCalc); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_26cy_initialTemperature(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_28cy_setInitialTemperature(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_tValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_30cy_finalTemperature(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_32cy_setFinalTemperature(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_tValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_34cy_incrementTemperature(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_36cy_setIncrementTemperature(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_tValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_38cy_initialPressure(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_40cy_setInitialPressure(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_pValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_42cy_finalPressure(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_44cy_setFinalPressure(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_pValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_46cy_incrementPressure(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_48cy_setIncrementPressure(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_pValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_50cy_initialVolume(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_52cy_setInitialVolume(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_vValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_54cy_finalVolume(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_56cy_setFinalVolume(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_vValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_58cy_incrementVolume(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_60cy_setIncrementVolume(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_vValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_62cy_initialEnthalpy(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_64cy_setInitialEnthalpy(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_hValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_66cy_finalEnthalpy(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_68cy_setFinalEnthalpy(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_hValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_70cy_incrementEnthalpy(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_72cy_setIncrementEnthalpy(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_hValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_74cy_initialEntropy(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_76cy_setInitialEntropy(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_sValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_78cy_finalEntropy(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_80cy_setFinalEntropy(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_sValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_82cy_incrementEntropy(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_84cy_setIncrementEntropy(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_sValue); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_86cy_isIsochoric(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_88cy_setIsochoric(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_90cy_isIsenthalpic(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_92cy_setIsenthalpic(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_94cy_isIsotropic(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_96cy_setIsotropic(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_98cy_bulkCompAsList(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_100cy_setBulkCompAsList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_102cy_liquidCompAsListOfLists(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_104cy_setLiquidCompAsListOfLists(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_106cy_nLiquidCoexist(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_108cy_setnLiquidCoexist(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_110cy_liquidMass(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_112cy_setLiquidMass(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_114cy_multipleLiqs(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_116cy_setMultipleLiqs(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_118cy_nSolidCoexistAsList(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_120cy_setnSolidCoexistAsList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_122cy_solidCompAsListOfLists(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_124cy_setSolidCompAsListOfLists(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_126cy_solidMass(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_128cy_setSolidMass(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_130cy_incSolidsAsList(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_132cy_setIncSolidsAsList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_134cy_fo2(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_136cy_setfo2(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_138cy_fo2Path(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_140cy_setfo2Path(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_142cy_fo2Delta(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_144cy_setfo2Delta(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_146cy_oxygenContent(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_148cy_setOxygenContent(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_150cy_fractionateSol(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_152cy_setFractionateSol(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_154cy_fractionateLiq(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_156cy_setFractionateLiq(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_158cy_fractionateFlu(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_160cy_setFractionateFlu(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_162cy_nFracCoexistAsList(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_164cy_setnFracCoexistAsList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_166cy_fracSCompAsListOfLists(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_168cy_setFracSCompAsListOfLists(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_170cy_fracLCompAsList(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_172cy_setFracLCompAsList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_174cy_fracMass(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_5melts_15rMELTSframework_176cy_setFracMass(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_int_0;
static PyObject *__pyx_k_;
static PyObject *__pyx_k__2;
static PyObject *__pyx_k__3;
static PyObject *__pyx_k__4;
static PyObject *__pyx_k__5;
static PyObject *__pyx_k__6;
static PyObject *__pyx_k__7;
static PyObject *__pyx_k__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__52;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__58;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_tuple__64;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__70;
static PyObject *__pyx_tuple__73;
static PyObject *__pyx_tuple__76;
static PyObject *__pyx_tuple__79;
static PyObject *__pyx_tuple__82;
static PyObject *__pyx_tuple__84;
static PyObject *__pyx_tuple__86;
static PyObject *__pyx_tuple__88;
static PyObject *__pyx_tuple__90;
static PyObject *__pyx_tuple__93;
static PyObject *__pyx_tuple__96;
static PyObject *__pyx_tuple__99;
static PyObject *__pyx_tuple__101;
static PyObject *__pyx_tuple__103;
static PyObject *__pyx_tuple__105;
static PyObject *__pyx_tuple__107;
static PyObject *__pyx_tuple__110;
static PyObject *__pyx_tuple__112;
static PyObject *__pyx_tuple__114;
static PyObject *__pyx_tuple__117;
static PyObject *__pyx_tuple__120;
static PyObject *__pyx_tuple__123;
static PyObject *__pyx_tuple__126;
static PyObject *__pyx_tuple__129;
static PyObject *__pyx_tuple__132;
static PyObject *__pyx_tuple__135;
static PyObject *__pyx_tuple__137;
static PyObject *__pyx_tuple__139;
static PyObject *__pyx_tuple__141;
static PyObject *__pyx_tuple__143;
static PyObject *__pyx_tuple__145;
static PyObject *__pyx_tuple__147;
static PyObject *__pyx_tuple__150;
static PyObject *__pyx_codeobj__11;
static PyObject *__pyx_codeobj__12;
static PyObject *__pyx_codeobj__13;
static PyObject *__pyx_codeobj__14;
static PyObject *__pyx_codeobj__15;
static PyObject *__pyx_codeobj__16;
static PyObject *__pyx_codeobj__17;
static PyObject *__pyx_codeobj__19;
static PyObject *__pyx_codeobj__21;
static PyObject *__pyx_codeobj__23;
static PyObject *__pyx_codeobj__25;
static PyObject *__pyx_codeobj__27;
static PyObject *__pyx_codeobj__29;
static PyObject *__pyx_codeobj__30;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__33;
static PyObject *__pyx_codeobj__35;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__38;
static PyObject *__pyx_codeobj__39;
static PyObject *__pyx_codeobj__41;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__45;
static PyObject *__pyx_codeobj__47;
static PyObject *__pyx_codeobj__48;
static PyObject *__pyx_codeobj__50;
static PyObject *__pyx_codeobj__51;
static PyObject *__pyx_codeobj__53;
static PyObject *__pyx_codeobj__54;
static PyObject *__pyx_codeobj__56;
static PyObject *__pyx_codeobj__57;
static PyObject *__pyx_codeobj__59;
static PyObject *__pyx_codeobj__60;
static PyObject *__pyx_codeobj__62;
static PyObject *__pyx_codeobj__63;
static PyObject *__pyx_codeobj__65;
static PyObject *__pyx_codeobj__66;
static PyObject *__pyx_codeobj__68;
static PyObject *__pyx_codeobj__69;
static PyObject *__pyx_codeobj__71;
static PyObject *__pyx_codeobj__72;
static PyObject *__pyx_codeobj__74;
static PyObject *__pyx_codeobj__75;
static PyObject *__pyx_codeobj__77;
static PyObject *__pyx_codeobj__78;
static PyObject *__pyx_codeobj__80;
static PyObject *__pyx_codeobj__81;
static PyObject *__pyx_codeobj__83;
static PyObject *__pyx_codeobj__85;
static PyObject *__pyx_codeobj__87;
static PyObject *__pyx_codeobj__89;
static PyObject *__pyx_codeobj__91;
static PyObject *__pyx_codeobj__92;
static PyObject *__pyx_codeobj__94;
static PyObject *__pyx_codeobj__95;
static PyObject *__pyx_codeobj__97;
static PyObject *__pyx_codeobj__98;
static PyObject *__pyx_codeobj__100;
static PyObject *__pyx_codeobj__102;
static PyObject *__pyx_codeobj__104;
static PyObject *__pyx_codeobj__106;
static PyObject *__pyx_codeobj__108;
static PyObject *__pyx_codeobj__109;
static PyObject *__pyx_codeobj__111;
static PyObject *__pyx_codeobj__113;
static PyObject *__pyx_codeobj__115;
static PyObject *__pyx_codeobj__116;
static PyObject *__pyx_codeobj__118;
static PyObject *__pyx_codeobj__119;
static PyObject *__pyx_codeobj__121;
static PyObject *__pyx_codeobj__122;
static PyObject *__pyx_codeobj__124;
static PyObject *__pyx_codeobj__125;
static PyObject *__pyx_codeobj__127;
static PyObject *__pyx_codeobj__128;
static PyObject *__pyx_codeobj__130;
static PyObject *__pyx_codeobj__131;
static PyObject *__pyx_codeobj__133;
static PyObject *__pyx_codeobj__134;
static PyObject *__pyx_codeobj__136;
static PyObject *__pyx_codeobj__138;
static PyObject *__pyx_codeobj__140;
static PyObject *__pyx_codeobj__142;
static PyObject *__pyx_codeobj__144;
static PyObject *__pyx_codeobj__146;
static PyObject *__pyx_codeobj__148;
static PyObject *__pyx_codeobj__149;
static PyObject *__pyx_codeobj__151;
/* Late includes */

/* "melts/rMELTSframework.pyx":110
 * # here is the "wrapper" signature
 * 
 * def cy_MELTScalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return MELTScalculationModeConstant()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_1cy_MELTScalculationModeConstant(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_cy_MELTScalculationModeConstant[] = "cy_MELTScalculationModeConstant()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_1cy_MELTScalculationModeConstant = {"cy_MELTScalculationModeConstant", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_1cy_MELTScalculationModeConstant, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_cy_MELTScalculationModeConstant};
static PyObject *__pyx_pw_5melts_15rMELTSframework_1cy_MELTScalculationModeConstant(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_MELTScalculationModeConstant (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_cy_MELTScalculationModeConstant(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_cy_MELTScalculationModeConstant(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_MELTScalculationModeConstant", 0);

  /* "melts/rMELTSframework.pyx":111
 * 
 * def cy_MELTScalculationModeConstant():
 * 	return MELTScalculationModeConstant()             # <<<<<<<<<<<<<<
 * 
 * def cy_pMELTScalculationModeConstant():
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(MELTScalculationModeConstant()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":110
 * # here is the "wrapper" signature
 * 
 * def cy_MELTScalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return MELTScalculationModeConstant()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_MELTScalculationModeConstant", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":113
 * 	return MELTScalculationModeConstant()
 * 
 * def cy_pMELTScalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return pMELTScalculationModeConstant()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_3cy_pMELTScalculationModeConstant(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_2cy_pMELTScalculationModeConstant[] = "cy_pMELTScalculationModeConstant()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_3cy_pMELTScalculationModeConstant = {"cy_pMELTScalculationModeConstant", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_3cy_pMELTScalculationModeConstant, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_2cy_pMELTScalculationModeConstant};
static PyObject *__pyx_pw_5melts_15rMELTSframework_3cy_pMELTScalculationModeConstant(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_pMELTScalculationModeConstant (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_2cy_pMELTScalculationModeConstant(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_2cy_pMELTScalculationModeConstant(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_pMELTScalculationModeConstant", 0);

  /* "melts/rMELTSframework.pyx":114
 * 
 * def cy_pMELTScalculationModeConstant():
 * 	return pMELTScalculationModeConstant()             # <<<<<<<<<<<<<<
 * 
 * def cy_xMELTScalculationModeConstant():
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(pMELTScalculationModeConstant()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":113
 * 	return MELTScalculationModeConstant()
 * 
 * def cy_pMELTScalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return pMELTScalculationModeConstant()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_pMELTScalculationModeConstant", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":116
 * 	return pMELTScalculationModeConstant()
 * 
 * def cy_xMELTScalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return xMELTScalculationModeConstant()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_5cy_xMELTScalculationModeConstant(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_4cy_xMELTScalculationModeConstant[] = "cy_xMELTScalculationModeConstant()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_5cy_xMELTScalculationModeConstant = {"cy_xMELTScalculationModeConstant", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_5cy_xMELTScalculationModeConstant, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_4cy_xMELTScalculationModeConstant};
static PyObject *__pyx_pw_5melts_15rMELTSframework_5cy_xMELTScalculationModeConstant(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_xMELTScalculationModeConstant (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_4cy_xMELTScalculationModeConstant(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_4cy_xMELTScalculationModeConstant(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_xMELTScalculationModeConstant", 0);

  /* "melts/rMELTSframework.pyx":117
 * 
 * def cy_xMELTScalculationModeConstant():
 * 	return xMELTScalculationModeConstant()             # <<<<<<<<<<<<<<
 * 
 * def cy_MELTSandCO2calculationModeConstant():
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(xMELTScalculationModeConstant()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":116
 * 	return pMELTScalculationModeConstant()
 * 
 * def cy_xMELTScalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return xMELTScalculationModeConstant()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_xMELTScalculationModeConstant", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":119
 * 	return xMELTScalculationModeConstant()
 * 
 * def cy_MELTSandCO2calculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return MELTSandCO2calculationModeConstant()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_7cy_MELTSandCO2calculationModeConstant(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_6cy_MELTSandCO2calculationModeConstant[] = "cy_MELTSandCO2calculationModeConstant()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_7cy_MELTSandCO2calculationModeConstant = {"cy_MELTSandCO2calculationModeConstant", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_7cy_MELTSandCO2calculationModeConstant, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_6cy_MELTSandCO2calculationModeConstant};
static PyObject *__pyx_pw_5melts_15rMELTSframework_7cy_MELTSandCO2calculationModeConstant(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_MELTSandCO2calculationModeConstant (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_6cy_MELTSandCO2calculationModeConstant(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_6cy_MELTSandCO2calculationModeConstant(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_MELTSandCO2calculationModeConstant", 0);

  /* "melts/rMELTSframework.pyx":120
 * 
 * def cy_MELTSandCO2calculationModeConstant():
 * 	return MELTSandCO2calculationModeConstant()             # <<<<<<<<<<<<<<
 * 
 * def cy_MELTSandCO2_H2OcalculationModeConstant():
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(MELTSandCO2calculationModeConstant()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":119
 * 	return xMELTScalculationModeConstant()
 * 
 * def cy_MELTSandCO2calculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return MELTSandCO2calculationModeConstant()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_MELTSandCO2calculationModeConstant", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":122
 * 	return MELTSandCO2calculationModeConstant()
 * 
 * def cy_MELTSandCO2_H2OcalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return MELTSandCO2_H2OcalculationModeConstant()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_9cy_MELTSandCO2_H2OcalculationModeConstant(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_8cy_MELTSandCO2_H2OcalculationModeConstant[] = "cy_MELTSandCO2_H2OcalculationModeConstant()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_9cy_MELTSandCO2_H2OcalculationModeConstant = {"cy_MELTSandCO2_H2OcalculationModeConstant", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_9cy_MELTSandCO2_H2OcalculationModeConstant, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_8cy_MELTSandCO2_H2OcalculationModeConstant};
static PyObject *__pyx_pw_5melts_15rMELTSframework_9cy_MELTSandCO2_H2OcalculationModeConstant(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_MELTSandCO2_H2OcalculationModeConstant (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_8cy_MELTSandCO2_H2OcalculationModeConstant(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_8cy_MELTSandCO2_H2OcalculationModeConstant(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_MELTSandCO2_H2OcalculationModeConstant", 0);

  /* "melts/rMELTSframework.pyx":123
 * 
 * def cy_MELTSandCO2_H2OcalculationModeConstant():
 * 	return MELTSandCO2_H2OcalculationModeConstant()             # <<<<<<<<<<<<<<
 * 
 * def cy_num_oxides():
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(MELTSandCO2_H2OcalculationModeConstant()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":122
 * 	return MELTSandCO2calculationModeConstant()
 * 
 * def cy_MELTSandCO2_H2OcalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return MELTSandCO2_H2OcalculationModeConstant()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_MELTSandCO2_H2OcalculationModeConstant", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":125
 * 	return MELTSandCO2_H2OcalculationModeConstant()
 * 
 * def cy_num_oxides():             # <<<<<<<<<<<<<<
 * 	return num_oxides()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_11cy_num_oxides(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_10cy_num_oxides[] = "cy_num_oxides()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_11cy_num_oxides = {"cy_num_oxides", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_11cy_num_oxides, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_10cy_num_oxides};
static PyObject *__pyx_pw_5melts_15rMELTSframework_11cy_num_oxides(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_num_oxides (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_10cy_num_oxides(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_10cy_num_oxides(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_num_oxides", 0);

  /* "melts/rMELTSframework.pyx":126
 * 
 * def cy_num_oxides():
 * 	return num_oxides()             # <<<<<<<<<<<<<<
 * 
 * def cy_num_phase_components():
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(num_oxides()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":125
 * 	return MELTSandCO2_H2OcalculationModeConstant()
 * 
 * def cy_num_oxides():             # <<<<<<<<<<<<<<
 * 	return num_oxides()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_num_oxides", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":128
 * 	return num_oxides()
 * 
 * def cy_num_phase_components():             # <<<<<<<<<<<<<<
 * 	return num_phase_components()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_13cy_num_phase_components(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_12cy_num_phase_components[] = "cy_num_phase_components()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_13cy_num_phase_components = {"cy_num_phase_components", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_13cy_num_phase_components, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_12cy_num_phase_components};
static PyObject *__pyx_pw_5melts_15rMELTSframework_13cy_num_phase_components(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_num_phase_components (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_12cy_num_phase_components(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_12cy_num_phase_components(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_num_phase_components", 0);

  /* "melts/rMELTSframework.pyx":129
 * 
 * def cy_num_phase_components():
 * 	return num_phase_components()             # <<<<<<<<<<<<<<
 * 
 * def cy_oxideListAsStrings():
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(num_phase_components()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":128
 * 	return num_oxides()
 * 
 * def cy_num_phase_components():             # <<<<<<<<<<<<<<
 * 	return num_phase_components()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_num_phase_components", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":131
 * 	return num_phase_components()
 * 
 * def cy_oxideListAsStrings():             # <<<<<<<<<<<<<<
 * 	cdef char **names = oxideListAsStrings()
 * 	n = num_oxides()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_15cy_oxideListAsStrings(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_14cy_oxideListAsStrings[] = "cy_oxideListAsStrings()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_15cy_oxideListAsStrings = {"cy_oxideListAsStrings", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_15cy_oxideListAsStrings, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_14cy_oxideListAsStrings};
static PyObject *__pyx_pw_5melts_15rMELTSframework_15cy_oxideListAsStrings(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_oxideListAsStrings (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_14cy_oxideListAsStrings(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_14cy_oxideListAsStrings(CYTHON_UNUSED PyObject *__pyx_self) {
  char **__pyx_v_names;
  int __pyx_v_n;
  PyObject *__pyx_v_result = NULL;
  long __pyx_v_i;
  PyObject *__pyx_v_entry = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  long __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_oxideListAsStrings", 0);

  /* "melts/rMELTSframework.pyx":132
 * 
 * def cy_oxideListAsStrings():
 * 	cdef char **names = oxideListAsStrings()             # <<<<<<<<<<<<<<
 * 	n = num_oxides()
 * 	result = []
 */
  __pyx_v_names = oxideListAsStrings();

  /* "melts/rMELTSframework.pyx":133
 * def cy_oxideListAsStrings():
 * 	cdef char **names = oxideListAsStrings()
 * 	n = num_oxides()             # <<<<<<<<<<<<<<
 * 	result = []
 * 	for i in range(0,n):
 */
  __pyx_v_n = num_oxides();

  /* "melts/rMELTSframework.pyx":134
 * 	cdef char **names = oxideListAsStrings()
 * 	n = num_oxides()
 * 	result = []             # <<<<<<<<<<<<<<
 * 	for i in range(0,n):
 * 		entry = <bytes> names[i]
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":135
 * 	n = num_oxides()
 * 	result = []
 * 	for i in range(0,n):             # <<<<<<<<<<<<<<
 * 		entry = <bytes> names[i]
 * 		result.append(entry.decode('UTF-8'))
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "melts/rMELTSframework.pyx":136
 * 	result = []
 * 	for i in range(0,n):
 * 		entry = <bytes> names[i]             # <<<<<<<<<<<<<<
 * 		result.append(entry.decode('UTF-8'))
 * 	return result
 */
    __pyx_t_1 = __Pyx_PyBytes_FromString((__pyx_v_names[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __pyx_t_1;
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_entry, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "melts/rMELTSframework.pyx":137
 * 	for i in range(0,n):
 * 		entry = <bytes> names[i]
 * 		result.append(entry.decode('UTF-8'))             # <<<<<<<<<<<<<<
 * 	return result
 * 
 */
    if (unlikely(__pyx_v_entry == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
      __PYX_ERR(0, 137, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_decode_bytes(__pyx_v_entry, 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_result, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }

  /* "melts/rMELTSframework.pyx":138
 * 		entry = <bytes> names[i]
 * 		result.append(entry.decode('UTF-8'))
 * 	return result             # <<<<<<<<<<<<<<
 * 
 * def cy_phaseListAsStrings(int modelSelection):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":131
 * 	return num_phase_components()
 * 
 * def cy_oxideListAsStrings():             # <<<<<<<<<<<<<<
 * 	cdef char **names = oxideListAsStrings()
 * 	n = num_oxides()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_oxideListAsStrings", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_entry);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":140
 * 	return result
 * 
 * def cy_phaseListAsStrings(int modelSelection):             # <<<<<<<<<<<<<<
 * 	cdef char **names = phaseListAsStrings(modelSelection)
 * 	n = num_phases(modelSelection)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_17cy_phaseListAsStrings(PyObject *__pyx_self, PyObject *__pyx_arg_modelSelection); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_16cy_phaseListAsStrings[] = "cy_phaseListAsStrings(int modelSelection)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_17cy_phaseListAsStrings = {"cy_phaseListAsStrings", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_17cy_phaseListAsStrings, METH_O, __pyx_doc_5melts_15rMELTSframework_16cy_phaseListAsStrings};
static PyObject *__pyx_pw_5melts_15rMELTSframework_17cy_phaseListAsStrings(PyObject *__pyx_self, PyObject *__pyx_arg_modelSelection) {
  int __pyx_v_modelSelection;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_phaseListAsStrings (wrapper)", 0);
  assert(__pyx_arg_modelSelection); {
    __pyx_v_modelSelection = __Pyx_PyInt_As_int(__pyx_arg_modelSelection); if (unlikely((__pyx_v_modelSelection == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 140, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_phaseListAsStrings", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_16cy_phaseListAsStrings(__pyx_self, ((int)__pyx_v_modelSelection));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_16cy_phaseListAsStrings(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_modelSelection) {
  char **__pyx_v_names;
  int __pyx_v_n;
  PyObject *__pyx_v_result = NULL;
  long __pyx_v_i;
  PyObject *__pyx_v_entry = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  long __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_phaseListAsStrings", 0);

  /* "melts/rMELTSframework.pyx":141
 * 
 * def cy_phaseListAsStrings(int modelSelection):
 * 	cdef char **names = phaseListAsStrings(modelSelection)             # <<<<<<<<<<<<<<
 * 	n = num_phases(modelSelection)
 * 	result = []
 */
  __pyx_v_names = phaseListAsStrings(__pyx_v_modelSelection);

  /* "melts/rMELTSframework.pyx":142
 * def cy_phaseListAsStrings(int modelSelection):
 * 	cdef char **names = phaseListAsStrings(modelSelection)
 * 	n = num_phases(modelSelection)             # <<<<<<<<<<<<<<
 * 	result = []
 * 	for i in range(0,n):
 */
  __pyx_v_n = num_phases(__pyx_v_modelSelection);

  /* "melts/rMELTSframework.pyx":143
 * 	cdef char **names = phaseListAsStrings(modelSelection)
 * 	n = num_phases(modelSelection)
 * 	result = []             # <<<<<<<<<<<<<<
 * 	for i in range(0,n):
 * 		entry = <bytes> names[i]
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":144
 * 	n = num_phases(modelSelection)
 * 	result = []
 * 	for i in range(0,n):             # <<<<<<<<<<<<<<
 * 		entry = <bytes> names[i]
 * 		result.append(entry.decode('UTF-8'))
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "melts/rMELTSframework.pyx":145
 * 	result = []
 * 	for i in range(0,n):
 * 		entry = <bytes> names[i]             # <<<<<<<<<<<<<<
 * 		result.append(entry.decode('UTF-8'))
 * 	return result
 */
    __pyx_t_1 = __Pyx_PyBytes_FromString((__pyx_v_names[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 145, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __pyx_t_1;
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_entry, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "melts/rMELTSframework.pyx":146
 * 	for i in range(0,n):
 * 		entry = <bytes> names[i]
 * 		result.append(entry.decode('UTF-8'))             # <<<<<<<<<<<<<<
 * 	return result
 * 
 */
    if (unlikely(__pyx_v_entry == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "decode");
      __PYX_ERR(0, 146, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_decode_bytes(__pyx_v_entry, 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_result, __pyx_t_5); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 146, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }

  /* "melts/rMELTSframework.pyx":147
 * 		entry = <bytes> names[i]
 * 		result.append(entry.decode('UTF-8'))
 * 	return result             # <<<<<<<<<<<<<<
 * 
 * def cy_init(int modelSelection):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":140
 * 	return result
 * 
 * def cy_phaseListAsStrings(int modelSelection):             # <<<<<<<<<<<<<<
 * 	cdef char **names = phaseListAsStrings(modelSelection)
 * 	n = num_phases(modelSelection)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_phaseListAsStrings", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_entry);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":149
 * 	return result
 * 
 * def cy_init(int modelSelection):             # <<<<<<<<<<<<<<
 * 	return init(modelSelection)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_19cy_init(PyObject *__pyx_self, PyObject *__pyx_arg_modelSelection); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_18cy_init[] = "cy_init(int modelSelection)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_19cy_init = {"cy_init", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_19cy_init, METH_O, __pyx_doc_5melts_15rMELTSframework_18cy_init};
static PyObject *__pyx_pw_5melts_15rMELTSframework_19cy_init(PyObject *__pyx_self, PyObject *__pyx_arg_modelSelection) {
  int __pyx_v_modelSelection;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_init (wrapper)", 0);
  assert(__pyx_arg_modelSelection); {
    __pyx_v_modelSelection = __Pyx_PyInt_As_int(__pyx_arg_modelSelection); if (unlikely((__pyx_v_modelSelection == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 149, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_18cy_init(__pyx_self, ((int)__pyx_v_modelSelection));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_18cy_init(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_modelSelection) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_init", 0);

  /* "melts/rMELTSframework.pyx":150
 * 
 * def cy_init(int modelSelection):
 * 	return init(modelSelection)             # <<<<<<<<<<<<<<
 * 
 * def cy_parseAndLoadDataStructuresFromXMLstring(inputXMLstring):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(init(__pyx_v_modelSelection)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":149
 * 	return result
 * 
 * def cy_init(int modelSelection):             # <<<<<<<<<<<<<<
 * 	return init(modelSelection)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_init", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":152
 * 	return init(modelSelection)
 * 
 * def cy_parseAndLoadDataStructuresFromXMLstring(inputXMLstring):             # <<<<<<<<<<<<<<
 * 	return parseAndLoadDataStructuresFromXMLstring(inputXMLstring.encode('utf-8'))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_21cy_parseAndLoadDataStructuresFromXMLstring(PyObject *__pyx_self, PyObject *__pyx_v_inputXMLstring); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_20cy_parseAndLoadDataStructuresFromXMLstring[] = "cy_parseAndLoadDataStructuresFromXMLstring(inputXMLstring)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_21cy_parseAndLoadDataStructuresFromXMLstring = {"cy_parseAndLoadDataStructuresFromXMLstring", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_21cy_parseAndLoadDataStructuresFromXMLstring, METH_O, __pyx_doc_5melts_15rMELTSframework_20cy_parseAndLoadDataStructuresFromXMLstring};
static PyObject *__pyx_pw_5melts_15rMELTSframework_21cy_parseAndLoadDataStructuresFromXMLstring(PyObject *__pyx_self, PyObject *__pyx_v_inputXMLstring) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_parseAndLoadDataStructuresFromXMLstring (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_20cy_parseAndLoadDataStructuresFromXMLstring(__pyx_self, ((PyObject *)__pyx_v_inputXMLstring));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_20cy_parseAndLoadDataStructuresFromXMLstring(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_inputXMLstring) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_parseAndLoadDataStructuresFromXMLstring", 0);

  /* "melts/rMELTSframework.pyx":153
 * 
 * def cy_parseAndLoadDataStructuresFromXMLstring(inputXMLstring):
 * 	return parseAndLoadDataStructuresFromXMLstring(inputXMLstring.encode('utf-8'))             # <<<<<<<<<<<<<<
 * 
 * def cy_writeDataStructuresToXMLstring(sessionID):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_inputXMLstring, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 153, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(parseAndLoadDataStructuresFromXMLstring(__pyx_t_4)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":152
 * 	return init(modelSelection)
 * 
 * def cy_parseAndLoadDataStructuresFromXMLstring(inputXMLstring):             # <<<<<<<<<<<<<<
 * 	return parseAndLoadDataStructuresFromXMLstring(inputXMLstring.encode('utf-8'))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_parseAndLoadDataStructuresFromXMLstring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":155
 * 	return parseAndLoadDataStructuresFromXMLstring(inputXMLstring.encode('utf-8'))
 * 
 * def cy_writeDataStructuresToXMLstring(sessionID):             # <<<<<<<<<<<<<<
 * 	result_b = writeDataStructuresToXMLstring(sessionID.encode('utf-8'))
 * 	result = result_b.decode('utf-8')
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_23cy_writeDataStructuresToXMLstring(PyObject *__pyx_self, PyObject *__pyx_v_sessionID); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_22cy_writeDataStructuresToXMLstring[] = "cy_writeDataStructuresToXMLstring(sessionID)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_23cy_writeDataStructuresToXMLstring = {"cy_writeDataStructuresToXMLstring", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_23cy_writeDataStructuresToXMLstring, METH_O, __pyx_doc_5melts_15rMELTSframework_22cy_writeDataStructuresToXMLstring};
static PyObject *__pyx_pw_5melts_15rMELTSframework_23cy_writeDataStructuresToXMLstring(PyObject *__pyx_self, PyObject *__pyx_v_sessionID) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_writeDataStructuresToXMLstring (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_22cy_writeDataStructuresToXMLstring(__pyx_self, ((PyObject *)__pyx_v_sessionID));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_22cy_writeDataStructuresToXMLstring(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_sessionID) {
  char *__pyx_v_result_b;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_writeDataStructuresToXMLstring", 0);

  /* "melts/rMELTSframework.pyx":156
 * 
 * def cy_writeDataStructuresToXMLstring(sessionID):
 * 	result_b = writeDataStructuresToXMLstring(sessionID.encode('utf-8'))             # <<<<<<<<<<<<<<
 * 	result = result_b.decode('utf-8')
 * 	free(result_b)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_sessionID, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u_utf_8);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_AsWritableString(__pyx_t_1); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 156, __pyx_L1_error)
  __pyx_v_result_b = writeDataStructuresToXMLstring(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":157
 * def cy_writeDataStructuresToXMLstring(sessionID):
 * 	result_b = writeDataStructuresToXMLstring(sessionID.encode('utf-8'))
 * 	result = result_b.decode('utf-8')             # <<<<<<<<<<<<<<
 * 	free(result_b)
 * 	return result
 */
  __pyx_t_1 = __Pyx_decode_c_string(__pyx_v_result_b, 0, strlen(__pyx_v_result_b), NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":158
 * 	result_b = writeDataStructuresToXMLstring(sessionID.encode('utf-8'))
 * 	result = result_b.decode('utf-8')
 * 	free(result_b)             # <<<<<<<<<<<<<<
 * 	return result
 * 
 */
  free(__pyx_v_result_b);

  /* "melts/rMELTSframework.pyx":159
 * 	result = result_b.decode('utf-8')
 * 	free(result_b)
 * 	return result             # <<<<<<<<<<<<<<
 * 
 * def cy_performMELTScalculation(typeOfCalc):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":155
 * 	return parseAndLoadDataStructuresFromXMLstring(inputXMLstring.encode('utf-8'))
 * 
 * def cy_writeDataStructuresToXMLstring(sessionID):             # <<<<<<<<<<<<<<
 * 	result_b = writeDataStructuresToXMLstring(sessionID.encode('utf-8'))
 * 	result = result_b.decode('utf-8')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_writeDataStructuresToXMLstring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":161
 * 	return result
 * 
 * def cy_performMELTScalculation(typeOfCalc):             # <<<<<<<<<<<<<<
 * 	return performMELTScalculation(typeOfCalc)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_25cy_performMELTScalculation(PyObject *__pyx_self, PyObject *__pyx_v_typeOfCalc); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_24cy_performMELTScalculation[] = "cy_performMELTScalculation(typeOfCalc)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_25cy_performMELTScalculation = {"cy_performMELTScalculation", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_25cy_performMELTScalculation, METH_O, __pyx_doc_5melts_15rMELTSframework_24cy_performMELTScalculation};
static PyObject *__pyx_pw_5melts_15rMELTSframework_25cy_performMELTScalculation(PyObject *__pyx_self, PyObject *__pyx_v_typeOfCalc) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_performMELTScalculation (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_24cy_performMELTScalculation(__pyx_self, ((PyObject *)__pyx_v_typeOfCalc));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_24cy_performMELTScalculation(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_typeOfCalc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_performMELTScalculation", 0);

  /* "melts/rMELTSframework.pyx":162
 * 
 * def cy_performMELTScalculation(typeOfCalc):
 * 	return performMELTScalculation(typeOfCalc)             # <<<<<<<<<<<<<<
 * 
 * def cy_initialTemperature():
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_typeOfCalc); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 162, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(performMELTScalculation(__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":161
 * 	return result
 * 
 * def cy_performMELTScalculation(typeOfCalc):             # <<<<<<<<<<<<<<
 * 	return performMELTScalculation(typeOfCalc)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_performMELTScalculation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":164
 * 	return performMELTScalculation(typeOfCalc)
 * 
 * def cy_initialTemperature():             # <<<<<<<<<<<<<<
 * 	return initialTemperature()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_27cy_initialTemperature(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_26cy_initialTemperature[] = "cy_initialTemperature()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_27cy_initialTemperature = {"cy_initialTemperature", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_27cy_initialTemperature, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_26cy_initialTemperature};
static PyObject *__pyx_pw_5melts_15rMELTSframework_27cy_initialTemperature(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_initialTemperature (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_26cy_initialTemperature(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_26cy_initialTemperature(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_initialTemperature", 0);

  /* "melts/rMELTSframework.pyx":165
 * 
 * def cy_initialTemperature():
 * 	return initialTemperature()             # <<<<<<<<<<<<<<
 * 
 * def cy_setInitialTemperature(double tValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(initialTemperature()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":164
 * 	return performMELTScalculation(typeOfCalc)
 * 
 * def cy_initialTemperature():             # <<<<<<<<<<<<<<
 * 	return initialTemperature()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_initialTemperature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":167
 * 	return initialTemperature()
 * 
 * def cy_setInitialTemperature(double tValue):             # <<<<<<<<<<<<<<
 * 	setInitialTemperature(tValue)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_29cy_setInitialTemperature(PyObject *__pyx_self, PyObject *__pyx_arg_tValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_28cy_setInitialTemperature[] = "cy_setInitialTemperature(double tValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_29cy_setInitialTemperature = {"cy_setInitialTemperature", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_29cy_setInitialTemperature, METH_O, __pyx_doc_5melts_15rMELTSframework_28cy_setInitialTemperature};
static PyObject *__pyx_pw_5melts_15rMELTSframework_29cy_setInitialTemperature(PyObject *__pyx_self, PyObject *__pyx_arg_tValue) {
  double __pyx_v_tValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setInitialTemperature (wrapper)", 0);
  assert(__pyx_arg_tValue); {
    __pyx_v_tValue = __pyx_PyFloat_AsDouble(__pyx_arg_tValue); if (unlikely((__pyx_v_tValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 167, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setInitialTemperature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_28cy_setInitialTemperature(__pyx_self, ((double)__pyx_v_tValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_28cy_setInitialTemperature(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_tValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setInitialTemperature", 0);

  /* "melts/rMELTSframework.pyx":168
 * 
 * def cy_setInitialTemperature(double tValue):
 * 	setInitialTemperature(tValue)             # <<<<<<<<<<<<<<
 * 
 * def cy_finalTemperature():
 */
  setInitialTemperature(__pyx_v_tValue);

  /* "melts/rMELTSframework.pyx":167
 * 	return initialTemperature()
 * 
 * def cy_setInitialTemperature(double tValue):             # <<<<<<<<<<<<<<
 * 	setInitialTemperature(tValue)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":170
 * 	setInitialTemperature(tValue)
 * 
 * def cy_finalTemperature():             # <<<<<<<<<<<<<<
 * 	return finalTemperature()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_31cy_finalTemperature(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_30cy_finalTemperature[] = "cy_finalTemperature()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_31cy_finalTemperature = {"cy_finalTemperature", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_31cy_finalTemperature, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_30cy_finalTemperature};
static PyObject *__pyx_pw_5melts_15rMELTSframework_31cy_finalTemperature(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_finalTemperature (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_30cy_finalTemperature(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_30cy_finalTemperature(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_finalTemperature", 0);

  /* "melts/rMELTSframework.pyx":171
 * 
 * def cy_finalTemperature():
 * 	return finalTemperature()             # <<<<<<<<<<<<<<
 * 
 * def cy_setFinalTemperature(double tValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(finalTemperature()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":170
 * 	setInitialTemperature(tValue)
 * 
 * def cy_finalTemperature():             # <<<<<<<<<<<<<<
 * 	return finalTemperature()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_finalTemperature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":173
 * 	return finalTemperature()
 * 
 * def cy_setFinalTemperature(double tValue):             # <<<<<<<<<<<<<<
 * 	setFinalTemperature(tValue)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_33cy_setFinalTemperature(PyObject *__pyx_self, PyObject *__pyx_arg_tValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_32cy_setFinalTemperature[] = "cy_setFinalTemperature(double tValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_33cy_setFinalTemperature = {"cy_setFinalTemperature", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_33cy_setFinalTemperature, METH_O, __pyx_doc_5melts_15rMELTSframework_32cy_setFinalTemperature};
static PyObject *__pyx_pw_5melts_15rMELTSframework_33cy_setFinalTemperature(PyObject *__pyx_self, PyObject *__pyx_arg_tValue) {
  double __pyx_v_tValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFinalTemperature (wrapper)", 0);
  assert(__pyx_arg_tValue); {
    __pyx_v_tValue = __pyx_PyFloat_AsDouble(__pyx_arg_tValue); if (unlikely((__pyx_v_tValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 173, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setFinalTemperature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_32cy_setFinalTemperature(__pyx_self, ((double)__pyx_v_tValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_32cy_setFinalTemperature(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_tValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFinalTemperature", 0);

  /* "melts/rMELTSframework.pyx":174
 * 
 * def cy_setFinalTemperature(double tValue):
 * 	setFinalTemperature(tValue)             # <<<<<<<<<<<<<<
 * 
 * def cy_incrementTemperature():
 */
  setFinalTemperature(__pyx_v_tValue);

  /* "melts/rMELTSframework.pyx":173
 * 	return finalTemperature()
 * 
 * def cy_setFinalTemperature(double tValue):             # <<<<<<<<<<<<<<
 * 	setFinalTemperature(tValue)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":176
 * 	setFinalTemperature(tValue)
 * 
 * def cy_incrementTemperature():             # <<<<<<<<<<<<<<
 * 	return incrementTemperature()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_35cy_incrementTemperature(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_34cy_incrementTemperature[] = "cy_incrementTemperature()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_35cy_incrementTemperature = {"cy_incrementTemperature", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_35cy_incrementTemperature, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_34cy_incrementTemperature};
static PyObject *__pyx_pw_5melts_15rMELTSframework_35cy_incrementTemperature(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_incrementTemperature (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_34cy_incrementTemperature(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_34cy_incrementTemperature(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_incrementTemperature", 0);

  /* "melts/rMELTSframework.pyx":177
 * 
 * def cy_incrementTemperature():
 * 	return incrementTemperature()             # <<<<<<<<<<<<<<
 * 
 * def cy_setIncrementTemperature(double tValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(incrementTemperature()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":176
 * 	setFinalTemperature(tValue)
 * 
 * def cy_incrementTemperature():             # <<<<<<<<<<<<<<
 * 	return incrementTemperature()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_incrementTemperature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":179
 * 	return incrementTemperature()
 * 
 * def cy_setIncrementTemperature(double tValue):             # <<<<<<<<<<<<<<
 * 	setIncrementTemperature(tValue);
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_37cy_setIncrementTemperature(PyObject *__pyx_self, PyObject *__pyx_arg_tValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_36cy_setIncrementTemperature[] = "cy_setIncrementTemperature(double tValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_37cy_setIncrementTemperature = {"cy_setIncrementTemperature", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_37cy_setIncrementTemperature, METH_O, __pyx_doc_5melts_15rMELTSframework_36cy_setIncrementTemperature};
static PyObject *__pyx_pw_5melts_15rMELTSframework_37cy_setIncrementTemperature(PyObject *__pyx_self, PyObject *__pyx_arg_tValue) {
  double __pyx_v_tValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIncrementTemperature (wrapper)", 0);
  assert(__pyx_arg_tValue); {
    __pyx_v_tValue = __pyx_PyFloat_AsDouble(__pyx_arg_tValue); if (unlikely((__pyx_v_tValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 179, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setIncrementTemperature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_36cy_setIncrementTemperature(__pyx_self, ((double)__pyx_v_tValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_36cy_setIncrementTemperature(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_tValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIncrementTemperature", 0);

  /* "melts/rMELTSframework.pyx":180
 * 
 * def cy_setIncrementTemperature(double tValue):
 * 	setIncrementTemperature(tValue);             # <<<<<<<<<<<<<<
 * 
 * def cy_initialPressure():
 */
  setIncrementTemperature(__pyx_v_tValue);

  /* "melts/rMELTSframework.pyx":179
 * 	return incrementTemperature()
 * 
 * def cy_setIncrementTemperature(double tValue):             # <<<<<<<<<<<<<<
 * 	setIncrementTemperature(tValue);
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":182
 * 	setIncrementTemperature(tValue);
 * 
 * def cy_initialPressure():             # <<<<<<<<<<<<<<
 * 	return initialPressure()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_39cy_initialPressure(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_38cy_initialPressure[] = "cy_initialPressure()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_39cy_initialPressure = {"cy_initialPressure", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_39cy_initialPressure, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_38cy_initialPressure};
static PyObject *__pyx_pw_5melts_15rMELTSframework_39cy_initialPressure(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_initialPressure (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_38cy_initialPressure(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_38cy_initialPressure(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_initialPressure", 0);

  /* "melts/rMELTSframework.pyx":183
 * 
 * def cy_initialPressure():
 * 	return initialPressure()             # <<<<<<<<<<<<<<
 * 
 * def cy_setInitialPressure(double pValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(initialPressure()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":182
 * 	setIncrementTemperature(tValue);
 * 
 * def cy_initialPressure():             # <<<<<<<<<<<<<<
 * 	return initialPressure()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_initialPressure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":185
 * 	return initialPressure()
 * 
 * def cy_setInitialPressure(double pValue):             # <<<<<<<<<<<<<<
 * 	setInitialPressure(pValue)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_41cy_setInitialPressure(PyObject *__pyx_self, PyObject *__pyx_arg_pValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_40cy_setInitialPressure[] = "cy_setInitialPressure(double pValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_41cy_setInitialPressure = {"cy_setInitialPressure", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_41cy_setInitialPressure, METH_O, __pyx_doc_5melts_15rMELTSframework_40cy_setInitialPressure};
static PyObject *__pyx_pw_5melts_15rMELTSframework_41cy_setInitialPressure(PyObject *__pyx_self, PyObject *__pyx_arg_pValue) {
  double __pyx_v_pValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setInitialPressure (wrapper)", 0);
  assert(__pyx_arg_pValue); {
    __pyx_v_pValue = __pyx_PyFloat_AsDouble(__pyx_arg_pValue); if (unlikely((__pyx_v_pValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 185, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setInitialPressure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_40cy_setInitialPressure(__pyx_self, ((double)__pyx_v_pValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_40cy_setInitialPressure(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_pValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setInitialPressure", 0);

  /* "melts/rMELTSframework.pyx":186
 * 
 * def cy_setInitialPressure(double pValue):
 * 	setInitialPressure(pValue)             # <<<<<<<<<<<<<<
 * 
 * def cy_finalPressure():
 */
  setInitialPressure(__pyx_v_pValue);

  /* "melts/rMELTSframework.pyx":185
 * 	return initialPressure()
 * 
 * def cy_setInitialPressure(double pValue):             # <<<<<<<<<<<<<<
 * 	setInitialPressure(pValue)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":188
 * 	setInitialPressure(pValue)
 * 
 * def cy_finalPressure():             # <<<<<<<<<<<<<<
 * 	return finalPressure()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_43cy_finalPressure(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_42cy_finalPressure[] = "cy_finalPressure()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_43cy_finalPressure = {"cy_finalPressure", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_43cy_finalPressure, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_42cy_finalPressure};
static PyObject *__pyx_pw_5melts_15rMELTSframework_43cy_finalPressure(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_finalPressure (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_42cy_finalPressure(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_42cy_finalPressure(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_finalPressure", 0);

  /* "melts/rMELTSframework.pyx":189
 * 
 * def cy_finalPressure():
 * 	return finalPressure()             # <<<<<<<<<<<<<<
 * 
 * def cy_setFinalPressure(double pValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(finalPressure()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":188
 * 	setInitialPressure(pValue)
 * 
 * def cy_finalPressure():             # <<<<<<<<<<<<<<
 * 	return finalPressure()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_finalPressure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":191
 * 	return finalPressure()
 * 
 * def cy_setFinalPressure(double pValue):             # <<<<<<<<<<<<<<
 * 	setFinalPressure(pValue)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_45cy_setFinalPressure(PyObject *__pyx_self, PyObject *__pyx_arg_pValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_44cy_setFinalPressure[] = "cy_setFinalPressure(double pValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_45cy_setFinalPressure = {"cy_setFinalPressure", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_45cy_setFinalPressure, METH_O, __pyx_doc_5melts_15rMELTSframework_44cy_setFinalPressure};
static PyObject *__pyx_pw_5melts_15rMELTSframework_45cy_setFinalPressure(PyObject *__pyx_self, PyObject *__pyx_arg_pValue) {
  double __pyx_v_pValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFinalPressure (wrapper)", 0);
  assert(__pyx_arg_pValue); {
    __pyx_v_pValue = __pyx_PyFloat_AsDouble(__pyx_arg_pValue); if (unlikely((__pyx_v_pValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 191, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setFinalPressure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_44cy_setFinalPressure(__pyx_self, ((double)__pyx_v_pValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_44cy_setFinalPressure(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_pValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFinalPressure", 0);

  /* "melts/rMELTSframework.pyx":192
 * 
 * def cy_setFinalPressure(double pValue):
 * 	setFinalPressure(pValue)             # <<<<<<<<<<<<<<
 * 
 * def cy_incrementPressure():
 */
  setFinalPressure(__pyx_v_pValue);

  /* "melts/rMELTSframework.pyx":191
 * 	return finalPressure()
 * 
 * def cy_setFinalPressure(double pValue):             # <<<<<<<<<<<<<<
 * 	setFinalPressure(pValue)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":194
 * 	setFinalPressure(pValue)
 * 
 * def cy_incrementPressure():             # <<<<<<<<<<<<<<
 * 	return incrementPressure()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_47cy_incrementPressure(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_46cy_incrementPressure[] = "cy_incrementPressure()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_47cy_incrementPressure = {"cy_incrementPressure", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_47cy_incrementPressure, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_46cy_incrementPressure};
static PyObject *__pyx_pw_5melts_15rMELTSframework_47cy_incrementPressure(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_incrementPressure (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_46cy_incrementPressure(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_46cy_incrementPressure(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_incrementPressure", 0);

  /* "melts/rMELTSframework.pyx":195
 * 
 * def cy_incrementPressure():
 * 	return incrementPressure()             # <<<<<<<<<<<<<<
 * 
 * def cy_setIncrementPressure(double pValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(incrementPressure()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":194
 * 	setFinalPressure(pValue)
 * 
 * def cy_incrementPressure():             # <<<<<<<<<<<<<<
 * 	return incrementPressure()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_incrementPressure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":197
 * 	return incrementPressure()
 * 
 * def cy_setIncrementPressure(double pValue):             # <<<<<<<<<<<<<<
 * 	setIncrementPressure(pValue)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_49cy_setIncrementPressure(PyObject *__pyx_self, PyObject *__pyx_arg_pValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_48cy_setIncrementPressure[] = "cy_setIncrementPressure(double pValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_49cy_setIncrementPressure = {"cy_setIncrementPressure", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_49cy_setIncrementPressure, METH_O, __pyx_doc_5melts_15rMELTSframework_48cy_setIncrementPressure};
static PyObject *__pyx_pw_5melts_15rMELTSframework_49cy_setIncrementPressure(PyObject *__pyx_self, PyObject *__pyx_arg_pValue) {
  double __pyx_v_pValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIncrementPressure (wrapper)", 0);
  assert(__pyx_arg_pValue); {
    __pyx_v_pValue = __pyx_PyFloat_AsDouble(__pyx_arg_pValue); if (unlikely((__pyx_v_pValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 197, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setIncrementPressure", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_48cy_setIncrementPressure(__pyx_self, ((double)__pyx_v_pValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_48cy_setIncrementPressure(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_pValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIncrementPressure", 0);

  /* "melts/rMELTSframework.pyx":198
 * 
 * def cy_setIncrementPressure(double pValue):
 * 	setIncrementPressure(pValue)             # <<<<<<<<<<<<<<
 * 
 * def cy_initialVolume():
 */
  setIncrementPressure(__pyx_v_pValue);

  /* "melts/rMELTSframework.pyx":197
 * 	return incrementPressure()
 * 
 * def cy_setIncrementPressure(double pValue):             # <<<<<<<<<<<<<<
 * 	setIncrementPressure(pValue)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":200
 * 	setIncrementPressure(pValue)
 * 
 * def cy_initialVolume():             # <<<<<<<<<<<<<<
 * 	return initialVolume()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_51cy_initialVolume(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_50cy_initialVolume[] = "cy_initialVolume()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_51cy_initialVolume = {"cy_initialVolume", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_51cy_initialVolume, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_50cy_initialVolume};
static PyObject *__pyx_pw_5melts_15rMELTSframework_51cy_initialVolume(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_initialVolume (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_50cy_initialVolume(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_50cy_initialVolume(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_initialVolume", 0);

  /* "melts/rMELTSframework.pyx":201
 * 
 * def cy_initialVolume():
 * 	return initialVolume()             # <<<<<<<<<<<<<<
 * 
 * def cy_setInitialVolume(double vValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(initialVolume()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":200
 * 	setIncrementPressure(pValue)
 * 
 * def cy_initialVolume():             # <<<<<<<<<<<<<<
 * 	return initialVolume()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_initialVolume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":203
 * 	return initialVolume()
 * 
 * def cy_setInitialVolume(double vValue):             # <<<<<<<<<<<<<<
 * 	setInitialVolume(vValue)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_53cy_setInitialVolume(PyObject *__pyx_self, PyObject *__pyx_arg_vValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_52cy_setInitialVolume[] = "cy_setInitialVolume(double vValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_53cy_setInitialVolume = {"cy_setInitialVolume", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_53cy_setInitialVolume, METH_O, __pyx_doc_5melts_15rMELTSframework_52cy_setInitialVolume};
static PyObject *__pyx_pw_5melts_15rMELTSframework_53cy_setInitialVolume(PyObject *__pyx_self, PyObject *__pyx_arg_vValue) {
  double __pyx_v_vValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setInitialVolume (wrapper)", 0);
  assert(__pyx_arg_vValue); {
    __pyx_v_vValue = __pyx_PyFloat_AsDouble(__pyx_arg_vValue); if (unlikely((__pyx_v_vValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 203, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setInitialVolume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_52cy_setInitialVolume(__pyx_self, ((double)__pyx_v_vValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_52cy_setInitialVolume(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_vValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setInitialVolume", 0);

  /* "melts/rMELTSframework.pyx":204
 * 
 * def cy_setInitialVolume(double vValue):
 * 	setInitialVolume(vValue)             # <<<<<<<<<<<<<<
 * 
 * def cy_finalVolume():
 */
  setInitialVolume(__pyx_v_vValue);

  /* "melts/rMELTSframework.pyx":203
 * 	return initialVolume()
 * 
 * def cy_setInitialVolume(double vValue):             # <<<<<<<<<<<<<<
 * 	setInitialVolume(vValue)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":206
 * 	setInitialVolume(vValue)
 * 
 * def cy_finalVolume():             # <<<<<<<<<<<<<<
 * 	return finalVolume()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_55cy_finalVolume(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_54cy_finalVolume[] = "cy_finalVolume()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_55cy_finalVolume = {"cy_finalVolume", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_55cy_finalVolume, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_54cy_finalVolume};
static PyObject *__pyx_pw_5melts_15rMELTSframework_55cy_finalVolume(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_finalVolume (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_54cy_finalVolume(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_54cy_finalVolume(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_finalVolume", 0);

  /* "melts/rMELTSframework.pyx":207
 * 
 * def cy_finalVolume():
 * 	return finalVolume()             # <<<<<<<<<<<<<<
 * 
 * def cy_setFinalVolume(double vValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(finalVolume()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":206
 * 	setInitialVolume(vValue)
 * 
 * def cy_finalVolume():             # <<<<<<<<<<<<<<
 * 	return finalVolume()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_finalVolume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":209
 * 	return finalVolume()
 * 
 * def cy_setFinalVolume(double vValue):             # <<<<<<<<<<<<<<
 * 	setFinalVolume(vValue)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_57cy_setFinalVolume(PyObject *__pyx_self, PyObject *__pyx_arg_vValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_56cy_setFinalVolume[] = "cy_setFinalVolume(double vValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_57cy_setFinalVolume = {"cy_setFinalVolume", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_57cy_setFinalVolume, METH_O, __pyx_doc_5melts_15rMELTSframework_56cy_setFinalVolume};
static PyObject *__pyx_pw_5melts_15rMELTSframework_57cy_setFinalVolume(PyObject *__pyx_self, PyObject *__pyx_arg_vValue) {
  double __pyx_v_vValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFinalVolume (wrapper)", 0);
  assert(__pyx_arg_vValue); {
    __pyx_v_vValue = __pyx_PyFloat_AsDouble(__pyx_arg_vValue); if (unlikely((__pyx_v_vValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 209, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setFinalVolume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_56cy_setFinalVolume(__pyx_self, ((double)__pyx_v_vValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_56cy_setFinalVolume(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_vValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFinalVolume", 0);

  /* "melts/rMELTSframework.pyx":210
 * 
 * def cy_setFinalVolume(double vValue):
 * 	setFinalVolume(vValue)             # <<<<<<<<<<<<<<
 * 
 * def cy_incrementVolume():
 */
  setFinalVolume(__pyx_v_vValue);

  /* "melts/rMELTSframework.pyx":209
 * 	return finalVolume()
 * 
 * def cy_setFinalVolume(double vValue):             # <<<<<<<<<<<<<<
 * 	setFinalVolume(vValue)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":212
 * 	setFinalVolume(vValue)
 * 
 * def cy_incrementVolume():             # <<<<<<<<<<<<<<
 * 	return incrementVolume()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_59cy_incrementVolume(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_58cy_incrementVolume[] = "cy_incrementVolume()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_59cy_incrementVolume = {"cy_incrementVolume", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_59cy_incrementVolume, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_58cy_incrementVolume};
static PyObject *__pyx_pw_5melts_15rMELTSframework_59cy_incrementVolume(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_incrementVolume (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_58cy_incrementVolume(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_58cy_incrementVolume(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_incrementVolume", 0);

  /* "melts/rMELTSframework.pyx":213
 * 
 * def cy_incrementVolume():
 * 	return incrementVolume()             # <<<<<<<<<<<<<<
 * 
 * def cy_setIncrementVolume(double vValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(incrementVolume()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":212
 * 	setFinalVolume(vValue)
 * 
 * def cy_incrementVolume():             # <<<<<<<<<<<<<<
 * 	return incrementVolume()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_incrementVolume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":215
 * 	return incrementVolume()
 * 
 * def cy_setIncrementVolume(double vValue):             # <<<<<<<<<<<<<<
 * 	setIncrementVolume(vValue)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_61cy_setIncrementVolume(PyObject *__pyx_self, PyObject *__pyx_arg_vValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_60cy_setIncrementVolume[] = "cy_setIncrementVolume(double vValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_61cy_setIncrementVolume = {"cy_setIncrementVolume", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_61cy_setIncrementVolume, METH_O, __pyx_doc_5melts_15rMELTSframework_60cy_setIncrementVolume};
static PyObject *__pyx_pw_5melts_15rMELTSframework_61cy_setIncrementVolume(PyObject *__pyx_self, PyObject *__pyx_arg_vValue) {
  double __pyx_v_vValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIncrementVolume (wrapper)", 0);
  assert(__pyx_arg_vValue); {
    __pyx_v_vValue = __pyx_PyFloat_AsDouble(__pyx_arg_vValue); if (unlikely((__pyx_v_vValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 215, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setIncrementVolume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_60cy_setIncrementVolume(__pyx_self, ((double)__pyx_v_vValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_60cy_setIncrementVolume(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_vValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIncrementVolume", 0);

  /* "melts/rMELTSframework.pyx":216
 * 
 * def cy_setIncrementVolume(double vValue):
 * 	setIncrementVolume(vValue)             # <<<<<<<<<<<<<<
 * 
 * def cy_initialEnthalpy():
 */
  setIncrementVolume(__pyx_v_vValue);

  /* "melts/rMELTSframework.pyx":215
 * 	return incrementVolume()
 * 
 * def cy_setIncrementVolume(double vValue):             # <<<<<<<<<<<<<<
 * 	setIncrementVolume(vValue)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":218
 * 	setIncrementVolume(vValue)
 * 
 * def cy_initialEnthalpy():             # <<<<<<<<<<<<<<
 * 	return initialEnthalpy()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_63cy_initialEnthalpy(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_62cy_initialEnthalpy[] = "cy_initialEnthalpy()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_63cy_initialEnthalpy = {"cy_initialEnthalpy", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_63cy_initialEnthalpy, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_62cy_initialEnthalpy};
static PyObject *__pyx_pw_5melts_15rMELTSframework_63cy_initialEnthalpy(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_initialEnthalpy (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_62cy_initialEnthalpy(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_62cy_initialEnthalpy(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_initialEnthalpy", 0);

  /* "melts/rMELTSframework.pyx":219
 * 
 * def cy_initialEnthalpy():
 * 	return initialEnthalpy()             # <<<<<<<<<<<<<<
 * 
 * def cy_setInitialEnthalpy(double hValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(initialEnthalpy()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":218
 * 	setIncrementVolume(vValue)
 * 
 * def cy_initialEnthalpy():             # <<<<<<<<<<<<<<
 * 	return initialEnthalpy()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_initialEnthalpy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":221
 * 	return initialEnthalpy()
 * 
 * def cy_setInitialEnthalpy(double hValue):             # <<<<<<<<<<<<<<
 * 	setInitialEnthalpy(hValue)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_65cy_setInitialEnthalpy(PyObject *__pyx_self, PyObject *__pyx_arg_hValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_64cy_setInitialEnthalpy[] = "cy_setInitialEnthalpy(double hValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_65cy_setInitialEnthalpy = {"cy_setInitialEnthalpy", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_65cy_setInitialEnthalpy, METH_O, __pyx_doc_5melts_15rMELTSframework_64cy_setInitialEnthalpy};
static PyObject *__pyx_pw_5melts_15rMELTSframework_65cy_setInitialEnthalpy(PyObject *__pyx_self, PyObject *__pyx_arg_hValue) {
  double __pyx_v_hValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setInitialEnthalpy (wrapper)", 0);
  assert(__pyx_arg_hValue); {
    __pyx_v_hValue = __pyx_PyFloat_AsDouble(__pyx_arg_hValue); if (unlikely((__pyx_v_hValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 221, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setInitialEnthalpy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_64cy_setInitialEnthalpy(__pyx_self, ((double)__pyx_v_hValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_64cy_setInitialEnthalpy(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_hValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setInitialEnthalpy", 0);

  /* "melts/rMELTSframework.pyx":222
 * 
 * def cy_setInitialEnthalpy(double hValue):
 * 	setInitialEnthalpy(hValue)             # <<<<<<<<<<<<<<
 * 
 * def cy_finalEnthalpy():
 */
  setInitialEnthalpy(__pyx_v_hValue);

  /* "melts/rMELTSframework.pyx":221
 * 	return initialEnthalpy()
 * 
 * def cy_setInitialEnthalpy(double hValue):             # <<<<<<<<<<<<<<
 * 	setInitialEnthalpy(hValue)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":224
 * 	setInitialEnthalpy(hValue)
 * 
 * def cy_finalEnthalpy():             # <<<<<<<<<<<<<<
 * 	return finalEnthalpy()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_67cy_finalEnthalpy(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_66cy_finalEnthalpy[] = "cy_finalEnthalpy()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_67cy_finalEnthalpy = {"cy_finalEnthalpy", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_67cy_finalEnthalpy, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_66cy_finalEnthalpy};
static PyObject *__pyx_pw_5melts_15rMELTSframework_67cy_finalEnthalpy(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_finalEnthalpy (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_66cy_finalEnthalpy(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_66cy_finalEnthalpy(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_finalEnthalpy", 0);

  /* "melts/rMELTSframework.pyx":225
 * 
 * def cy_finalEnthalpy():
 * 	return finalEnthalpy()             # <<<<<<<<<<<<<<
 * 
 * def cy_setFinalEnthalpy(double hValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(finalEnthalpy()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":224
 * 	setInitialEnthalpy(hValue)
 * 
 * def cy_finalEnthalpy():             # <<<<<<<<<<<<<<
 * 	return finalEnthalpy()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_finalEnthalpy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":227
 * 	return finalEnthalpy()
 * 
 * def cy_setFinalEnthalpy(double hValue):             # <<<<<<<<<<<<<<
 * 	setFinalEnthalpy(hValue)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_69cy_setFinalEnthalpy(PyObject *__pyx_self, PyObject *__pyx_arg_hValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_68cy_setFinalEnthalpy[] = "cy_setFinalEnthalpy(double hValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_69cy_setFinalEnthalpy = {"cy_setFinalEnthalpy", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_69cy_setFinalEnthalpy, METH_O, __pyx_doc_5melts_15rMELTSframework_68cy_setFinalEnthalpy};
static PyObject *__pyx_pw_5melts_15rMELTSframework_69cy_setFinalEnthalpy(PyObject *__pyx_self, PyObject *__pyx_arg_hValue) {
  double __pyx_v_hValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFinalEnthalpy (wrapper)", 0);
  assert(__pyx_arg_hValue); {
    __pyx_v_hValue = __pyx_PyFloat_AsDouble(__pyx_arg_hValue); if (unlikely((__pyx_v_hValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 227, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setFinalEnthalpy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_68cy_setFinalEnthalpy(__pyx_self, ((double)__pyx_v_hValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_68cy_setFinalEnthalpy(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_hValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFinalEnthalpy", 0);

  /* "melts/rMELTSframework.pyx":228
 * 
 * def cy_setFinalEnthalpy(double hValue):
 * 	setFinalEnthalpy(hValue)             # <<<<<<<<<<<<<<
 * 
 * def cy_incrementEnthalpy():
 */
  setFinalEnthalpy(__pyx_v_hValue);

  /* "melts/rMELTSframework.pyx":227
 * 	return finalEnthalpy()
 * 
 * def cy_setFinalEnthalpy(double hValue):             # <<<<<<<<<<<<<<
 * 	setFinalEnthalpy(hValue)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":230
 * 	setFinalEnthalpy(hValue)
 * 
 * def cy_incrementEnthalpy():             # <<<<<<<<<<<<<<
 * 	return incrementEnthalpy()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_71cy_incrementEnthalpy(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_70cy_incrementEnthalpy[] = "cy_incrementEnthalpy()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_71cy_incrementEnthalpy = {"cy_incrementEnthalpy", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_71cy_incrementEnthalpy, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_70cy_incrementEnthalpy};
static PyObject *__pyx_pw_5melts_15rMELTSframework_71cy_incrementEnthalpy(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_incrementEnthalpy (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_70cy_incrementEnthalpy(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_70cy_incrementEnthalpy(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_incrementEnthalpy", 0);

  /* "melts/rMELTSframework.pyx":231
 * 
 * def cy_incrementEnthalpy():
 * 	return incrementEnthalpy()             # <<<<<<<<<<<<<<
 * 
 * def cy_setIncrementEnthalpy(double hValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(incrementEnthalpy()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":230
 * 	setFinalEnthalpy(hValue)
 * 
 * def cy_incrementEnthalpy():             # <<<<<<<<<<<<<<
 * 	return incrementEnthalpy()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_incrementEnthalpy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":233
 * 	return incrementEnthalpy()
 * 
 * def cy_setIncrementEnthalpy(double hValue):             # <<<<<<<<<<<<<<
 * 	setIncrementEnthalpy(hValue)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_73cy_setIncrementEnthalpy(PyObject *__pyx_self, PyObject *__pyx_arg_hValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_72cy_setIncrementEnthalpy[] = "cy_setIncrementEnthalpy(double hValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_73cy_setIncrementEnthalpy = {"cy_setIncrementEnthalpy", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_73cy_setIncrementEnthalpy, METH_O, __pyx_doc_5melts_15rMELTSframework_72cy_setIncrementEnthalpy};
static PyObject *__pyx_pw_5melts_15rMELTSframework_73cy_setIncrementEnthalpy(PyObject *__pyx_self, PyObject *__pyx_arg_hValue) {
  double __pyx_v_hValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIncrementEnthalpy (wrapper)", 0);
  assert(__pyx_arg_hValue); {
    __pyx_v_hValue = __pyx_PyFloat_AsDouble(__pyx_arg_hValue); if (unlikely((__pyx_v_hValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 233, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setIncrementEnthalpy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_72cy_setIncrementEnthalpy(__pyx_self, ((double)__pyx_v_hValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_72cy_setIncrementEnthalpy(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_hValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIncrementEnthalpy", 0);

  /* "melts/rMELTSframework.pyx":234
 * 
 * def cy_setIncrementEnthalpy(double hValue):
 * 	setIncrementEnthalpy(hValue)             # <<<<<<<<<<<<<<
 * 
 * def cy_initialEntropy():
 */
  setIncrementEnthalpy(__pyx_v_hValue);

  /* "melts/rMELTSframework.pyx":233
 * 	return incrementEnthalpy()
 * 
 * def cy_setIncrementEnthalpy(double hValue):             # <<<<<<<<<<<<<<
 * 	setIncrementEnthalpy(hValue)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":236
 * 	setIncrementEnthalpy(hValue)
 * 
 * def cy_initialEntropy():             # <<<<<<<<<<<<<<
 * 	return initialEntropy()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_75cy_initialEntropy(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_74cy_initialEntropy[] = "cy_initialEntropy()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_75cy_initialEntropy = {"cy_initialEntropy", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_75cy_initialEntropy, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_74cy_initialEntropy};
static PyObject *__pyx_pw_5melts_15rMELTSframework_75cy_initialEntropy(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_initialEntropy (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_74cy_initialEntropy(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_74cy_initialEntropy(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_initialEntropy", 0);

  /* "melts/rMELTSframework.pyx":237
 * 
 * def cy_initialEntropy():
 * 	return initialEntropy()             # <<<<<<<<<<<<<<
 * 
 * def cy_setInitialEntropy(double sValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(initialEntropy()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":236
 * 	setIncrementEnthalpy(hValue)
 * 
 * def cy_initialEntropy():             # <<<<<<<<<<<<<<
 * 	return initialEntropy()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_initialEntropy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":239
 * 	return initialEntropy()
 * 
 * def cy_setInitialEntropy(double sValue):             # <<<<<<<<<<<<<<
 * 	setInitialEntropy(sValue)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_77cy_setInitialEntropy(PyObject *__pyx_self, PyObject *__pyx_arg_sValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_76cy_setInitialEntropy[] = "cy_setInitialEntropy(double sValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_77cy_setInitialEntropy = {"cy_setInitialEntropy", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_77cy_setInitialEntropy, METH_O, __pyx_doc_5melts_15rMELTSframework_76cy_setInitialEntropy};
static PyObject *__pyx_pw_5melts_15rMELTSframework_77cy_setInitialEntropy(PyObject *__pyx_self, PyObject *__pyx_arg_sValue) {
  double __pyx_v_sValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setInitialEntropy (wrapper)", 0);
  assert(__pyx_arg_sValue); {
    __pyx_v_sValue = __pyx_PyFloat_AsDouble(__pyx_arg_sValue); if (unlikely((__pyx_v_sValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 239, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setInitialEntropy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_76cy_setInitialEntropy(__pyx_self, ((double)__pyx_v_sValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_76cy_setInitialEntropy(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_sValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setInitialEntropy", 0);

  /* "melts/rMELTSframework.pyx":240
 * 
 * def cy_setInitialEntropy(double sValue):
 * 	setInitialEntropy(sValue)             # <<<<<<<<<<<<<<
 * 
 * def cy_finalEntropy():
 */
  setInitialEntropy(__pyx_v_sValue);

  /* "melts/rMELTSframework.pyx":239
 * 	return initialEntropy()
 * 
 * def cy_setInitialEntropy(double sValue):             # <<<<<<<<<<<<<<
 * 	setInitialEntropy(sValue)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":242
 * 	setInitialEntropy(sValue)
 * 
 * def cy_finalEntropy():             # <<<<<<<<<<<<<<
 * 	return finalEntropy()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_79cy_finalEntropy(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_78cy_finalEntropy[] = "cy_finalEntropy()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_79cy_finalEntropy = {"cy_finalEntropy", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_79cy_finalEntropy, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_78cy_finalEntropy};
static PyObject *__pyx_pw_5melts_15rMELTSframework_79cy_finalEntropy(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_finalEntropy (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_78cy_finalEntropy(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_78cy_finalEntropy(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_finalEntropy", 0);

  /* "melts/rMELTSframework.pyx":243
 * 
 * def cy_finalEntropy():
 * 	return finalEntropy()             # <<<<<<<<<<<<<<
 * 
 * def cy_setFinalEntropy(double sValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(finalEntropy()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 243, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":242
 * 	setInitialEntropy(sValue)
 * 
 * def cy_finalEntropy():             # <<<<<<<<<<<<<<
 * 	return finalEntropy()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_finalEntropy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":245
 * 	return finalEntropy()
 * 
 * def cy_setFinalEntropy(double sValue):             # <<<<<<<<<<<<<<
 * 	setFinalEntropy(sValue)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_81cy_setFinalEntropy(PyObject *__pyx_self, PyObject *__pyx_arg_sValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_80cy_setFinalEntropy[] = "cy_setFinalEntropy(double sValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_81cy_setFinalEntropy = {"cy_setFinalEntropy", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_81cy_setFinalEntropy, METH_O, __pyx_doc_5melts_15rMELTSframework_80cy_setFinalEntropy};
static PyObject *__pyx_pw_5melts_15rMELTSframework_81cy_setFinalEntropy(PyObject *__pyx_self, PyObject *__pyx_arg_sValue) {
  double __pyx_v_sValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFinalEntropy (wrapper)", 0);
  assert(__pyx_arg_sValue); {
    __pyx_v_sValue = __pyx_PyFloat_AsDouble(__pyx_arg_sValue); if (unlikely((__pyx_v_sValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 245, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setFinalEntropy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_80cy_setFinalEntropy(__pyx_self, ((double)__pyx_v_sValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_80cy_setFinalEntropy(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_sValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFinalEntropy", 0);

  /* "melts/rMELTSframework.pyx":246
 * 
 * def cy_setFinalEntropy(double sValue):
 * 	setFinalEntropy(sValue)             # <<<<<<<<<<<<<<
 * 
 * def cy_incrementEntropy():
 */
  setFinalEntropy(__pyx_v_sValue);

  /* "melts/rMELTSframework.pyx":245
 * 	return finalEntropy()
 * 
 * def cy_setFinalEntropy(double sValue):             # <<<<<<<<<<<<<<
 * 	setFinalEntropy(sValue)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":248
 * 	setFinalEntropy(sValue)
 * 
 * def cy_incrementEntropy():             # <<<<<<<<<<<<<<
 * 	return incrementEntropy()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_83cy_incrementEntropy(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_82cy_incrementEntropy[] = "cy_incrementEntropy()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_83cy_incrementEntropy = {"cy_incrementEntropy", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_83cy_incrementEntropy, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_82cy_incrementEntropy};
static PyObject *__pyx_pw_5melts_15rMELTSframework_83cy_incrementEntropy(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_incrementEntropy (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_82cy_incrementEntropy(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_82cy_incrementEntropy(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_incrementEntropy", 0);

  /* "melts/rMELTSframework.pyx":249
 * 
 * def cy_incrementEntropy():
 * 	return incrementEntropy()             # <<<<<<<<<<<<<<
 * 
 * def cy_setIncrementEntropy(double sValue):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(incrementEntropy()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":248
 * 	setFinalEntropy(sValue)
 * 
 * def cy_incrementEntropy():             # <<<<<<<<<<<<<<
 * 	return incrementEntropy()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_incrementEntropy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":251
 * 	return incrementEntropy()
 * 
 * def cy_setIncrementEntropy(double sValue):             # <<<<<<<<<<<<<<
 * 	setIncrementEntropy(sValue)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_85cy_setIncrementEntropy(PyObject *__pyx_self, PyObject *__pyx_arg_sValue); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_84cy_setIncrementEntropy[] = "cy_setIncrementEntropy(double sValue)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_85cy_setIncrementEntropy = {"cy_setIncrementEntropy", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_85cy_setIncrementEntropy, METH_O, __pyx_doc_5melts_15rMELTSframework_84cy_setIncrementEntropy};
static PyObject *__pyx_pw_5melts_15rMELTSframework_85cy_setIncrementEntropy(PyObject *__pyx_self, PyObject *__pyx_arg_sValue) {
  double __pyx_v_sValue;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIncrementEntropy (wrapper)", 0);
  assert(__pyx_arg_sValue); {
    __pyx_v_sValue = __pyx_PyFloat_AsDouble(__pyx_arg_sValue); if (unlikely((__pyx_v_sValue == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 251, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setIncrementEntropy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_84cy_setIncrementEntropy(__pyx_self, ((double)__pyx_v_sValue));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_84cy_setIncrementEntropy(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_sValue) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIncrementEntropy", 0);

  /* "melts/rMELTSframework.pyx":252
 * 
 * def cy_setIncrementEntropy(double sValue):
 * 	setIncrementEntropy(sValue)             # <<<<<<<<<<<<<<
 * 
 * def cy_isIsochoric():
 */
  setIncrementEntropy(__pyx_v_sValue);

  /* "melts/rMELTSframework.pyx":251
 * 	return incrementEntropy()
 * 
 * def cy_setIncrementEntropy(double sValue):             # <<<<<<<<<<<<<<
 * 	setIncrementEntropy(sValue)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":254
 * 	setIncrementEntropy(sValue)
 * 
 * def cy_isIsochoric():             # <<<<<<<<<<<<<<
 * 	return isIsochoric()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_87cy_isIsochoric(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_86cy_isIsochoric[] = "cy_isIsochoric()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_87cy_isIsochoric = {"cy_isIsochoric", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_87cy_isIsochoric, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_86cy_isIsochoric};
static PyObject *__pyx_pw_5melts_15rMELTSframework_87cy_isIsochoric(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_isIsochoric (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_86cy_isIsochoric(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_86cy_isIsochoric(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_isIsochoric", 0);

  /* "melts/rMELTSframework.pyx":255
 * 
 * def cy_isIsochoric():
 * 	return isIsochoric()             # <<<<<<<<<<<<<<
 * 
 * def cy_setIsochoric(int value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(isIsochoric()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":254
 * 	setIncrementEntropy(sValue)
 * 
 * def cy_isIsochoric():             # <<<<<<<<<<<<<<
 * 	return isIsochoric()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_isIsochoric", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":257
 * 	return isIsochoric()
 * 
 * def cy_setIsochoric(int value):             # <<<<<<<<<<<<<<
 * 	setIsochoric(value)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_89cy_setIsochoric(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_88cy_setIsochoric[] = "cy_setIsochoric(int value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_89cy_setIsochoric = {"cy_setIsochoric", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_89cy_setIsochoric, METH_O, __pyx_doc_5melts_15rMELTSframework_88cy_setIsochoric};
static PyObject *__pyx_pw_5melts_15rMELTSframework_89cy_setIsochoric(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  int __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIsochoric (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyInt_As_int(__pyx_arg_value); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 257, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setIsochoric", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_88cy_setIsochoric(__pyx_self, ((int)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_88cy_setIsochoric(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIsochoric", 0);

  /* "melts/rMELTSframework.pyx":258
 * 
 * def cy_setIsochoric(int value):
 * 	setIsochoric(value)             # <<<<<<<<<<<<<<
 * 
 * def cy_isIsenthalpic():
 */
  setIsochoric(__pyx_v_value);

  /* "melts/rMELTSframework.pyx":257
 * 	return isIsochoric()
 * 
 * def cy_setIsochoric(int value):             # <<<<<<<<<<<<<<
 * 	setIsochoric(value)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":260
 * 	setIsochoric(value)
 * 
 * def cy_isIsenthalpic():             # <<<<<<<<<<<<<<
 * 	return isIsenthalpic()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_91cy_isIsenthalpic(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_90cy_isIsenthalpic[] = "cy_isIsenthalpic()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_91cy_isIsenthalpic = {"cy_isIsenthalpic", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_91cy_isIsenthalpic, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_90cy_isIsenthalpic};
static PyObject *__pyx_pw_5melts_15rMELTSframework_91cy_isIsenthalpic(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_isIsenthalpic (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_90cy_isIsenthalpic(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_90cy_isIsenthalpic(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_isIsenthalpic", 0);

  /* "melts/rMELTSframework.pyx":261
 * 
 * def cy_isIsenthalpic():
 * 	return isIsenthalpic()             # <<<<<<<<<<<<<<
 * 
 * def cy_setIsenthalpic(int value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(isIsenthalpic()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":260
 * 	setIsochoric(value)
 * 
 * def cy_isIsenthalpic():             # <<<<<<<<<<<<<<
 * 	return isIsenthalpic()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_isIsenthalpic", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":263
 * 	return isIsenthalpic()
 * 
 * def cy_setIsenthalpic(int value):             # <<<<<<<<<<<<<<
 * 	setIsenthalpic(value)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_93cy_setIsenthalpic(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_92cy_setIsenthalpic[] = "cy_setIsenthalpic(int value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_93cy_setIsenthalpic = {"cy_setIsenthalpic", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_93cy_setIsenthalpic, METH_O, __pyx_doc_5melts_15rMELTSframework_92cy_setIsenthalpic};
static PyObject *__pyx_pw_5melts_15rMELTSframework_93cy_setIsenthalpic(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  int __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIsenthalpic (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyInt_As_int(__pyx_arg_value); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 263, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setIsenthalpic", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_92cy_setIsenthalpic(__pyx_self, ((int)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_92cy_setIsenthalpic(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIsenthalpic", 0);

  /* "melts/rMELTSframework.pyx":264
 * 
 * def cy_setIsenthalpic(int value):
 * 	setIsenthalpic(value)             # <<<<<<<<<<<<<<
 * 
 * def cy_isIsotropic():
 */
  setIsenthalpic(__pyx_v_value);

  /* "melts/rMELTSframework.pyx":263
 * 	return isIsenthalpic()
 * 
 * def cy_setIsenthalpic(int value):             # <<<<<<<<<<<<<<
 * 	setIsenthalpic(value)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":266
 * 	setIsenthalpic(value)
 * 
 * def cy_isIsotropic():             # <<<<<<<<<<<<<<
 * 	return isIsotropic()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_95cy_isIsotropic(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_94cy_isIsotropic[] = "cy_isIsotropic()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_95cy_isIsotropic = {"cy_isIsotropic", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_95cy_isIsotropic, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_94cy_isIsotropic};
static PyObject *__pyx_pw_5melts_15rMELTSframework_95cy_isIsotropic(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_isIsotropic (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_94cy_isIsotropic(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_94cy_isIsotropic(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_isIsotropic", 0);

  /* "melts/rMELTSframework.pyx":267
 * 
 * def cy_isIsotropic():
 * 	return isIsotropic()             # <<<<<<<<<<<<<<
 * 
 * def cy_setIsotropic(int value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(isIsotropic()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":266
 * 	setIsenthalpic(value)
 * 
 * def cy_isIsotropic():             # <<<<<<<<<<<<<<
 * 	return isIsotropic()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_isIsotropic", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":269
 * 	return isIsotropic()
 * 
 * def cy_setIsotropic(int value):             # <<<<<<<<<<<<<<
 * 	setIsotropic(value)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_97cy_setIsotropic(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_96cy_setIsotropic[] = "cy_setIsotropic(int value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_97cy_setIsotropic = {"cy_setIsotropic", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_97cy_setIsotropic, METH_O, __pyx_doc_5melts_15rMELTSframework_96cy_setIsotropic};
static PyObject *__pyx_pw_5melts_15rMELTSframework_97cy_setIsotropic(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  int __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIsotropic (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyInt_As_int(__pyx_arg_value); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 269, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setIsotropic", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_96cy_setIsotropic(__pyx_self, ((int)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_96cy_setIsotropic(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIsotropic", 0);

  /* "melts/rMELTSframework.pyx":270
 * 
 * def cy_setIsotropic(int value):
 * 	setIsotropic(value)             # <<<<<<<<<<<<<<
 * 
 * def cy_bulkCompAsList():
 */
  setIsotropic(__pyx_v_value);

  /* "melts/rMELTSframework.pyx":269
 * 	return isIsotropic()
 * 
 * def cy_setIsotropic(int value):             # <<<<<<<<<<<<<<
 * 	setIsotropic(value)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":272
 * 	setIsotropic(value)
 * 
 * def cy_bulkCompAsList():             # <<<<<<<<<<<<<<
 * 	cdef double *bc = bulkCompAsPtr()
 * 	n = num_oxides()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_99cy_bulkCompAsList(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_98cy_bulkCompAsList[] = "cy_bulkCompAsList()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_99cy_bulkCompAsList = {"cy_bulkCompAsList", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_99cy_bulkCompAsList, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_98cy_bulkCompAsList};
static PyObject *__pyx_pw_5melts_15rMELTSframework_99cy_bulkCompAsList(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_bulkCompAsList (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_98cy_bulkCompAsList(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_98cy_bulkCompAsList(CYTHON_UNUSED PyObject *__pyx_self) {
  double *__pyx_v_bc;
  int __pyx_v_n;
  PyObject *__pyx_v_result = NULL;
  long __pyx_v_i;
  double __pyx_v_entry;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  long __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_bulkCompAsList", 0);

  /* "melts/rMELTSframework.pyx":273
 * 
 * def cy_bulkCompAsList():
 * 	cdef double *bc = bulkCompAsPtr()             # <<<<<<<<<<<<<<
 * 	n = num_oxides()
 * 	result = []
 */
  __pyx_v_bc = bulkCompAsPtr();

  /* "melts/rMELTSframework.pyx":274
 * def cy_bulkCompAsList():
 * 	cdef double *bc = bulkCompAsPtr()
 * 	n = num_oxides()             # <<<<<<<<<<<<<<
 * 	result = []
 * 	for i in range(0,n):
 */
  __pyx_v_n = num_oxides();

  /* "melts/rMELTSframework.pyx":275
 * 	cdef double *bc = bulkCompAsPtr()
 * 	n = num_oxides()
 * 	result = []             # <<<<<<<<<<<<<<
 * 	for i in range(0,n):
 * 		entry = <double> bc[i]
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":276
 * 	n = num_oxides()
 * 	result = []
 * 	for i in range(0,n):             # <<<<<<<<<<<<<<
 * 		entry = <double> bc[i]
 * 		result.append(entry)
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "melts/rMELTSframework.pyx":277
 * 	result = []
 * 	for i in range(0,n):
 * 		entry = <double> bc[i]             # <<<<<<<<<<<<<<
 * 		result.append(entry)
 * 	return result
 */
    __pyx_v_entry = ((double)(__pyx_v_bc[__pyx_v_i]));

    /* "melts/rMELTSframework.pyx":278
 * 	for i in range(0,n):
 * 		entry = <double> bc[i]
 * 		result.append(entry)             # <<<<<<<<<<<<<<
 * 	return result
 * 
 */
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_entry); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 278, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 278, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "melts/rMELTSframework.pyx":279
 * 		entry = <double> bc[i]
 * 		result.append(entry)
 * 	return result             # <<<<<<<<<<<<<<
 * 
 * def cy_setBulkCompAsList(list values = []):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":272
 * 	setIsotropic(value)
 * 
 * def cy_bulkCompAsList():             # <<<<<<<<<<<<<<
 * 	cdef double *bc = bulkCompAsPtr()
 * 	n = num_oxides()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_bulkCompAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":281
 * 	return result
 * 
 * def cy_setBulkCompAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	n = num_oxides()
 * 	cdef double *bc = <double *>malloc(n*sizeof(double))
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_101cy_setBulkCompAsList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_100cy_setBulkCompAsList[] = "cy_setBulkCompAsList(list values=[])";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_101cy_setBulkCompAsList = {"cy_setBulkCompAsList", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5melts_15rMELTSframework_101cy_setBulkCompAsList, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5melts_15rMELTSframework_100cy_setBulkCompAsList};
static PyObject *__pyx_pw_5melts_15rMELTSframework_101cy_setBulkCompAsList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_values = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setBulkCompAsList (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_values,0};
    PyObject* values[1] = {0};
    values[0] = __pyx_k_;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cy_setBulkCompAsList") < 0)) __PYX_ERR(0, 281, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_values = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cy_setBulkCompAsList", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 281, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setBulkCompAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_values), (&PyList_Type), 1, "values", 1))) __PYX_ERR(0, 281, __pyx_L1_error)
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_100cy_setBulkCompAsList(__pyx_self, __pyx_v_values);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_100cy_setBulkCompAsList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values) {
  PyObject *__pyx_v_n = NULL;
  double *__pyx_v_bc;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  double __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_setBulkCompAsList", 0);

  /* "melts/rMELTSframework.pyx":282
 * 
 * def cy_setBulkCompAsList(list values = []):
 * 	n = num_oxides()             # <<<<<<<<<<<<<<
 * 	cdef double *bc = <double *>malloc(n*sizeof(double))
 * 	for i in range (0,n):
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(num_oxides()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_n = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":283
 * def cy_setBulkCompAsList(list values = []):
 * 	n = num_oxides()
 * 	cdef double *bc = <double *>malloc(n*sizeof(double))             # <<<<<<<<<<<<<<
 * 	for i in range (0,n):
 * 		bc[i] = <double> values[<int>i]
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t((sizeof(double))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_n, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 283, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_bc = ((double *)malloc(__pyx_t_3));

  /* "melts/rMELTSframework.pyx":284
 * 	n = num_oxides()
 * 	cdef double *bc = <double *>malloc(n*sizeof(double))
 * 	for i in range (0,n):             # <<<<<<<<<<<<<<
 * 		bc[i] = <double> values[<int>i]
 * 	setBulkCompAsPtr(<double *> bc)
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_0);
  __Pyx_INCREF(__pyx_v_n);
  __Pyx_GIVEREF(__pyx_v_n);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_n);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 284, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 284, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 284, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 284, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "melts/rMELTSframework.pyx":285
 * 	cdef double *bc = <double *>malloc(n*sizeof(double))
 * 	for i in range (0,n):
 * 		bc[i] = <double> values[<int>i]             # <<<<<<<<<<<<<<
 * 	setBulkCompAsPtr(<double *> bc)
 * 	# free (bc) memory management in C function
 */
    if (unlikely(__pyx_v_values == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 285, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_i); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 285, __pyx_L1_error)
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_values, ((int)__pyx_t_6), int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 285, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_8 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 285, __pyx_L1_error)
    (__pyx_v_bc[__pyx_t_8]) = ((double)__pyx_t_7);

    /* "melts/rMELTSframework.pyx":284
 * 	n = num_oxides()
 * 	cdef double *bc = <double *>malloc(n*sizeof(double))
 * 	for i in range (0,n):             # <<<<<<<<<<<<<<
 * 		bc[i] = <double> values[<int>i]
 * 	setBulkCompAsPtr(<double *> bc)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "melts/rMELTSframework.pyx":286
 * 	for i in range (0,n):
 * 		bc[i] = <double> values[<int>i]
 * 	setBulkCompAsPtr(<double *> bc)             # <<<<<<<<<<<<<<
 * 	# free (bc) memory management in C function
 * 
 */
  setBulkCompAsPtr(((double *)__pyx_v_bc));

  /* "melts/rMELTSframework.pyx":281
 * 	return result
 * 
 * def cy_setBulkCompAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	n = num_oxides()
 * 	cdef double *bc = <double *>malloc(n*sizeof(double))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setBulkCompAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_n);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":289
 * 	# free (bc) memory management in C function
 * 
 * def cy_liquidCompAsListOfLists():             # <<<<<<<<<<<<<<
 * 	cdef double **lc = liquidCompAsPtr()
 * 	nl = nLiquidCoexist()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_103cy_liquidCompAsListOfLists(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_102cy_liquidCompAsListOfLists[] = "cy_liquidCompAsListOfLists()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_103cy_liquidCompAsListOfLists = {"cy_liquidCompAsListOfLists", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_103cy_liquidCompAsListOfLists, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_102cy_liquidCompAsListOfLists};
static PyObject *__pyx_pw_5melts_15rMELTSframework_103cy_liquidCompAsListOfLists(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_liquidCompAsListOfLists (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_102cy_liquidCompAsListOfLists(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_102cy_liquidCompAsListOfLists(CYTHON_UNUSED PyObject *__pyx_self) {
  double **__pyx_v_lc;
  int __pyx_v_nl;
  int __pyx_v_nc;
  PyObject *__pyx_v_result = NULL;
  long __pyx_v_i;
  long __pyx_7genexpr__pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  long __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  long __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_liquidCompAsListOfLists", 0);

  /* "melts/rMELTSframework.pyx":290
 * 
 * def cy_liquidCompAsListOfLists():
 * 	cdef double **lc = liquidCompAsPtr()             # <<<<<<<<<<<<<<
 * 	nl = nLiquidCoexist()
 * 	nc = num_oxides()
 */
  __pyx_v_lc = liquidCompAsPtr();

  /* "melts/rMELTSframework.pyx":291
 * def cy_liquidCompAsListOfLists():
 * 	cdef double **lc = liquidCompAsPtr()
 * 	nl = nLiquidCoexist()             # <<<<<<<<<<<<<<
 * 	nc = num_oxides()
 * 	result = []
 */
  __pyx_v_nl = nLiquidCoexist();

  /* "melts/rMELTSframework.pyx":292
 * 	cdef double **lc = liquidCompAsPtr()
 * 	nl = nLiquidCoexist()
 * 	nc = num_oxides()             # <<<<<<<<<<<<<<
 * 	result = []
 * 	for i in range(0, nl):
 */
  __pyx_v_nc = num_oxides();

  /* "melts/rMELTSframework.pyx":293
 * 	nl = nLiquidCoexist()
 * 	nc = num_oxides()
 * 	result = []             # <<<<<<<<<<<<<<
 * 	for i in range(0, nl):
 * 		result.append([lc[i][j] for j in range(0, nc)])
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":294
 * 	nc = num_oxides()
 * 	result = []
 * 	for i in range(0, nl):             # <<<<<<<<<<<<<<
 * 		result.append([lc[i][j] for j in range(0, nc)])
 * 	return result
 */
  __pyx_t_2 = __pyx_v_nl;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "melts/rMELTSframework.pyx":295
 * 	result = []
 * 	for i in range(0, nl):
 * 		result.append([lc[i][j] for j in range(0, nc)])             # <<<<<<<<<<<<<<
 * 	return result
 * 
 */
    { /* enter inner scope */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 295, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __pyx_v_nc;
      __pyx_t_6 = __pyx_t_5;
      for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
        __pyx_7genexpr__pyx_v_j = __pyx_t_7;
        __pyx_t_8 = PyFloat_FromDouble(((__pyx_v_lc[__pyx_v_i])[__pyx_7genexpr__pyx_v_j])); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 295, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 295, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
    } /* exit inner scope */
    __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 295, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "melts/rMELTSframework.pyx":296
 * 	for i in range(0, nl):
 * 		result.append([lc[i][j] for j in range(0, nc)])
 * 	return result             # <<<<<<<<<<<<<<
 * 
 * def cy_setLiquidCompAsListOfLists(list values = [[]]):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":289
 * 	# free (bc) memory management in C function
 * 
 * def cy_liquidCompAsListOfLists():             # <<<<<<<<<<<<<<
 * 	cdef double **lc = liquidCompAsPtr()
 * 	nl = nLiquidCoexist()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_liquidCompAsListOfLists", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":298
 * 	return result
 * 
 * def cy_setLiquidCompAsListOfLists(list values = [[]]):             # <<<<<<<<<<<<<<
 * 	nl = nLiquidCoexist()
 * 	nc = num_oxides()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_105cy_setLiquidCompAsListOfLists(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_104cy_setLiquidCompAsListOfLists[] = "cy_setLiquidCompAsListOfLists(list values=[[]])";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_105cy_setLiquidCompAsListOfLists = {"cy_setLiquidCompAsListOfLists", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5melts_15rMELTSframework_105cy_setLiquidCompAsListOfLists, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5melts_15rMELTSframework_104cy_setLiquidCompAsListOfLists};
static PyObject *__pyx_pw_5melts_15rMELTSframework_105cy_setLiquidCompAsListOfLists(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_values = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setLiquidCompAsListOfLists (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_values,0};
    PyObject* values[1] = {0};
    values[0] = __pyx_k__2;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cy_setLiquidCompAsListOfLists") < 0)) __PYX_ERR(0, 298, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_values = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cy_setLiquidCompAsListOfLists", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 298, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setLiquidCompAsListOfLists", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_values), (&PyList_Type), 1, "values", 1))) __PYX_ERR(0, 298, __pyx_L1_error)
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_104cy_setLiquidCompAsListOfLists(__pyx_self, __pyx_v_values);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_104cy_setLiquidCompAsListOfLists(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values) {
  PyObject *__pyx_v_nl = NULL;
  PyObject *__pyx_v_nc = NULL;
  double **__pyx_v_lc;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  double __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_setLiquidCompAsListOfLists", 0);

  /* "melts/rMELTSframework.pyx":299
 * 
 * def cy_setLiquidCompAsListOfLists(list values = [[]]):
 * 	nl = nLiquidCoexist()             # <<<<<<<<<<<<<<
 * 	nc = num_oxides()
 * 	cdef double **lc = <double **>malloc(nl*sizeof(double *))
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(nLiquidCoexist()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_nl = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":300
 * def cy_setLiquidCompAsListOfLists(list values = [[]]):
 * 	nl = nLiquidCoexist()
 * 	nc = num_oxides()             # <<<<<<<<<<<<<<
 * 	cdef double **lc = <double **>malloc(nl*sizeof(double *))
 * 	for i in range(0, nl):
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(num_oxides()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_nc = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":301
 * 	nl = nLiquidCoexist()
 * 	nc = num_oxides()
 * 	cdef double **lc = <double **>malloc(nl*sizeof(double *))             # <<<<<<<<<<<<<<
 * 	for i in range(0, nl):
 * 		lc[i] = <double *>malloc(nc*sizeof(double))
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t((sizeof(double *))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_nl, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_lc = ((double **)malloc(__pyx_t_3));

  /* "melts/rMELTSframework.pyx":302
 * 	nc = num_oxides()
 * 	cdef double **lc = <double **>malloc(nl*sizeof(double *))
 * 	for i in range(0, nl):             # <<<<<<<<<<<<<<
 * 		lc[i] = <double *>malloc(nc*sizeof(double))
 * 		for j in range(0, nc):
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_0);
  __Pyx_INCREF(__pyx_v_nl);
  __Pyx_GIVEREF(__pyx_v_nl);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_nl);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 302, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 302, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 302, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 302, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 302, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "melts/rMELTSframework.pyx":303
 * 	cdef double **lc = <double **>malloc(nl*sizeof(double *))
 * 	for i in range(0, nl):
 * 		lc[i] = <double *>malloc(nc*sizeof(double))             # <<<<<<<<<<<<<<
 * 		for j in range(0, nc):
 * 			lc[i][j] = values[<int>i][<int>j]
 */
    __pyx_t_1 = __Pyx_PyInt_FromSize_t((sizeof(double))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyNumber_Multiply(__pyx_v_nc, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __Pyx_PyInt_As_size_t(__pyx_t_6); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_7 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 303, __pyx_L1_error)
    (__pyx_v_lc[__pyx_t_7]) = ((double *)malloc(__pyx_t_3));

    /* "melts/rMELTSframework.pyx":304
 * 	for i in range(0, nl):
 * 		lc[i] = <double *>malloc(nc*sizeof(double))
 * 		for j in range(0, nc):             # <<<<<<<<<<<<<<
 * 			lc[i][j] = values[<int>i][<int>j]
 * 	setLiquidCompAsPtr(<double **> lc)
 */
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_0);
    __Pyx_INCREF(__pyx_v_nc);
    __Pyx_GIVEREF(__pyx_v_nc);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_nc);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_6 = __pyx_t_1; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 304, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 304, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_6))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 304, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 304, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_8(__pyx_t_6);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 304, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "melts/rMELTSframework.pyx":305
 * 		lc[i] = <double *>malloc(nc*sizeof(double))
 * 		for j in range(0, nc):
 * 			lc[i][j] = values[<int>i][<int>j]             # <<<<<<<<<<<<<<
 * 	setLiquidCompAsPtr(<double **> lc)
 * 	#for i in range(0, nl):
 */
      if (unlikely(__pyx_v_values == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 305, __pyx_L1_error)
      }
      __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_v_i); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L1_error)
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_values, ((int)__pyx_t_9), int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_v_j); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L1_error)
      __pyx_t_10 = __Pyx_GetItemInt(__pyx_t_1, ((int)__pyx_t_9), int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_t_10); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L1_error)
      __pyx_t_13 = __Pyx_PyIndex_AsSsize_t(__pyx_v_j); if (unlikely((__pyx_t_13 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 305, __pyx_L1_error)
      ((__pyx_v_lc[__pyx_t_12])[__pyx_t_13]) = __pyx_t_11;

      /* "melts/rMELTSframework.pyx":304
 * 	for i in range(0, nl):
 * 		lc[i] = <double *>malloc(nc*sizeof(double))
 * 		for j in range(0, nc):             # <<<<<<<<<<<<<<
 * 			lc[i][j] = values[<int>i][<int>j]
 * 	setLiquidCompAsPtr(<double **> lc)
 */
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "melts/rMELTSframework.pyx":302
 * 	nc = num_oxides()
 * 	cdef double **lc = <double **>malloc(nl*sizeof(double *))
 * 	for i in range(0, nl):             # <<<<<<<<<<<<<<
 * 		lc[i] = <double *>malloc(nc*sizeof(double))
 * 		for j in range(0, nc):
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "melts/rMELTSframework.pyx":306
 * 		for j in range(0, nc):
 * 			lc[i][j] = values[<int>i][<int>j]
 * 	setLiquidCompAsPtr(<double **> lc)             # <<<<<<<<<<<<<<
 * 	#for i in range(0, nl):
 * 	#	free(lc[i])
 */
  setLiquidCompAsPtr(((double **)__pyx_v_lc));

  /* "melts/rMELTSframework.pyx":298
 * 	return result
 * 
 * def cy_setLiquidCompAsListOfLists(list values = [[]]):             # <<<<<<<<<<<<<<
 * 	nl = nLiquidCoexist()
 * 	nc = num_oxides()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setLiquidCompAsListOfLists", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nl);
  __Pyx_XDECREF(__pyx_v_nc);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":311
 * 	#free (lc)  memory management in C function
 * 
 * def cy_nLiquidCoexist():             # <<<<<<<<<<<<<<
 * 	return nLiquidCoexist()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_107cy_nLiquidCoexist(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_106cy_nLiquidCoexist[] = "cy_nLiquidCoexist()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_107cy_nLiquidCoexist = {"cy_nLiquidCoexist", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_107cy_nLiquidCoexist, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_106cy_nLiquidCoexist};
static PyObject *__pyx_pw_5melts_15rMELTSframework_107cy_nLiquidCoexist(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_nLiquidCoexist (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_106cy_nLiquidCoexist(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_106cy_nLiquidCoexist(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_nLiquidCoexist", 0);

  /* "melts/rMELTSframework.pyx":312
 * 
 * def cy_nLiquidCoexist():
 * 	return nLiquidCoexist()             # <<<<<<<<<<<<<<
 * 
 * def cy_setnLiquidCoexist(int value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(nLiquidCoexist()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":311
 * 	#free (lc)  memory management in C function
 * 
 * def cy_nLiquidCoexist():             # <<<<<<<<<<<<<<
 * 	return nLiquidCoexist()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_nLiquidCoexist", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":314
 * 	return nLiquidCoexist()
 * 
 * def cy_setnLiquidCoexist(int value):             # <<<<<<<<<<<<<<
 * 	setnLiquidCoexist(<int> value)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_109cy_setnLiquidCoexist(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_108cy_setnLiquidCoexist[] = "cy_setnLiquidCoexist(int value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_109cy_setnLiquidCoexist = {"cy_setnLiquidCoexist", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_109cy_setnLiquidCoexist, METH_O, __pyx_doc_5melts_15rMELTSframework_108cy_setnLiquidCoexist};
static PyObject *__pyx_pw_5melts_15rMELTSframework_109cy_setnLiquidCoexist(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  int __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setnLiquidCoexist (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyInt_As_int(__pyx_arg_value); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 314, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setnLiquidCoexist", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_108cy_setnLiquidCoexist(__pyx_self, ((int)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_108cy_setnLiquidCoexist(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setnLiquidCoexist", 0);

  /* "melts/rMELTSframework.pyx":315
 * 
 * def cy_setnLiquidCoexist(int value):
 * 	setnLiquidCoexist(<int> value)             # <<<<<<<<<<<<<<
 * 
 * def cy_liquidMass():
 */
  setnLiquidCoexist(((int)__pyx_v_value));

  /* "melts/rMELTSframework.pyx":314
 * 	return nLiquidCoexist()
 * 
 * def cy_setnLiquidCoexist(int value):             # <<<<<<<<<<<<<<
 * 	setnLiquidCoexist(<int> value)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":317
 * 	setnLiquidCoexist(<int> value)
 * 
 * def cy_liquidMass():             # <<<<<<<<<<<<<<
 * 	return <double> liquidMass()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_111cy_liquidMass(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_110cy_liquidMass[] = "cy_liquidMass()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_111cy_liquidMass = {"cy_liquidMass", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_111cy_liquidMass, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_110cy_liquidMass};
static PyObject *__pyx_pw_5melts_15rMELTSframework_111cy_liquidMass(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_liquidMass (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_110cy_liquidMass(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_110cy_liquidMass(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_liquidMass", 0);

  /* "melts/rMELTSframework.pyx":318
 * 
 * def cy_liquidMass():
 * 	return <double> liquidMass()             # <<<<<<<<<<<<<<
 * 
 * def cy_setLiquidMass(double value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(((double)liquidMass())); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":317
 * 	setnLiquidCoexist(<int> value)
 * 
 * def cy_liquidMass():             # <<<<<<<<<<<<<<
 * 	return <double> liquidMass()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_liquidMass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":320
 * 	return <double> liquidMass()
 * 
 * def cy_setLiquidMass(double value):             # <<<<<<<<<<<<<<
 * 	setLiquidMass(<double> value);
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_113cy_setLiquidMass(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_112cy_setLiquidMass[] = "cy_setLiquidMass(double value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_113cy_setLiquidMass = {"cy_setLiquidMass", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_113cy_setLiquidMass, METH_O, __pyx_doc_5melts_15rMELTSframework_112cy_setLiquidMass};
static PyObject *__pyx_pw_5melts_15rMELTSframework_113cy_setLiquidMass(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setLiquidMass (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 320, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setLiquidMass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_112cy_setLiquidMass(__pyx_self, ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_112cy_setLiquidMass(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setLiquidMass", 0);

  /* "melts/rMELTSframework.pyx":321
 * 
 * def cy_setLiquidMass(double value):
 * 	setLiquidMass(<double> value);             # <<<<<<<<<<<<<<
 * 
 * def cy_multipleLiqs():
 */
  setLiquidMass(((double)__pyx_v_value));

  /* "melts/rMELTSframework.pyx":320
 * 	return <double> liquidMass()
 * 
 * def cy_setLiquidMass(double value):             # <<<<<<<<<<<<<<
 * 	setLiquidMass(<double> value);
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":323
 * 	setLiquidMass(<double> value);
 * 
 * def cy_multipleLiqs():             # <<<<<<<<<<<<<<
 * 	return False if multipleLiqs() == 0 else True;
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_115cy_multipleLiqs(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_114cy_multipleLiqs[] = "cy_multipleLiqs()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_115cy_multipleLiqs = {"cy_multipleLiqs", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_115cy_multipleLiqs, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_114cy_multipleLiqs};
static PyObject *__pyx_pw_5melts_15rMELTSframework_115cy_multipleLiqs(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_multipleLiqs (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_114cy_multipleLiqs(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_114cy_multipleLiqs(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cy_multipleLiqs", 0);

  /* "melts/rMELTSframework.pyx":324
 * 
 * def cy_multipleLiqs():
 * 	return False if multipleLiqs() == 0 else True;             # <<<<<<<<<<<<<<
 * 
 * def cy_setMultipleLiqs(bint value):
 */
  __Pyx_XDECREF(__pyx_r);
  if (((multipleLiqs() == 0) != 0)) {
    __Pyx_INCREF(Py_False);
    __pyx_t_1 = Py_False;
  } else {
    __Pyx_INCREF(Py_True);
    __pyx_t_1 = Py_True;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":323
 * 	setLiquidMass(<double> value);
 * 
 * def cy_multipleLiqs():             # <<<<<<<<<<<<<<
 * 	return False if multipleLiqs() == 0 else True;
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":326
 * 	return False if multipleLiqs() == 0 else True;
 * 
 * def cy_setMultipleLiqs(bint value):             # <<<<<<<<<<<<<<
 * 	setMultipleLiqs(<int> (1 if value else 0))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_117cy_setMultipleLiqs(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_116cy_setMultipleLiqs[] = "cy_setMultipleLiqs(bool value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_117cy_setMultipleLiqs = {"cy_setMultipleLiqs", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_117cy_setMultipleLiqs, METH_O, __pyx_doc_5melts_15rMELTSframework_116cy_setMultipleLiqs};
static PyObject *__pyx_pw_5melts_15rMELTSframework_117cy_setMultipleLiqs(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  int __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setMultipleLiqs (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyObject_IsTrue(__pyx_arg_value); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 326, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setMultipleLiqs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_116cy_setMultipleLiqs(__pyx_self, ((int)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_116cy_setMultipleLiqs(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  __Pyx_RefNannySetupContext("cy_setMultipleLiqs", 0);

  /* "melts/rMELTSframework.pyx":327
 * 
 * def cy_setMultipleLiqs(bint value):
 * 	setMultipleLiqs(<int> (1 if value else 0))             # <<<<<<<<<<<<<<
 * 
 * def cy_nSolidCoexistAsList():
 */
  if ((__pyx_v_value != 0)) {
    __pyx_t_1 = 1;
  } else {
    __pyx_t_1 = 0;
  }
  setMultipleLiqs(((int)__pyx_t_1));

  /* "melts/rMELTSframework.pyx":326
 * 	return False if multipleLiqs() == 0 else True;
 * 
 * def cy_setMultipleLiqs(bint value):             # <<<<<<<<<<<<<<
 * 	setMultipleLiqs(<int> (1 if value else 0))
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":329
 * 	setMultipleLiqs(<int> (1 if value else 0))
 * 
 * def cy_nSolidCoexistAsList():             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_119cy_nSolidCoexistAsList(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_118cy_nSolidCoexistAsList[] = "cy_nSolidCoexistAsList()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_119cy_nSolidCoexistAsList = {"cy_nSolidCoexistAsList", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_119cy_nSolidCoexistAsList, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_118cy_nSolidCoexistAsList};
static PyObject *__pyx_pw_5melts_15rMELTSframework_119cy_nSolidCoexistAsList(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_nSolidCoexistAsList (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_118cy_nSolidCoexistAsList(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_118cy_nSolidCoexistAsList(CYTHON_UNUSED PyObject *__pyx_self) {
  int *__pyx_v_ns;
  int __pyx_v_npc;
  PyObject *__pyx_v_result = NULL;
  long __pyx_8genexpr1__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  long __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_nSolidCoexistAsList", 0);

  /* "melts/rMELTSframework.pyx":330
 * 
 * def cy_nSolidCoexistAsList():
 * 	cdef int *ns = nSolidCoexist()             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	result = [<int> ns[i] for i in range(0,npc)]
 */
  __pyx_v_ns = nSolidCoexist();

  /* "melts/rMELTSframework.pyx":331
 * def cy_nSolidCoexistAsList():
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()             # <<<<<<<<<<<<<<
 * 	result = [<int> ns[i] for i in range(0,npc)]
 * 	return result
 */
  __pyx_v_npc = num_phase_components();

  /* "melts/rMELTSframework.pyx":332
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()
 * 	result = [<int> ns[i] for i in range(0,npc)]             # <<<<<<<<<<<<<<
 * 	return result
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 332, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_npc;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_8genexpr1__pyx_v_i = __pyx_t_4;
      __pyx_t_5 = __Pyx_PyInt_From_int(((int)(__pyx_v_ns[__pyx_8genexpr1__pyx_v_i]))); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 332, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  } /* exit inner scope */
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":333
 * 	npc = num_phase_components()
 * 	result = [<int> ns[i] for i in range(0,npc)]
 * 	return result             # <<<<<<<<<<<<<<
 * 
 * def cy_setnSolidCoexistAsList(list values = []):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":329
 * 	setMultipleLiqs(<int> (1 if value else 0))
 * 
 * def cy_nSolidCoexistAsList():             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_nSolidCoexistAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":335
 * 	return result
 * 
 * def cy_setnSolidCoexistAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_121cy_setnSolidCoexistAsList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_120cy_setnSolidCoexistAsList[] = "cy_setnSolidCoexistAsList(list values=[])";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_121cy_setnSolidCoexistAsList = {"cy_setnSolidCoexistAsList", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5melts_15rMELTSframework_121cy_setnSolidCoexistAsList, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5melts_15rMELTSframework_120cy_setnSolidCoexistAsList};
static PyObject *__pyx_pw_5melts_15rMELTSframework_121cy_setnSolidCoexistAsList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_values = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setnSolidCoexistAsList (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_values,0};
    PyObject* values[1] = {0};
    values[0] = __pyx_k__3;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cy_setnSolidCoexistAsList") < 0)) __PYX_ERR(0, 335, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_values = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cy_setnSolidCoexistAsList", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 335, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setnSolidCoexistAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_values), (&PyList_Type), 1, "values", 1))) __PYX_ERR(0, 335, __pyx_L1_error)
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_120cy_setnSolidCoexistAsList(__pyx_self, __pyx_v_values);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_120cy_setnSolidCoexistAsList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values) {
  PyObject *__pyx_v_npc = NULL;
  int *__pyx_v_ns;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_setnSolidCoexistAsList", 0);

  /* "melts/rMELTSframework.pyx":336
 * 
 * def cy_setnSolidCoexistAsList(list values = []):
 * 	npc = num_phase_components()             # <<<<<<<<<<<<<<
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))
 * 	for i in range(0,npc):
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(num_phase_components()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_npc = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":337
 * def cy_setnSolidCoexistAsList(list values = []):
 * 	npc = num_phase_components()
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))             # <<<<<<<<<<<<<<
 * 	for i in range(0,npc):
 * 		ns[i] = <int> values[<int>i]
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t((sizeof(int))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_npc, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_ns = ((int *)malloc(__pyx_t_3));

  /* "melts/rMELTSframework.pyx":338
 * 	npc = num_phase_components()
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))
 * 	for i in range(0,npc):             # <<<<<<<<<<<<<<
 * 		ns[i] = <int> values[<int>i]
 * 	setnSolidCoexist(<int *> ns)
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_0);
  __Pyx_INCREF(__pyx_v_npc);
  __Pyx_GIVEREF(__pyx_v_npc);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_npc);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 338, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 338, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 338, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 338, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "melts/rMELTSframework.pyx":339
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))
 * 	for i in range(0,npc):
 * 		ns[i] = <int> values[<int>i]             # <<<<<<<<<<<<<<
 * 	setnSolidCoexist(<int *> ns)
 * 	# free (ns)  memory management in C function
 */
    if (unlikely(__pyx_v_values == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 339, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_v_i); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 339, __pyx_L1_error)
    __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_values, ((int)__pyx_t_6), int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 339, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 339, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_7 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 339, __pyx_L1_error)
    (__pyx_v_ns[__pyx_t_7]) = ((int)__pyx_t_6);

    /* "melts/rMELTSframework.pyx":338
 * 	npc = num_phase_components()
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))
 * 	for i in range(0,npc):             # <<<<<<<<<<<<<<
 * 		ns[i] = <int> values[<int>i]
 * 	setnSolidCoexist(<int *> ns)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "melts/rMELTSframework.pyx":340
 * 	for i in range(0,npc):
 * 		ns[i] = <int> values[<int>i]
 * 	setnSolidCoexist(<int *> ns)             # <<<<<<<<<<<<<<
 * 	# free (ns)  memory management in C function
 * 
 */
  setnSolidCoexist(((int *)__pyx_v_ns));

  /* "melts/rMELTSframework.pyx":335
 * 	return result
 * 
 * def cy_setnSolidCoexistAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setnSolidCoexistAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_npc);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":343
 * 	# free (ns)  memory management in C function
 * 
 * def cy_solidCompAsListOfLists():             # <<<<<<<<<<<<<<
 * 	cdef double **sc = solidCompAsPtr()
 * 	cdef int *ns = nSolidCoexist()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_123cy_solidCompAsListOfLists(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_122cy_solidCompAsListOfLists[] = "cy_solidCompAsListOfLists()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_123cy_solidCompAsListOfLists = {"cy_solidCompAsListOfLists", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_123cy_solidCompAsListOfLists, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_122cy_solidCompAsListOfLists};
static PyObject *__pyx_pw_5melts_15rMELTSframework_123cy_solidCompAsListOfLists(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_solidCompAsListOfLists (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_122cy_solidCompAsListOfLists(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_122cy_solidCompAsListOfLists(CYTHON_UNUSED PyObject *__pyx_self) {
  double **__pyx_v_sc;
  int *__pyx_v_ns;
  int __pyx_v_npc;
  PyObject *__pyx_v_result = NULL;
  long __pyx_v_i;
  long __pyx_8genexpr2__pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  long __pyx_t_4;
  long __pyx_t_5;
  int __pyx_t_6;
  long __pyx_t_7;
  long __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_solidCompAsListOfLists", 0);

  /* "melts/rMELTSframework.pyx":344
 * 
 * def cy_solidCompAsListOfLists():
 * 	cdef double **sc = solidCompAsPtr()             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()
 */
  __pyx_v_sc = solidCompAsPtr();

  /* "melts/rMELTSframework.pyx":345
 * def cy_solidCompAsListOfLists():
 * 	cdef double **sc = solidCompAsPtr()
 * 	cdef int *ns = nSolidCoexist()             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	result = []
 */
  __pyx_v_ns = nSolidCoexist();

  /* "melts/rMELTSframework.pyx":346
 * 	cdef double **sc = solidCompAsPtr()
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()             # <<<<<<<<<<<<<<
 * 	result = []
 * 	for i in range(0, npc):
 */
  __pyx_v_npc = num_phase_components();

  /* "melts/rMELTSframework.pyx":347
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()
 * 	result = []             # <<<<<<<<<<<<<<
 * 	for i in range(0, npc):
 * 		result.append([sc[i][j] for j in range(0, max(ns[i],1))])
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":348
 * 	npc = num_phase_components()
 * 	result = []
 * 	for i in range(0, npc):             # <<<<<<<<<<<<<<
 * 		result.append([sc[i][j] for j in range(0, max(ns[i],1))])
 * 	return result
 */
  __pyx_t_2 = __pyx_v_npc;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "melts/rMELTSframework.pyx":349
 * 	result = []
 * 	for i in range(0, npc):
 * 		result.append([sc[i][j] for j in range(0, max(ns[i],1))])             # <<<<<<<<<<<<<<
 * 	return result
 * 
 */
    { /* enter inner scope */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 349, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = 1;
      __pyx_t_6 = (__pyx_v_ns[__pyx_v_i]);
      if (((__pyx_t_5 > __pyx_t_6) != 0)) {
        __pyx_t_7 = __pyx_t_5;
      } else {
        __pyx_t_7 = __pyx_t_6;
      }
      __pyx_t_5 = __pyx_t_7;
      __pyx_t_7 = __pyx_t_5;
      for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
        __pyx_8genexpr2__pyx_v_j = __pyx_t_8;
        __pyx_t_9 = PyFloat_FromDouble(((__pyx_v_sc[__pyx_v_i])[__pyx_8genexpr2__pyx_v_j])); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_9))) __PYX_ERR(0, 349, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
    } /* exit inner scope */
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 349, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "melts/rMELTSframework.pyx":350
 * 	for i in range(0, npc):
 * 		result.append([sc[i][j] for j in range(0, max(ns[i],1))])
 * 	return result             # <<<<<<<<<<<<<<
 * 
 * def cy_setSolidCompAsListOfLists(list values = [[]]):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":343
 * 	# free (ns)  memory management in C function
 * 
 * def cy_solidCompAsListOfLists():             # <<<<<<<<<<<<<<
 * 	cdef double **sc = solidCompAsPtr()
 * 	cdef int *ns = nSolidCoexist()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_solidCompAsListOfLists", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":352
 * 	return result
 * 
 * def cy_setSolidCompAsListOfLists(list values = [[]]):             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_125cy_setSolidCompAsListOfLists(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_124cy_setSolidCompAsListOfLists[] = "cy_setSolidCompAsListOfLists(list values=[[]])";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_125cy_setSolidCompAsListOfLists = {"cy_setSolidCompAsListOfLists", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5melts_15rMELTSframework_125cy_setSolidCompAsListOfLists, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5melts_15rMELTSframework_124cy_setSolidCompAsListOfLists};
static PyObject *__pyx_pw_5melts_15rMELTSframework_125cy_setSolidCompAsListOfLists(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_values = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setSolidCompAsListOfLists (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_values,0};
    PyObject* values[1] = {0};
    values[0] = __pyx_k__4;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cy_setSolidCompAsListOfLists") < 0)) __PYX_ERR(0, 352, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_values = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cy_setSolidCompAsListOfLists", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 352, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setSolidCompAsListOfLists", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_values), (&PyList_Type), 1, "values", 1))) __PYX_ERR(0, 352, __pyx_L1_error)
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_124cy_setSolidCompAsListOfLists(__pyx_self, __pyx_v_values);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_124cy_setSolidCompAsListOfLists(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values) {
  int *__pyx_v_ns;
  PyObject *__pyx_v_npc = NULL;
  double **__pyx_v_sc;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  long __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  long __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  double __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_setSolidCompAsListOfLists", 0);

  /* "melts/rMELTSframework.pyx":353
 * 
 * def cy_setSolidCompAsListOfLists(list values = [[]]):
 * 	cdef int *ns = nSolidCoexist()             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef double **sc = <double **>malloc(npc*sizeof(double *))
 */
  __pyx_v_ns = nSolidCoexist();

  /* "melts/rMELTSframework.pyx":354
 * def cy_setSolidCompAsListOfLists(list values = [[]]):
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()             # <<<<<<<<<<<<<<
 * 	cdef double **sc = <double **>malloc(npc*sizeof(double *))
 * 	for i in range(0, npc):
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(num_phase_components()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_npc = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":355
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()
 * 	cdef double **sc = <double **>malloc(npc*sizeof(double *))             # <<<<<<<<<<<<<<
 * 	for i in range(0, npc):
 * 		sc[i] = <double *>malloc(max(ns[i],1)*sizeof(double))
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t((sizeof(double *))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_npc, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_sc = ((double **)malloc(__pyx_t_3));

  /* "melts/rMELTSframework.pyx":356
 * 	npc = num_phase_components()
 * 	cdef double **sc = <double **>malloc(npc*sizeof(double *))
 * 	for i in range(0, npc):             # <<<<<<<<<<<<<<
 * 		sc[i] = <double *>malloc(max(ns[i],1)*sizeof(double))
 * 		for j in range(0, max(ns[i],1)):
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_0);
  __Pyx_INCREF(__pyx_v_npc);
  __Pyx_GIVEREF(__pyx_v_npc);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_npc);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 356, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 356, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 356, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 356, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 356, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 356, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "melts/rMELTSframework.pyx":357
 * 	cdef double **sc = <double **>malloc(npc*sizeof(double *))
 * 	for i in range(0, npc):
 * 		sc[i] = <double *>malloc(max(ns[i],1)*sizeof(double))             # <<<<<<<<<<<<<<
 * 		for j in range(0, max(ns[i],1)):
 * 			sc[i][j] = values[<int>i][<int>j]
 */
    __pyx_t_6 = 1;
    __pyx_t_7 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_7 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 357, __pyx_L1_error)
    __pyx_t_8 = (__pyx_v_ns[__pyx_t_7]);
    if (((__pyx_t_6 > __pyx_t_8) != 0)) {
      __pyx_t_9 = __pyx_t_6;
    } else {
      __pyx_t_9 = __pyx_t_8;
    }
    __pyx_t_7 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_7 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 357, __pyx_L1_error)
    (__pyx_v_sc[__pyx_t_7]) = ((double *)malloc((__pyx_t_9 * (sizeof(double)))));

    /* "melts/rMELTSframework.pyx":358
 * 	for i in range(0, npc):
 * 		sc[i] = <double *>malloc(max(ns[i],1)*sizeof(double))
 * 		for j in range(0, max(ns[i],1)):             # <<<<<<<<<<<<<<
 * 			sc[i][j] = values[<int>i][<int>j]
 * 	setSolidCompAsPtr(<double **> sc)
 */
    __pyx_t_9 = 1;
    __pyx_t_7 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_7 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 358, __pyx_L1_error)
    __pyx_t_8 = (__pyx_v_ns[__pyx_t_7]);
    if (((__pyx_t_9 > __pyx_t_8) != 0)) {
      __pyx_t_6 = __pyx_t_9;
    } else {
      __pyx_t_6 = __pyx_t_8;
    }
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_int_0);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_10 = __pyx_t_1; __Pyx_INCREF(__pyx_t_10); __pyx_t_7 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_10 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = Py_TYPE(__pyx_t_10)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 358, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_10))) {
          if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_10)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_10, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 358, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_10, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_10)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_10, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 358, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_10, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_11(__pyx_t_10);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 358, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "melts/rMELTSframework.pyx":359
 * 		sc[i] = <double *>malloc(max(ns[i],1)*sizeof(double))
 * 		for j in range(0, max(ns[i],1)):
 * 			sc[i][j] = values[<int>i][<int>j]             # <<<<<<<<<<<<<<
 * 	setSolidCompAsPtr(<double **> sc)
 * 	#for i in range(0, npc):
 */
      if (unlikely(__pyx_v_values == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 359, __pyx_L1_error)
      }
      __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_v_i); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 359, __pyx_L1_error)
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_values, ((int)__pyx_t_8), int, 1, __Pyx_PyInt_From_int, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 359, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_v_j); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 359, __pyx_L1_error)
      __pyx_t_12 = __Pyx_GetItemInt(__pyx_t_1, ((int)__pyx_t_8), int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 359, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_13 = __pyx_PyFloat_AsDouble(__pyx_t_12); if (unlikely((__pyx_t_13 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 359, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_14 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_14 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 359, __pyx_L1_error)
      __pyx_t_15 = __Pyx_PyIndex_AsSsize_t(__pyx_v_j); if (unlikely((__pyx_t_15 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 359, __pyx_L1_error)
      ((__pyx_v_sc[__pyx_t_14])[__pyx_t_15]) = __pyx_t_13;

      /* "melts/rMELTSframework.pyx":358
 * 	for i in range(0, npc):
 * 		sc[i] = <double *>malloc(max(ns[i],1)*sizeof(double))
 * 		for j in range(0, max(ns[i],1)):             # <<<<<<<<<<<<<<
 * 			sc[i][j] = values[<int>i][<int>j]
 * 	setSolidCompAsPtr(<double **> sc)
 */
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "melts/rMELTSframework.pyx":356
 * 	npc = num_phase_components()
 * 	cdef double **sc = <double **>malloc(npc*sizeof(double *))
 * 	for i in range(0, npc):             # <<<<<<<<<<<<<<
 * 		sc[i] = <double *>malloc(max(ns[i],1)*sizeof(double))
 * 		for j in range(0, max(ns[i],1)):
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "melts/rMELTSframework.pyx":360
 * 		for j in range(0, max(ns[i],1)):
 * 			sc[i][j] = values[<int>i][<int>j]
 * 	setSolidCompAsPtr(<double **> sc)             # <<<<<<<<<<<<<<
 * 	#for i in range(0, npc):
 * 	#	free(sc[i])
 */
  setSolidCompAsPtr(((double **)__pyx_v_sc));

  /* "melts/rMELTSframework.pyx":352
 * 	return result
 * 
 * def cy_setSolidCompAsListOfLists(list values = [[]]):             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setSolidCompAsListOfLists", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_npc);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":365
 * 	#free (sc)  memory management in C function
 * 
 * def cy_solidMass():             # <<<<<<<<<<<<<<
 * 	return <double> solidMass()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_127cy_solidMass(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_126cy_solidMass[] = "cy_solidMass()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_127cy_solidMass = {"cy_solidMass", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_127cy_solidMass, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_126cy_solidMass};
static PyObject *__pyx_pw_5melts_15rMELTSframework_127cy_solidMass(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_solidMass (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_126cy_solidMass(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_126cy_solidMass(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_solidMass", 0);

  /* "melts/rMELTSframework.pyx":366
 * 
 * def cy_solidMass():
 * 	return <double> solidMass()             # <<<<<<<<<<<<<<
 * 
 * def cy_setSolidMass(double value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(((double)solidMass())); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":365
 * 	#free (sc)  memory management in C function
 * 
 * def cy_solidMass():             # <<<<<<<<<<<<<<
 * 	return <double> solidMass()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_solidMass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":368
 * 	return <double> solidMass()
 * 
 * def cy_setSolidMass(double value):             # <<<<<<<<<<<<<<
 * 	setSolidMass(<double> value);
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_129cy_setSolidMass(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_128cy_setSolidMass[] = "cy_setSolidMass(double value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_129cy_setSolidMass = {"cy_setSolidMass", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_129cy_setSolidMass, METH_O, __pyx_doc_5melts_15rMELTSframework_128cy_setSolidMass};
static PyObject *__pyx_pw_5melts_15rMELTSframework_129cy_setSolidMass(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setSolidMass (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 368, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setSolidMass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_128cy_setSolidMass(__pyx_self, ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_128cy_setSolidMass(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setSolidMass", 0);

  /* "melts/rMELTSframework.pyx":369
 * 
 * def cy_setSolidMass(double value):
 * 	setSolidMass(<double> value);             # <<<<<<<<<<<<<<
 * 
 * def cy_incSolidsAsList():
 */
  setSolidMass(((double)__pyx_v_value));

  /* "melts/rMELTSframework.pyx":368
 * 	return <double> solidMass()
 * 
 * def cy_setSolidMass(double value):             # <<<<<<<<<<<<<<
 * 	setSolidMass(<double> value);
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":371
 * 	setSolidMass(<double> value);
 * 
 * def cy_incSolidsAsList():             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef int *inc = incSolids();
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_131cy_incSolidsAsList(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_130cy_incSolidsAsList[] = "cy_incSolidsAsList()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_131cy_incSolidsAsList = {"cy_incSolidsAsList", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_131cy_incSolidsAsList, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_130cy_incSolidsAsList};
static PyObject *__pyx_pw_5melts_15rMELTSframework_131cy_incSolidsAsList(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_incSolidsAsList (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_130cy_incSolidsAsList(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_130cy_incSolidsAsList(CYTHON_UNUSED PyObject *__pyx_self) {
  int __pyx_v_npc;
  int *__pyx_v_inc;
  PyObject *__pyx_v_result = NULL;
  long __pyx_8genexpr3__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  long __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_incSolidsAsList", 0);

  /* "melts/rMELTSframework.pyx":372
 * 
 * def cy_incSolidsAsList():
 * 	npc = num_phase_components()             # <<<<<<<<<<<<<<
 * 	cdef int *inc = incSolids();
 * 	result = [True if inc[i] == 1 else False for i in range(0,npc)]
 */
  __pyx_v_npc = num_phase_components();

  /* "melts/rMELTSframework.pyx":373
 * def cy_incSolidsAsList():
 * 	npc = num_phase_components()
 * 	cdef int *inc = incSolids();             # <<<<<<<<<<<<<<
 * 	result = [True if inc[i] == 1 else False for i in range(0,npc)]
 * 	return result
 */
  __pyx_v_inc = incSolids();

  /* "melts/rMELTSframework.pyx":374
 * 	npc = num_phase_components()
 * 	cdef int *inc = incSolids();
 * 	result = [True if inc[i] == 1 else False for i in range(0,npc)]             # <<<<<<<<<<<<<<
 * 	return result
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 374, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_npc;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_8genexpr3__pyx_v_i = __pyx_t_4;
      if ((((__pyx_v_inc[__pyx_8genexpr3__pyx_v_i]) == 1) != 0)) {
        __Pyx_INCREF(Py_True);
        __pyx_t_5 = Py_True;
      } else {
        __Pyx_INCREF(Py_False);
        __pyx_t_5 = Py_False;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 374, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  } /* exit inner scope */
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":375
 * 	cdef int *inc = incSolids();
 * 	result = [True if inc[i] == 1 else False for i in range(0,npc)]
 * 	return result             # <<<<<<<<<<<<<<
 * 
 * def cy_setIncSolidsAsList(list values = []): # bint[:]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":371
 * 	setSolidMass(<double> value);
 * 
 * def cy_incSolidsAsList():             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef int *inc = incSolids();
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_incSolidsAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":377
 * 	return result
 * 
 * def cy_setIncSolidsAsList(list values = []): # bint[:]             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef int *inc = <int *>malloc(npc*sizeof(int))
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_133cy_setIncSolidsAsList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_132cy_setIncSolidsAsList[] = "cy_setIncSolidsAsList(list values=[])";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_133cy_setIncSolidsAsList = {"cy_setIncSolidsAsList", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5melts_15rMELTSframework_133cy_setIncSolidsAsList, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5melts_15rMELTSframework_132cy_setIncSolidsAsList};
static PyObject *__pyx_pw_5melts_15rMELTSframework_133cy_setIncSolidsAsList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_values = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setIncSolidsAsList (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_values,0};
    PyObject* values[1] = {0};
    values[0] = __pyx_k__5;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cy_setIncSolidsAsList") < 0)) __PYX_ERR(0, 377, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_values = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cy_setIncSolidsAsList", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 377, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setIncSolidsAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_values), (&PyList_Type), 1, "values", 1))) __PYX_ERR(0, 377, __pyx_L1_error)
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_132cy_setIncSolidsAsList(__pyx_self, __pyx_v_values);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_132cy_setIncSolidsAsList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values) {
  PyObject *__pyx_v_npc = NULL;
  int *__pyx_v_inc;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  long __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_setIncSolidsAsList", 0);

  /* "melts/rMELTSframework.pyx":378
 * 
 * def cy_setIncSolidsAsList(list values = []): # bint[:]
 * 	npc = num_phase_components()             # <<<<<<<<<<<<<<
 * 	cdef int *inc = <int *>malloc(npc*sizeof(int))
 * 	for i in range(0,npc):
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(num_phase_components()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_npc = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":379
 * def cy_setIncSolidsAsList(list values = []): # bint[:]
 * 	npc = num_phase_components()
 * 	cdef int *inc = <int *>malloc(npc*sizeof(int))             # <<<<<<<<<<<<<<
 * 	for i in range(0,npc):
 * 		inc[i] = <int> (1 if values[i] else 0)
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t((sizeof(int))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_npc, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_3 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_inc = ((int *)malloc(__pyx_t_3));

  /* "melts/rMELTSframework.pyx":380
 * 	npc = num_phase_components()
 * 	cdef int *inc = <int *>malloc(npc*sizeof(int))
 * 	for i in range(0,npc):             # <<<<<<<<<<<<<<
 * 		inc[i] = <int> (1 if values[i] else 0)
 * 	setIncSolids(<int *> inc)
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_int_0);
  __Pyx_INCREF(__pyx_v_npc);
  __Pyx_GIVEREF(__pyx_v_npc);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_npc);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
  } else {
    __pyx_t_4 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 380, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_5)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_4 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 380, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_4 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely(0 < 0)) __PYX_ERR(0, 380, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_5(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 380, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "melts/rMELTSframework.pyx":381
 * 	cdef int *inc = <int *>malloc(npc*sizeof(int))
 * 	for i in range(0,npc):
 * 		inc[i] = <int> (1 if values[i] else 0)             # <<<<<<<<<<<<<<
 * 	setIncSolids(<int *> inc)
 * 	# free (inc)  memory management in C function
 */
    if (unlikely(__pyx_v_values == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 381, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_values, __pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 381, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_7) {
      __pyx_t_6 = 1;
    } else {
      __pyx_t_6 = 0;
    }
    __pyx_t_8 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_8 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 381, __pyx_L1_error)
    (__pyx_v_inc[__pyx_t_8]) = ((int)__pyx_t_6);

    /* "melts/rMELTSframework.pyx":380
 * 	npc = num_phase_components()
 * 	cdef int *inc = <int *>malloc(npc*sizeof(int))
 * 	for i in range(0,npc):             # <<<<<<<<<<<<<<
 * 		inc[i] = <int> (1 if values[i] else 0)
 * 	setIncSolids(<int *> inc)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "melts/rMELTSframework.pyx":382
 * 	for i in range(0,npc):
 * 		inc[i] = <int> (1 if values[i] else 0)
 * 	setIncSolids(<int *> inc)             # <<<<<<<<<<<<<<
 * 	# free (inc)  memory management in C function
 * 
 */
  setIncSolids(((int *)__pyx_v_inc));

  /* "melts/rMELTSframework.pyx":377
 * 	return result
 * 
 * def cy_setIncSolidsAsList(list values = []): # bint[:]             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef int *inc = <int *>malloc(npc*sizeof(int))
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setIncSolidsAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_npc);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":385
 * 	# free (inc)  memory management in C function
 * 
 * def cy_fo2():             # <<<<<<<<<<<<<<
 * 	return fo2();
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_135cy_fo2(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_134cy_fo2[] = "cy_fo2()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_135cy_fo2 = {"cy_fo2", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_135cy_fo2, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_134cy_fo2};
static PyObject *__pyx_pw_5melts_15rMELTSframework_135cy_fo2(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_fo2 (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_134cy_fo2(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_134cy_fo2(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_fo2", 0);

  /* "melts/rMELTSframework.pyx":386
 * 
 * def cy_fo2():
 * 	return fo2();             # <<<<<<<<<<<<<<
 * 
 * def cy_setfo2(double value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(fo2()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":385
 * 	# free (inc)  memory management in C function
 * 
 * def cy_fo2():             # <<<<<<<<<<<<<<
 * 	return fo2();
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_fo2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":388
 * 	return fo2();
 * 
 * def cy_setfo2(double value):             # <<<<<<<<<<<<<<
 * 	setfo2(<double> value)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_137cy_setfo2(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_136cy_setfo2[] = "cy_setfo2(double value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_137cy_setfo2 = {"cy_setfo2", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_137cy_setfo2, METH_O, __pyx_doc_5melts_15rMELTSframework_136cy_setfo2};
static PyObject *__pyx_pw_5melts_15rMELTSframework_137cy_setfo2(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setfo2 (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 388, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setfo2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_136cy_setfo2(__pyx_self, ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_136cy_setfo2(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setfo2", 0);

  /* "melts/rMELTSframework.pyx":389
 * 
 * def cy_setfo2(double value):
 * 	setfo2(<double> value)             # <<<<<<<<<<<<<<
 * 
 * def cy_fo2Path():
 */
  setfo2(((double)__pyx_v_value));

  /* "melts/rMELTSframework.pyx":388
 * 	return fo2();
 * 
 * def cy_setfo2(double value):             # <<<<<<<<<<<<<<
 * 	setfo2(<double> value)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":391
 * 	setfo2(<double> value)
 * 
 * def cy_fo2Path():             # <<<<<<<<<<<<<<
 * 	return fo2Path();
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_139cy_fo2Path(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_138cy_fo2Path[] = "cy_fo2Path()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_139cy_fo2Path = {"cy_fo2Path", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_139cy_fo2Path, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_138cy_fo2Path};
static PyObject *__pyx_pw_5melts_15rMELTSframework_139cy_fo2Path(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_fo2Path (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_138cy_fo2Path(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_138cy_fo2Path(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_fo2Path", 0);

  /* "melts/rMELTSframework.pyx":392
 * 
 * def cy_fo2Path():
 * 	return fo2Path();             # <<<<<<<<<<<<<<
 * 
 * def cy_setfo2Path(int value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(fo2Path()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":391
 * 	setfo2(<double> value)
 * 
 * def cy_fo2Path():             # <<<<<<<<<<<<<<
 * 	return fo2Path();
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_fo2Path", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":394
 * 	return fo2Path();
 * 
 * def cy_setfo2Path(int value):             # <<<<<<<<<<<<<<
 * 	setfo2Path(<int> value)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_141cy_setfo2Path(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_140cy_setfo2Path[] = "cy_setfo2Path(int value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_141cy_setfo2Path = {"cy_setfo2Path", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_141cy_setfo2Path, METH_O, __pyx_doc_5melts_15rMELTSframework_140cy_setfo2Path};
static PyObject *__pyx_pw_5melts_15rMELTSframework_141cy_setfo2Path(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  int __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setfo2Path (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyInt_As_int(__pyx_arg_value); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 394, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setfo2Path", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_140cy_setfo2Path(__pyx_self, ((int)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_140cy_setfo2Path(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setfo2Path", 0);

  /* "melts/rMELTSframework.pyx":395
 * 
 * def cy_setfo2Path(int value):
 * 	setfo2Path(<int> value)             # <<<<<<<<<<<<<<
 * 
 * def cy_fo2Delta():
 */
  setfo2Path(((int)__pyx_v_value));

  /* "melts/rMELTSframework.pyx":394
 * 	return fo2Path();
 * 
 * def cy_setfo2Path(int value):             # <<<<<<<<<<<<<<
 * 	setfo2Path(<int> value)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":397
 * 	setfo2Path(<int> value)
 * 
 * def cy_fo2Delta():             # <<<<<<<<<<<<<<
 * 	return fo2Delta()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_143cy_fo2Delta(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_142cy_fo2Delta[] = "cy_fo2Delta()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_143cy_fo2Delta = {"cy_fo2Delta", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_143cy_fo2Delta, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_142cy_fo2Delta};
static PyObject *__pyx_pw_5melts_15rMELTSframework_143cy_fo2Delta(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_fo2Delta (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_142cy_fo2Delta(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_142cy_fo2Delta(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_fo2Delta", 0);

  /* "melts/rMELTSframework.pyx":398
 * 
 * def cy_fo2Delta():
 * 	return fo2Delta()             # <<<<<<<<<<<<<<
 * 
 * def cy_setfo2Delta(double value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(fo2Delta()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":397
 * 	setfo2Path(<int> value)
 * 
 * def cy_fo2Delta():             # <<<<<<<<<<<<<<
 * 	return fo2Delta()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_fo2Delta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":400
 * 	return fo2Delta()
 * 
 * def cy_setfo2Delta(double value):             # <<<<<<<<<<<<<<
 * 	setfo2Delta(<double> value)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_145cy_setfo2Delta(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_144cy_setfo2Delta[] = "cy_setfo2Delta(double value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_145cy_setfo2Delta = {"cy_setfo2Delta", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_145cy_setfo2Delta, METH_O, __pyx_doc_5melts_15rMELTSframework_144cy_setfo2Delta};
static PyObject *__pyx_pw_5melts_15rMELTSframework_145cy_setfo2Delta(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setfo2Delta (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 400, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setfo2Delta", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_144cy_setfo2Delta(__pyx_self, ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_144cy_setfo2Delta(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setfo2Delta", 0);

  /* "melts/rMELTSframework.pyx":401
 * 
 * def cy_setfo2Delta(double value):
 * 	setfo2Delta(<double> value)             # <<<<<<<<<<<<<<
 * 
 * def cy_oxygenContent():
 */
  setfo2Delta(((double)__pyx_v_value));

  /* "melts/rMELTSframework.pyx":400
 * 	return fo2Delta()
 * 
 * def cy_setfo2Delta(double value):             # <<<<<<<<<<<<<<
 * 	setfo2Delta(<double> value)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":403
 * 	setfo2Delta(<double> value)
 * 
 * def cy_oxygenContent():             # <<<<<<<<<<<<<<
 * 	return oxygenContent()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_147cy_oxygenContent(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_146cy_oxygenContent[] = "cy_oxygenContent()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_147cy_oxygenContent = {"cy_oxygenContent", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_147cy_oxygenContent, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_146cy_oxygenContent};
static PyObject *__pyx_pw_5melts_15rMELTSframework_147cy_oxygenContent(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_oxygenContent (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_146cy_oxygenContent(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_146cy_oxygenContent(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_oxygenContent", 0);

  /* "melts/rMELTSframework.pyx":404
 * 
 * def cy_oxygenContent():
 * 	return oxygenContent()             # <<<<<<<<<<<<<<
 * 
 * def cy_setOxygenContent(double value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(oxygenContent()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":403
 * 	setfo2Delta(<double> value)
 * 
 * def cy_oxygenContent():             # <<<<<<<<<<<<<<
 * 	return oxygenContent()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_oxygenContent", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":406
 * 	return oxygenContent()
 * 
 * def cy_setOxygenContent(double value):             # <<<<<<<<<<<<<<
 * 	setOxygenContent(<double> value)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_149cy_setOxygenContent(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_148cy_setOxygenContent[] = "cy_setOxygenContent(double value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_149cy_setOxygenContent = {"cy_setOxygenContent", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_149cy_setOxygenContent, METH_O, __pyx_doc_5melts_15rMELTSframework_148cy_setOxygenContent};
static PyObject *__pyx_pw_5melts_15rMELTSframework_149cy_setOxygenContent(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setOxygenContent (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 406, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setOxygenContent", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_148cy_setOxygenContent(__pyx_self, ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_148cy_setOxygenContent(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setOxygenContent", 0);

  /* "melts/rMELTSframework.pyx":407
 * 
 * def cy_setOxygenContent(double value):
 * 	setOxygenContent(<double> value)             # <<<<<<<<<<<<<<
 * 
 * def cy_fractionateSol():
 */
  setOxygenContent(((double)__pyx_v_value));

  /* "melts/rMELTSframework.pyx":406
 * 	return oxygenContent()
 * 
 * def cy_setOxygenContent(double value):             # <<<<<<<<<<<<<<
 * 	setOxygenContent(<double> value)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":409
 * 	setOxygenContent(<double> value)
 * 
 * def cy_fractionateSol():             # <<<<<<<<<<<<<<
 * 	return True if fractionateSol() == 1 else False
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_151cy_fractionateSol(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_150cy_fractionateSol[] = "cy_fractionateSol()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_151cy_fractionateSol = {"cy_fractionateSol", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_151cy_fractionateSol, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_150cy_fractionateSol};
static PyObject *__pyx_pw_5melts_15rMELTSframework_151cy_fractionateSol(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_fractionateSol (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_150cy_fractionateSol(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_150cy_fractionateSol(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cy_fractionateSol", 0);

  /* "melts/rMELTSframework.pyx":410
 * 
 * def cy_fractionateSol():
 * 	return True if fractionateSol() == 1 else False             # <<<<<<<<<<<<<<
 * 
 * def cy_setFractionateSol(bint value):
 */
  __Pyx_XDECREF(__pyx_r);
  if (((fractionateSol() == 1) != 0)) {
    __Pyx_INCREF(Py_True);
    __pyx_t_1 = Py_True;
  } else {
    __Pyx_INCREF(Py_False);
    __pyx_t_1 = Py_False;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":409
 * 	setOxygenContent(<double> value)
 * 
 * def cy_fractionateSol():             # <<<<<<<<<<<<<<
 * 	return True if fractionateSol() == 1 else False
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":412
 * 	return True if fractionateSol() == 1 else False
 * 
 * def cy_setFractionateSol(bint value):             # <<<<<<<<<<<<<<
 * 	setFractionateSol(<int>(1 if value else 0));
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_153cy_setFractionateSol(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_152cy_setFractionateSol[] = "cy_setFractionateSol(bool value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_153cy_setFractionateSol = {"cy_setFractionateSol", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_153cy_setFractionateSol, METH_O, __pyx_doc_5melts_15rMELTSframework_152cy_setFractionateSol};
static PyObject *__pyx_pw_5melts_15rMELTSframework_153cy_setFractionateSol(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  int __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFractionateSol (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyObject_IsTrue(__pyx_arg_value); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setFractionateSol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_152cy_setFractionateSol(__pyx_self, ((int)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_152cy_setFractionateSol(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  __Pyx_RefNannySetupContext("cy_setFractionateSol", 0);

  /* "melts/rMELTSframework.pyx":413
 * 
 * def cy_setFractionateSol(bint value):
 * 	setFractionateSol(<int>(1 if value else 0));             # <<<<<<<<<<<<<<
 * 
 * def cy_fractionateLiq():
 */
  if ((__pyx_v_value != 0)) {
    __pyx_t_1 = 1;
  } else {
    __pyx_t_1 = 0;
  }
  setFractionateSol(((int)__pyx_t_1));

  /* "melts/rMELTSframework.pyx":412
 * 	return True if fractionateSol() == 1 else False
 * 
 * def cy_setFractionateSol(bint value):             # <<<<<<<<<<<<<<
 * 	setFractionateSol(<int>(1 if value else 0));
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":415
 * 	setFractionateSol(<int>(1 if value else 0));
 * 
 * def cy_fractionateLiq():             # <<<<<<<<<<<<<<
 * 	return True if fractionateLiq() == 1 else False
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_155cy_fractionateLiq(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_154cy_fractionateLiq[] = "cy_fractionateLiq()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_155cy_fractionateLiq = {"cy_fractionateLiq", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_155cy_fractionateLiq, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_154cy_fractionateLiq};
static PyObject *__pyx_pw_5melts_15rMELTSframework_155cy_fractionateLiq(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_fractionateLiq (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_154cy_fractionateLiq(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_154cy_fractionateLiq(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cy_fractionateLiq", 0);

  /* "melts/rMELTSframework.pyx":416
 * 
 * def cy_fractionateLiq():
 * 	return True if fractionateLiq() == 1 else False             # <<<<<<<<<<<<<<
 * 
 * def cy_setFractionateLiq(bint value):
 */
  __Pyx_XDECREF(__pyx_r);
  if (((fractionateLiq() == 1) != 0)) {
    __Pyx_INCREF(Py_True);
    __pyx_t_1 = Py_True;
  } else {
    __Pyx_INCREF(Py_False);
    __pyx_t_1 = Py_False;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":415
 * 	setFractionateSol(<int>(1 if value else 0));
 * 
 * def cy_fractionateLiq():             # <<<<<<<<<<<<<<
 * 	return True if fractionateLiq() == 1 else False
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":418
 * 	return True if fractionateLiq() == 1 else False
 * 
 * def cy_setFractionateLiq(bint value):             # <<<<<<<<<<<<<<
 * 	setFractionateLiq(<int>(1 if value else 0))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_157cy_setFractionateLiq(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_156cy_setFractionateLiq[] = "cy_setFractionateLiq(bool value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_157cy_setFractionateLiq = {"cy_setFractionateLiq", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_157cy_setFractionateLiq, METH_O, __pyx_doc_5melts_15rMELTSframework_156cy_setFractionateLiq};
static PyObject *__pyx_pw_5melts_15rMELTSframework_157cy_setFractionateLiq(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  int __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFractionateLiq (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyObject_IsTrue(__pyx_arg_value); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 418, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setFractionateLiq", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_156cy_setFractionateLiq(__pyx_self, ((int)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_156cy_setFractionateLiq(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  __Pyx_RefNannySetupContext("cy_setFractionateLiq", 0);

  /* "melts/rMELTSframework.pyx":419
 * 
 * def cy_setFractionateLiq(bint value):
 * 	setFractionateLiq(<int>(1 if value else 0))             # <<<<<<<<<<<<<<
 * 
 * def cy_fractionateFlu():
 */
  if ((__pyx_v_value != 0)) {
    __pyx_t_1 = 1;
  } else {
    __pyx_t_1 = 0;
  }
  setFractionateLiq(((int)__pyx_t_1));

  /* "melts/rMELTSframework.pyx":418
 * 	return True if fractionateLiq() == 1 else False
 * 
 * def cy_setFractionateLiq(bint value):             # <<<<<<<<<<<<<<
 * 	setFractionateLiq(<int>(1 if value else 0))
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":421
 * 	setFractionateLiq(<int>(1 if value else 0))
 * 
 * def cy_fractionateFlu():             # <<<<<<<<<<<<<<
 * 	return True if fractionateFlu() == 1 else False
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_159cy_fractionateFlu(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_158cy_fractionateFlu[] = "cy_fractionateFlu()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_159cy_fractionateFlu = {"cy_fractionateFlu", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_159cy_fractionateFlu, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_158cy_fractionateFlu};
static PyObject *__pyx_pw_5melts_15rMELTSframework_159cy_fractionateFlu(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_fractionateFlu (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_158cy_fractionateFlu(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_158cy_fractionateFlu(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cy_fractionateFlu", 0);

  /* "melts/rMELTSframework.pyx":422
 * 
 * def cy_fractionateFlu():
 * 	return True if fractionateFlu() == 1 else False             # <<<<<<<<<<<<<<
 * 
 * def cy_setFractionateFlu(bint value):
 */
  __Pyx_XDECREF(__pyx_r);
  if (((fractionateFlu() == 1) != 0)) {
    __Pyx_INCREF(Py_True);
    __pyx_t_1 = Py_True;
  } else {
    __Pyx_INCREF(Py_False);
    __pyx_t_1 = Py_False;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":421
 * 	setFractionateLiq(<int>(1 if value else 0))
 * 
 * def cy_fractionateFlu():             # <<<<<<<<<<<<<<
 * 	return True if fractionateFlu() == 1 else False
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":424
 * 	return True if fractionateFlu() == 1 else False
 * 
 * def cy_setFractionateFlu(bint value):             # <<<<<<<<<<<<<<
 * 	setFractionateFlu(<int>(1 if value else 0));
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_161cy_setFractionateFlu(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_160cy_setFractionateFlu[] = "cy_setFractionateFlu(bool value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_161cy_setFractionateFlu = {"cy_setFractionateFlu", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_161cy_setFractionateFlu, METH_O, __pyx_doc_5melts_15rMELTSframework_160cy_setFractionateFlu};
static PyObject *__pyx_pw_5melts_15rMELTSframework_161cy_setFractionateFlu(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  int __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFractionateFlu (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __Pyx_PyObject_IsTrue(__pyx_arg_value); if (unlikely((__pyx_v_value == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 424, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setFractionateFlu", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_160cy_setFractionateFlu(__pyx_self, ((int)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_160cy_setFractionateFlu(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  __Pyx_RefNannySetupContext("cy_setFractionateFlu", 0);

  /* "melts/rMELTSframework.pyx":425
 * 
 * def cy_setFractionateFlu(bint value):
 * 	setFractionateFlu(<int>(1 if value else 0));             # <<<<<<<<<<<<<<
 * 
 * def cy_nFracCoexistAsList():
 */
  if ((__pyx_v_value != 0)) {
    __pyx_t_1 = 1;
  } else {
    __pyx_t_1 = 0;
  }
  setFractionateFlu(((int)__pyx_t_1));

  /* "melts/rMELTSframework.pyx":424
 * 	return True if fractionateFlu() == 1 else False
 * 
 * def cy_setFractionateFlu(bint value):             # <<<<<<<<<<<<<<
 * 	setFractionateFlu(<int>(1 if value else 0));
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":427
 * 	setFractionateFlu(<int>(1 if value else 0));
 * 
 * def cy_nFracCoexistAsList():             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nFracCoexist()
 * 	if ns == NULL:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_163cy_nFracCoexistAsList(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_162cy_nFracCoexistAsList[] = "cy_nFracCoexistAsList()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_163cy_nFracCoexistAsList = {"cy_nFracCoexistAsList", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_163cy_nFracCoexistAsList, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_162cy_nFracCoexistAsList};
static PyObject *__pyx_pw_5melts_15rMELTSframework_163cy_nFracCoexistAsList(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_nFracCoexistAsList (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_162cy_nFracCoexistAsList(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_162cy_nFracCoexistAsList(CYTHON_UNUSED PyObject *__pyx_self) {
  int *__pyx_v_ns;
  int __pyx_v_npc;
  PyObject *__pyx_v_result = NULL;
  long __pyx_8genexpr4__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  long __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_nFracCoexistAsList", 0);

  /* "melts/rMELTSframework.pyx":428
 * 
 * def cy_nFracCoexistAsList():
 * 	cdef int *ns = nFracCoexist()             # <<<<<<<<<<<<<<
 * 	if ns == NULL:
 * 		return None
 */
  __pyx_v_ns = nFracCoexist();

  /* "melts/rMELTSframework.pyx":429
 * def cy_nFracCoexistAsList():
 * 	cdef int *ns = nFracCoexist()
 * 	if ns == NULL:             # <<<<<<<<<<<<<<
 * 		return None
 * 	npc = num_phase_components()
 */
  __pyx_t_1 = ((__pyx_v_ns == NULL) != 0);
  if (__pyx_t_1) {

    /* "melts/rMELTSframework.pyx":430
 * 	cdef int *ns = nFracCoexist()
 * 	if ns == NULL:
 * 		return None             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	result = [<int> ns[i] for i in range(0,npc)]
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "melts/rMELTSframework.pyx":429
 * def cy_nFracCoexistAsList():
 * 	cdef int *ns = nFracCoexist()
 * 	if ns == NULL:             # <<<<<<<<<<<<<<
 * 		return None
 * 	npc = num_phase_components()
 */
  }

  /* "melts/rMELTSframework.pyx":431
 * 	if ns == NULL:
 * 		return None
 * 	npc = num_phase_components()             # <<<<<<<<<<<<<<
 * 	result = [<int> ns[i] for i in range(0,npc)]
 * 	return result
 */
  __pyx_v_npc = num_phase_components();

  /* "melts/rMELTSframework.pyx":432
 * 		return None
 * 	npc = num_phase_components()
 * 	result = [<int> ns[i] for i in range(0,npc)]             # <<<<<<<<<<<<<<
 * 	return result
 * 
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 432, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_v_npc;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_8genexpr4__pyx_v_i = __pyx_t_5;
      __pyx_t_6 = __Pyx_PyInt_From_int(((int)(__pyx_v_ns[__pyx_8genexpr4__pyx_v_i]))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 432, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 432, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  } /* exit inner scope */
  __pyx_v_result = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "melts/rMELTSframework.pyx":433
 * 	npc = num_phase_components()
 * 	result = [<int> ns[i] for i in range(0,npc)]
 * 	return result             # <<<<<<<<<<<<<<
 * 
 * def cy_setnFracCoexistAsList(list values = []):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":427
 * 	setFractionateFlu(<int>(1 if value else 0));
 * 
 * def cy_nFracCoexistAsList():             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nFracCoexist()
 * 	if ns == NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_nFracCoexistAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":435
 * 	return result
 * 
 * def cy_setnFracCoexistAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	if values is None:
 * 		return
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_165cy_setnFracCoexistAsList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_164cy_setnFracCoexistAsList[] = "cy_setnFracCoexistAsList(list values=[])";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_165cy_setnFracCoexistAsList = {"cy_setnFracCoexistAsList", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5melts_15rMELTSframework_165cy_setnFracCoexistAsList, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5melts_15rMELTSframework_164cy_setnFracCoexistAsList};
static PyObject *__pyx_pw_5melts_15rMELTSframework_165cy_setnFracCoexistAsList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_values = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setnFracCoexistAsList (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_values,0};
    PyObject* values[1] = {0};
    values[0] = __pyx_k__6;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cy_setnFracCoexistAsList") < 0)) __PYX_ERR(0, 435, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_values = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cy_setnFracCoexistAsList", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 435, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setnFracCoexistAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_values), (&PyList_Type), 1, "values", 1))) __PYX_ERR(0, 435, __pyx_L1_error)
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_164cy_setnFracCoexistAsList(__pyx_self, __pyx_v_values);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_164cy_setnFracCoexistAsList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values) {
  PyObject *__pyx_v_npc = NULL;
  int *__pyx_v_ns;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_setnFracCoexistAsList", 0);

  /* "melts/rMELTSframework.pyx":436
 * 
 * def cy_setnFracCoexistAsList(list values = []):
 * 	if values is None:             # <<<<<<<<<<<<<<
 * 		return
 * 	npc = num_phase_components()
 */
  __pyx_t_1 = (__pyx_v_values == ((PyObject*)Py_None));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "melts/rMELTSframework.pyx":437
 * def cy_setnFracCoexistAsList(list values = []):
 * 	if values is None:
 * 		return             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "melts/rMELTSframework.pyx":436
 * 
 * def cy_setnFracCoexistAsList(list values = []):
 * 	if values is None:             # <<<<<<<<<<<<<<
 * 		return
 * 	npc = num_phase_components()
 */
  }

  /* "melts/rMELTSframework.pyx":438
 * 	if values is None:
 * 		return
 * 	npc = num_phase_components()             # <<<<<<<<<<<<<<
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))
 * 	for i in range(0,npc):
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(num_phase_components()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_npc = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":439
 * 		return
 * 	npc = num_phase_components()
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))             # <<<<<<<<<<<<<<
 * 	for i in range(0,npc):
 * 		ns[i] = <int> values[i]
 */
  __pyx_t_3 = __Pyx_PyInt_FromSize_t((sizeof(int))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyNumber_Multiply(__pyx_v_npc, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_ns = ((int *)malloc(__pyx_t_5));

  /* "melts/rMELTSframework.pyx":440
 * 	npc = num_phase_components()
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))
 * 	for i in range(0,npc):             # <<<<<<<<<<<<<<
 * 		ns[i] = <int> values[i]
 * 	setnFracCoexist(<int *> ns)
 */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_int_0);
  __Pyx_INCREF(__pyx_v_npc);
  __Pyx_GIVEREF(__pyx_v_npc);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_npc);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 440, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 440, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 440, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_7(__pyx_t_4);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 440, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "melts/rMELTSframework.pyx":441
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))
 * 	for i in range(0,npc):
 * 		ns[i] = <int> values[i]             # <<<<<<<<<<<<<<
 * 	setnFracCoexist(<int *> ns)
 * 	# free (ns)  memory management in C function
 */
    if (unlikely(__pyx_v_values == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 441, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_values, __pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 441, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 441, __pyx_L1_error)
    (__pyx_v_ns[__pyx_t_9]) = ((int)__pyx_t_8);

    /* "melts/rMELTSframework.pyx":440
 * 	npc = num_phase_components()
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))
 * 	for i in range(0,npc):             # <<<<<<<<<<<<<<
 * 		ns[i] = <int> values[i]
 * 	setnFracCoexist(<int *> ns)
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "melts/rMELTSframework.pyx":442
 * 	for i in range(0,npc):
 * 		ns[i] = <int> values[i]
 * 	setnFracCoexist(<int *> ns)             # <<<<<<<<<<<<<<
 * 	# free (ns)  memory management in C function
 * 
 */
  setnFracCoexist(((int *)__pyx_v_ns));

  /* "melts/rMELTSframework.pyx":435
 * 	return result
 * 
 * def cy_setnFracCoexistAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	if values is None:
 * 		return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setnFracCoexistAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_npc);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":445
 * 	# free (ns)  memory management in C function
 * 
 * def cy_fracSCompAsListOfLists():             # <<<<<<<<<<<<<<
 * 	cdef double **sc = fracSComp()
 * 	cdef int *ns = nSolidCoexist()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_167cy_fracSCompAsListOfLists(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_166cy_fracSCompAsListOfLists[] = "cy_fracSCompAsListOfLists()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_167cy_fracSCompAsListOfLists = {"cy_fracSCompAsListOfLists", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_167cy_fracSCompAsListOfLists, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_166cy_fracSCompAsListOfLists};
static PyObject *__pyx_pw_5melts_15rMELTSframework_167cy_fracSCompAsListOfLists(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_fracSCompAsListOfLists (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_166cy_fracSCompAsListOfLists(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_166cy_fracSCompAsListOfLists(CYTHON_UNUSED PyObject *__pyx_self) {
  double **__pyx_v_sc;
  int *__pyx_v_ns;
  int __pyx_v_npc;
  PyObject *__pyx_v_result = NULL;
  long __pyx_v_i;
  long __pyx_8genexpr5__pyx_v_j;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  long __pyx_t_6;
  long __pyx_t_7;
  int __pyx_t_8;
  long __pyx_t_9;
  long __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_fracSCompAsListOfLists", 0);

  /* "melts/rMELTSframework.pyx":446
 * 
 * def cy_fracSCompAsListOfLists():
 * 	cdef double **sc = fracSComp()             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nSolidCoexist()
 * 	if sc == NULL or ns == NULL:
 */
  __pyx_v_sc = fracSComp();

  /* "melts/rMELTSframework.pyx":447
 * def cy_fracSCompAsListOfLists():
 * 	cdef double **sc = fracSComp()
 * 	cdef int *ns = nSolidCoexist()             # <<<<<<<<<<<<<<
 * 	if sc == NULL or ns == NULL:
 * 		return None
 */
  __pyx_v_ns = nSolidCoexist();

  /* "melts/rMELTSframework.pyx":448
 * 	cdef double **sc = fracSComp()
 * 	cdef int *ns = nSolidCoexist()
 * 	if sc == NULL or ns == NULL:             # <<<<<<<<<<<<<<
 * 		return None
 * 	npc = num_phase_components()
 */
  __pyx_t_2 = ((__pyx_v_sc == NULL) != 0);
  if (!__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_ns == NULL) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "melts/rMELTSframework.pyx":449
 * 	cdef int *ns = nSolidCoexist()
 * 	if sc == NULL or ns == NULL:
 * 		return None             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	result = []
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "melts/rMELTSframework.pyx":448
 * 	cdef double **sc = fracSComp()
 * 	cdef int *ns = nSolidCoexist()
 * 	if sc == NULL or ns == NULL:             # <<<<<<<<<<<<<<
 * 		return None
 * 	npc = num_phase_components()
 */
  }

  /* "melts/rMELTSframework.pyx":450
 * 	if sc == NULL or ns == NULL:
 * 		return None
 * 	npc = num_phase_components()             # <<<<<<<<<<<<<<
 * 	result = []
 * 	for i in range(0, npc):
 */
  __pyx_v_npc = num_phase_components();

  /* "melts/rMELTSframework.pyx":451
 * 		return None
 * 	npc = num_phase_components()
 * 	result = []             # <<<<<<<<<<<<<<
 * 	for i in range(0, npc):
 * 		result.append([sc[i][j] for j in range(0, max(ns[i],1))])
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_result = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":452
 * 	npc = num_phase_components()
 * 	result = []
 * 	for i in range(0, npc):             # <<<<<<<<<<<<<<
 * 		result.append([sc[i][j] for j in range(0, max(ns[i],1))])
 * 	return result
 */
  __pyx_t_4 = __pyx_v_npc;
  __pyx_t_5 = __pyx_t_4;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "melts/rMELTSframework.pyx":453
 * 	result = []
 * 	for i in range(0, npc):
 * 		result.append([sc[i][j] for j in range(0, max(ns[i],1))])             # <<<<<<<<<<<<<<
 * 	return result
 * 
 */
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 453, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = 1;
      __pyx_t_8 = (__pyx_v_ns[__pyx_v_i]);
      if (((__pyx_t_7 > __pyx_t_8) != 0)) {
        __pyx_t_9 = __pyx_t_7;
      } else {
        __pyx_t_9 = __pyx_t_8;
      }
      __pyx_t_7 = __pyx_t_9;
      __pyx_t_9 = __pyx_t_7;
      for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
        __pyx_8genexpr5__pyx_v_j = __pyx_t_10;
        __pyx_t_11 = PyFloat_FromDouble(((__pyx_v_sc[__pyx_v_i])[__pyx_8genexpr5__pyx_v_j])); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 453, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_11))) __PYX_ERR(0, 453, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
    } /* exit inner scope */
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }

  /* "melts/rMELTSframework.pyx":454
 * 	for i in range(0, npc):
 * 		result.append([sc[i][j] for j in range(0, max(ns[i],1))])
 * 	return result             # <<<<<<<<<<<<<<
 * 
 * def cy_setFracSCompAsListOfLists(list values = [[]]):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":445
 * 	# free (ns)  memory management in C function
 * 
 * def cy_fracSCompAsListOfLists():             # <<<<<<<<<<<<<<
 * 	cdef double **sc = fracSComp()
 * 	cdef int *ns = nSolidCoexist()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_fracSCompAsListOfLists", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":456
 * 	return result
 * 
 * def cy_setFracSCompAsListOfLists(list values = [[]]):             # <<<<<<<<<<<<<<
 * 	if values is None:
 * 		return
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_169cy_setFracSCompAsListOfLists(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_168cy_setFracSCompAsListOfLists[] = "cy_setFracSCompAsListOfLists(list values=[[]])";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_169cy_setFracSCompAsListOfLists = {"cy_setFracSCompAsListOfLists", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5melts_15rMELTSframework_169cy_setFracSCompAsListOfLists, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5melts_15rMELTSframework_168cy_setFracSCompAsListOfLists};
static PyObject *__pyx_pw_5melts_15rMELTSframework_169cy_setFracSCompAsListOfLists(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_values = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFracSCompAsListOfLists (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_values,0};
    PyObject* values[1] = {0};
    values[0] = __pyx_k__7;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cy_setFracSCompAsListOfLists") < 0)) __PYX_ERR(0, 456, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_values = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cy_setFracSCompAsListOfLists", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 456, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setFracSCompAsListOfLists", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_values), (&PyList_Type), 1, "values", 1))) __PYX_ERR(0, 456, __pyx_L1_error)
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_168cy_setFracSCompAsListOfLists(__pyx_self, __pyx_v_values);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_168cy_setFracSCompAsListOfLists(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values) {
  int *__pyx_v_ns;
  PyObject *__pyx_v_npc = NULL;
  double **__pyx_v_sc;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_j = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  long __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  long __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *(*__pyx_t_13)(PyObject *);
  PyObject *__pyx_t_14 = NULL;
  double __pyx_t_15;
  Py_ssize_t __pyx_t_16;
  Py_ssize_t __pyx_t_17;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_setFracSCompAsListOfLists", 0);

  /* "melts/rMELTSframework.pyx":457
 * 
 * def cy_setFracSCompAsListOfLists(list values = [[]]):
 * 	if values is None:             # <<<<<<<<<<<<<<
 * 		return
 * 	cdef int *ns = nSolidCoexist()
 */
  __pyx_t_1 = (__pyx_v_values == ((PyObject*)Py_None));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "melts/rMELTSframework.pyx":458
 * def cy_setFracSCompAsListOfLists(list values = [[]]):
 * 	if values is None:
 * 		return             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "melts/rMELTSframework.pyx":457
 * 
 * def cy_setFracSCompAsListOfLists(list values = [[]]):
 * 	if values is None:             # <<<<<<<<<<<<<<
 * 		return
 * 	cdef int *ns = nSolidCoexist()
 */
  }

  /* "melts/rMELTSframework.pyx":459
 * 	if values is None:
 * 		return
 * 	cdef int *ns = nSolidCoexist()             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef double **sc = <double **>malloc(npc*sizeof(double *))
 */
  __pyx_v_ns = nSolidCoexist();

  /* "melts/rMELTSframework.pyx":460
 * 		return
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()             # <<<<<<<<<<<<<<
 * 	cdef double **sc = <double **>malloc(npc*sizeof(double *))
 * 	for i in range(0, npc):
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(num_phase_components()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_npc = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":461
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()
 * 	cdef double **sc = <double **>malloc(npc*sizeof(double *))             # <<<<<<<<<<<<<<
 * 	for i in range(0, npc):
 * 		sc[i] = <double *>malloc(max(ns[i],1)*sizeof(double))
 */
  __pyx_t_3 = __Pyx_PyInt_FromSize_t((sizeof(double *))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyNumber_Multiply(__pyx_v_npc, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_sc = ((double **)malloc(__pyx_t_5));

  /* "melts/rMELTSframework.pyx":462
 * 	npc = num_phase_components()
 * 	cdef double **sc = <double **>malloc(npc*sizeof(double *))
 * 	for i in range(0, npc):             # <<<<<<<<<<<<<<
 * 		sc[i] = <double *>malloc(max(ns[i],1)*sizeof(double))
 * 		for j in range(0, max(ns[i],1)):
 */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_int_0);
  __Pyx_INCREF(__pyx_v_npc);
  __Pyx_GIVEREF(__pyx_v_npc);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_npc);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 462, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 462, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 462, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 462, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 462, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 462, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_7(__pyx_t_4);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 462, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "melts/rMELTSframework.pyx":463
 * 	cdef double **sc = <double **>malloc(npc*sizeof(double *))
 * 	for i in range(0, npc):
 * 		sc[i] = <double *>malloc(max(ns[i],1)*sizeof(double))             # <<<<<<<<<<<<<<
 * 		for j in range(0, max(ns[i],1)):
 * 			sc[i][j] = values[i][j]
 */
    __pyx_t_8 = 1;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 463, __pyx_L1_error)
    __pyx_t_10 = (__pyx_v_ns[__pyx_t_9]);
    if (((__pyx_t_8 > __pyx_t_10) != 0)) {
      __pyx_t_11 = __pyx_t_8;
    } else {
      __pyx_t_11 = __pyx_t_10;
    }
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 463, __pyx_L1_error)
    (__pyx_v_sc[__pyx_t_9]) = ((double *)malloc((__pyx_t_11 * (sizeof(double)))));

    /* "melts/rMELTSframework.pyx":464
 * 	for i in range(0, npc):
 * 		sc[i] = <double *>malloc(max(ns[i],1)*sizeof(double))
 * 		for j in range(0, max(ns[i],1)):             # <<<<<<<<<<<<<<
 * 			sc[i][j] = values[i][j]
 * 	setFracSComp(<double **> sc)
 */
    __pyx_t_11 = 1;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 464, __pyx_L1_error)
    __pyx_t_10 = (__pyx_v_ns[__pyx_t_9]);
    if (((__pyx_t_11 > __pyx_t_10) != 0)) {
      __pyx_t_8 = __pyx_t_11;
    } else {
      __pyx_t_8 = __pyx_t_10;
    }
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_GIVEREF(__pyx_int_0);
    PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_int_0);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_12 = __pyx_t_3; __Pyx_INCREF(__pyx_t_12); __pyx_t_9 = 0;
      __pyx_t_13 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_12 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 464, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_13 = Py_TYPE(__pyx_t_12)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 464, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_13)) {
        if (likely(PyList_CheckExact(__pyx_t_12))) {
          if (__pyx_t_9 >= PyList_GET_SIZE(__pyx_t_12)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_12, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 464, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_12, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 464, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_9 >= PyTuple_GET_SIZE(__pyx_t_12)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_12, __pyx_t_9); __Pyx_INCREF(__pyx_t_3); __pyx_t_9++; if (unlikely(0 < 0)) __PYX_ERR(0, 464, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_12, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 464, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_13(__pyx_t_12);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 464, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_j, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "melts/rMELTSframework.pyx":465
 * 		sc[i] = <double *>malloc(max(ns[i],1)*sizeof(double))
 * 		for j in range(0, max(ns[i],1)):
 * 			sc[i][j] = values[i][j]             # <<<<<<<<<<<<<<
 * 	setFracSComp(<double **> sc)
 * 	#for i in range(0, npc):
 */
      if (unlikely(__pyx_v_values == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 465, __pyx_L1_error)
      }
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_values, __pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 465, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_14 = __Pyx_PyObject_GetItem(__pyx_t_3, __pyx_v_j); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 465, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_15 = __pyx_PyFloat_AsDouble(__pyx_t_14); if (unlikely((__pyx_t_15 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 465, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_16 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_16 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 465, __pyx_L1_error)
      __pyx_t_17 = __Pyx_PyIndex_AsSsize_t(__pyx_v_j); if (unlikely((__pyx_t_17 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 465, __pyx_L1_error)
      ((__pyx_v_sc[__pyx_t_16])[__pyx_t_17]) = __pyx_t_15;

      /* "melts/rMELTSframework.pyx":464
 * 	for i in range(0, npc):
 * 		sc[i] = <double *>malloc(max(ns[i],1)*sizeof(double))
 * 		for j in range(0, max(ns[i],1)):             # <<<<<<<<<<<<<<
 * 			sc[i][j] = values[i][j]
 * 	setFracSComp(<double **> sc)
 */
    }
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

    /* "melts/rMELTSframework.pyx":462
 * 	npc = num_phase_components()
 * 	cdef double **sc = <double **>malloc(npc*sizeof(double *))
 * 	for i in range(0, npc):             # <<<<<<<<<<<<<<
 * 		sc[i] = <double *>malloc(max(ns[i],1)*sizeof(double))
 * 		for j in range(0, max(ns[i],1)):
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "melts/rMELTSframework.pyx":466
 * 		for j in range(0, max(ns[i],1)):
 * 			sc[i][j] = values[i][j]
 * 	setFracSComp(<double **> sc)             # <<<<<<<<<<<<<<
 * 	#for i in range(0, npc):
 * 	#	free(sc[i])
 */
  setFracSComp(((double **)__pyx_v_sc));

  /* "melts/rMELTSframework.pyx":456
 * 	return result
 * 
 * def cy_setFracSCompAsListOfLists(list values = [[]]):             # <<<<<<<<<<<<<<
 * 	if values is None:
 * 		return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setFracSCompAsListOfLists", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_npc);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_j);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":471
 * 	#free (sc)  memory management in C function
 * 
 * def cy_fracLCompAsList():             # <<<<<<<<<<<<<<
 * 	nc = num_oxides()
 * 	cdef double *nl = fracLComp()
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_171cy_fracLCompAsList(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_170cy_fracLCompAsList[] = "cy_fracLCompAsList()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_171cy_fracLCompAsList = {"cy_fracLCompAsList", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_171cy_fracLCompAsList, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_170cy_fracLCompAsList};
static PyObject *__pyx_pw_5melts_15rMELTSframework_171cy_fracLCompAsList(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_fracLCompAsList (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_170cy_fracLCompAsList(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_170cy_fracLCompAsList(CYTHON_UNUSED PyObject *__pyx_self) {
  int __pyx_v_nc;
  double *__pyx_v_nl;
  PyObject *__pyx_v_result = NULL;
  long __pyx_8genexpr6__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  long __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_fracLCompAsList", 0);

  /* "melts/rMELTSframework.pyx":472
 * 
 * def cy_fracLCompAsList():
 * 	nc = num_oxides()             # <<<<<<<<<<<<<<
 * 	cdef double *nl = fracLComp()
 * 	if nl == NULL:
 */
  __pyx_v_nc = num_oxides();

  /* "melts/rMELTSframework.pyx":473
 * def cy_fracLCompAsList():
 * 	nc = num_oxides()
 * 	cdef double *nl = fracLComp()             # <<<<<<<<<<<<<<
 * 	if nl == NULL:
 * 		return None
 */
  __pyx_v_nl = fracLComp();

  /* "melts/rMELTSframework.pyx":474
 * 	nc = num_oxides()
 * 	cdef double *nl = fracLComp()
 * 	if nl == NULL:             # <<<<<<<<<<<<<<
 * 		return None
 * 	result = [<double> nl[i] for i in range(0,nc)]
 */
  __pyx_t_1 = ((__pyx_v_nl == NULL) != 0);
  if (__pyx_t_1) {

    /* "melts/rMELTSframework.pyx":475
 * 	cdef double *nl = fracLComp()
 * 	if nl == NULL:
 * 		return None             # <<<<<<<<<<<<<<
 * 	result = [<double> nl[i] for i in range(0,nc)]
 * 	return result
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "melts/rMELTSframework.pyx":474
 * 	nc = num_oxides()
 * 	cdef double *nl = fracLComp()
 * 	if nl == NULL:             # <<<<<<<<<<<<<<
 * 		return None
 * 	result = [<double> nl[i] for i in range(0,nc)]
 */
  }

  /* "melts/rMELTSframework.pyx":476
 * 	if nl == NULL:
 * 		return None
 * 	result = [<double> nl[i] for i in range(0,nc)]             # <<<<<<<<<<<<<<
 * 	return result
 * 
 */
  { /* enter inner scope */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_v_nc;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_8genexpr6__pyx_v_i = __pyx_t_5;
      __pyx_t_6 = PyFloat_FromDouble(((double)(__pyx_v_nl[__pyx_8genexpr6__pyx_v_i]))); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 476, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 476, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  } /* exit inner scope */
  __pyx_v_result = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "melts/rMELTSframework.pyx":477
 * 		return None
 * 	result = [<double> nl[i] for i in range(0,nc)]
 * 	return result             # <<<<<<<<<<<<<<
 * 
 * def cy_setFracLCompAsList(list values = []):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":471
 * 	#free (sc)  memory management in C function
 * 
 * def cy_fracLCompAsList():             # <<<<<<<<<<<<<<
 * 	nc = num_oxides()
 * 	cdef double *nl = fracLComp()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_fracLCompAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":479
 * 	return result
 * 
 * def cy_setFracLCompAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	if values is None:
 * 		return
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_173cy_setFracLCompAsList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_172cy_setFracLCompAsList[] = "cy_setFracLCompAsList(list values=[])";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_173cy_setFracLCompAsList = {"cy_setFracLCompAsList", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5melts_15rMELTSframework_173cy_setFracLCompAsList, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5melts_15rMELTSframework_172cy_setFracLCompAsList};
static PyObject *__pyx_pw_5melts_15rMELTSframework_173cy_setFracLCompAsList(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_values = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFracLCompAsList (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_values,0};
    PyObject* values[1] = {0};
    values[0] = __pyx_k__8;
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_values);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cy_setFracLCompAsList") < 0)) __PYX_ERR(0, 479, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_values = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cy_setFracLCompAsList", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 479, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setFracLCompAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_values), (&PyList_Type), 1, "values", 1))) __PYX_ERR(0, 479, __pyx_L1_error)
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_172cy_setFracLCompAsList(__pyx_self, __pyx_v_values);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_172cy_setFracLCompAsList(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_values) {
  PyObject *__pyx_v_nc = NULL;
  double *__pyx_v_nl;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  double __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_setFracLCompAsList", 0);

  /* "melts/rMELTSframework.pyx":480
 * 
 * def cy_setFracLCompAsList(list values = []):
 * 	if values is None:             # <<<<<<<<<<<<<<
 * 		return
 * 	nc = num_oxides()
 */
  __pyx_t_1 = (__pyx_v_values == ((PyObject*)Py_None));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "melts/rMELTSframework.pyx":481
 * def cy_setFracLCompAsList(list values = []):
 * 	if values is None:
 * 		return             # <<<<<<<<<<<<<<
 * 	nc = num_oxides()
 * 	cdef double *nl = <double *>malloc(nc*sizeof(double))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "melts/rMELTSframework.pyx":480
 * 
 * def cy_setFracLCompAsList(list values = []):
 * 	if values is None:             # <<<<<<<<<<<<<<
 * 		return
 * 	nc = num_oxides()
 */
  }

  /* "melts/rMELTSframework.pyx":482
 * 	if values is None:
 * 		return
 * 	nc = num_oxides()             # <<<<<<<<<<<<<<
 * 	cdef double *nl = <double *>malloc(nc*sizeof(double))
 * 	for i in range(0,nc):
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(num_oxides()); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_nc = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":483
 * 		return
 * 	nc = num_oxides()
 * 	cdef double *nl = <double *>malloc(nc*sizeof(double))             # <<<<<<<<<<<<<<
 * 	for i in range(0,nc):
 * 		nl[i] = <double> values[i]
 */
  __pyx_t_3 = __Pyx_PyInt_FromSize_t((sizeof(double))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyNumber_Multiply(__pyx_v_nc, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyInt_As_size_t(__pyx_t_4); if (unlikely((__pyx_t_5 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 483, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_nl = ((double *)malloc(__pyx_t_5));

  /* "melts/rMELTSframework.pyx":484
 * 	nc = num_oxides()
 * 	cdef double *nl = <double *>malloc(nc*sizeof(double))
 * 	for i in range(0,nc):             # <<<<<<<<<<<<<<
 * 		nl[i] = <double> values[i]
 * 	setFracLComp(<double *> nl)
 */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_int_0);
  __Pyx_INCREF(__pyx_v_nc);
  __Pyx_GIVEREF(__pyx_v_nc);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_nc);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 484, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 484, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 484, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 484, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 484, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_7(__pyx_t_4);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 484, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "melts/rMELTSframework.pyx":485
 * 	cdef double *nl = <double *>malloc(nc*sizeof(double))
 * 	for i in range(0,nc):
 * 		nl[i] = <double> values[i]             # <<<<<<<<<<<<<<
 * 	setFracLComp(<double *> nl)
 * 	#free (nl)  memory management in C function
 */
    if (unlikely(__pyx_v_values == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 485, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v_values, __pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 485, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 485, __pyx_L1_error)
    (__pyx_v_nl[__pyx_t_9]) = ((double)__pyx_t_8);

    /* "melts/rMELTSframework.pyx":484
 * 	nc = num_oxides()
 * 	cdef double *nl = <double *>malloc(nc*sizeof(double))
 * 	for i in range(0,nc):             # <<<<<<<<<<<<<<
 * 		nl[i] = <double> values[i]
 * 	setFracLComp(<double *> nl)
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "melts/rMELTSframework.pyx":486
 * 	for i in range(0,nc):
 * 		nl[i] = <double> values[i]
 * 	setFracLComp(<double *> nl)             # <<<<<<<<<<<<<<
 * 	#free (nl)  memory management in C function
 * 
 */
  setFracLComp(((double *)__pyx_v_nl));

  /* "melts/rMELTSframework.pyx":479
 * 	return result
 * 
 * def cy_setFracLCompAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	if values is None:
 * 		return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setFracLCompAsList", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_nc);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":489
 * 	#free (nl)  memory management in C function
 * 
 * def cy_fracMass():             # <<<<<<<<<<<<<<
 * 	return <double> fracMass()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_175cy_fracMass(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_174cy_fracMass[] = "cy_fracMass()";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_175cy_fracMass = {"cy_fracMass", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_175cy_fracMass, METH_NOARGS, __pyx_doc_5melts_15rMELTSframework_174cy_fracMass};
static PyObject *__pyx_pw_5melts_15rMELTSframework_175cy_fracMass(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_fracMass (wrapper)", 0);
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_174cy_fracMass(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_174cy_fracMass(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cy_fracMass", 0);

  /* "melts/rMELTSframework.pyx":490
 * 
 * def cy_fracMass():
 * 	return <double> fracMass()             # <<<<<<<<<<<<<<
 * 
 * def cy_setFracMass(double value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(((double)fracMass())); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "melts/rMELTSframework.pyx":489
 * 	#free (nl)  memory management in C function
 * 
 * def cy_fracMass():             # <<<<<<<<<<<<<<
 * 	return <double> fracMass()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("melts.rMELTSframework.cy_fracMass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "melts/rMELTSframework.pyx":492
 * 	return <double> fracMass()
 * 
 * def cy_setFracMass(double value):             # <<<<<<<<<<<<<<
 * 	setFracMass = <double> value
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5melts_15rMELTSframework_177cy_setFracMass(PyObject *__pyx_self, PyObject *__pyx_arg_value); /*proto*/
static char __pyx_doc_5melts_15rMELTSframework_176cy_setFracMass[] = "cy_setFracMass(double value)";
static PyMethodDef __pyx_mdef_5melts_15rMELTSframework_177cy_setFracMass = {"cy_setFracMass", (PyCFunction)__pyx_pw_5melts_15rMELTSframework_177cy_setFracMass, METH_O, __pyx_doc_5melts_15rMELTSframework_176cy_setFracMass};
static PyObject *__pyx_pw_5melts_15rMELTSframework_177cy_setFracMass(PyObject *__pyx_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFracMass (wrapper)", 0);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 492, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("melts.rMELTSframework.cy_setFracMass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5melts_15rMELTSframework_176cy_setFracMass(__pyx_self, ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5melts_15rMELTSframework_176cy_setFracMass(CYTHON_UNUSED PyObject *__pyx_self, double __pyx_v_value) {
  CYTHON_UNUSED double __pyx_v_setFracMass;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cy_setFracMass", 0);

  /* "melts/rMELTSframework.pyx":493
 * 
 * def cy_setFracMass(double value):
 * 	setFracMass = <double> value             # <<<<<<<<<<<<<<
 * 
 */
  __pyx_v_setFracMass = ((double)__pyx_v_value);

  /* "melts/rMELTSframework.pyx":492
 * 	return <double> fracMass()
 * 
 * def cy_setFracMass(double value):             # <<<<<<<<<<<<<<
 * 	setFracMass = <double> value
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":735
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":736
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":735
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":738
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":739
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":738
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":741
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":742
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":741
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":744
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":745
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":744
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":747
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":748
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":747
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":750
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 0);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":751
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = (PyDataType_HASSUBARRAY(__pyx_v_d) != 0);
  if (__pyx_t_1) {

    /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":752
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":751
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":754
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":750
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":929
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":930
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":931
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  (void)(PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base));

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":929
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":933
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":934
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":935
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = ((__pyx_v_base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":936
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":935
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":937
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":933
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":941
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":943
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 943, __pyx_L3_error)

      /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":944
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 944, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":945
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 945, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 945, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":942
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":941
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":947
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":949
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 949, __pyx_L3_error)

      /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":950
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 950, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":951
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 951, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 951, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":948
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":947
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":953
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":955
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 955, __pyx_L3_error)

      /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":956
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 956, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":957
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef extern from *:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 957, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 957, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":954
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":953
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":967
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_timedelta64_object", 0);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":979
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":967
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":982
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_datetime64_object", 0);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":994
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":982
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":997
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":1004
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":997
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":1007
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":1011
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":1007
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":1014
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":1018
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":1014
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_rMELTSframework(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_rMELTSframework},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "rMELTSframework",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_s_bc, __pyx_k_bc, sizeof(__pyx_k_bc), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_ctypes, __pyx_k_ctypes, sizeof(__pyx_k_ctypes), 0, 0, 1, 1},
  {&__pyx_n_s_cy_MELTSandCO2_H2OcalculationMod, __pyx_k_cy_MELTSandCO2_H2OcalculationMod, sizeof(__pyx_k_cy_MELTSandCO2_H2OcalculationMod), 0, 0, 1, 1},
  {&__pyx_n_s_cy_MELTSandCO2calculationModeCon, __pyx_k_cy_MELTSandCO2calculationModeCon, sizeof(__pyx_k_cy_MELTSandCO2calculationModeCon), 0, 0, 1, 1},
  {&__pyx_n_s_cy_MELTScalculationModeConstant, __pyx_k_cy_MELTScalculationModeConstant, sizeof(__pyx_k_cy_MELTScalculationModeConstant), 0, 0, 1, 1},
  {&__pyx_n_s_cy_bulkCompAsList, __pyx_k_cy_bulkCompAsList, sizeof(__pyx_k_cy_bulkCompAsList), 0, 0, 1, 1},
  {&__pyx_n_s_cy_finalEnthalpy, __pyx_k_cy_finalEnthalpy, sizeof(__pyx_k_cy_finalEnthalpy), 0, 0, 1, 1},
  {&__pyx_n_s_cy_finalEntropy, __pyx_k_cy_finalEntropy, sizeof(__pyx_k_cy_finalEntropy), 0, 0, 1, 1},
  {&__pyx_n_s_cy_finalPressure, __pyx_k_cy_finalPressure, sizeof(__pyx_k_cy_finalPressure), 0, 0, 1, 1},
  {&__pyx_n_s_cy_finalTemperature, __pyx_k_cy_finalTemperature, sizeof(__pyx_k_cy_finalTemperature), 0, 0, 1, 1},
  {&__pyx_n_s_cy_finalVolume, __pyx_k_cy_finalVolume, sizeof(__pyx_k_cy_finalVolume), 0, 0, 1, 1},
  {&__pyx_n_s_cy_fo2, __pyx_k_cy_fo2, sizeof(__pyx_k_cy_fo2), 0, 0, 1, 1},
  {&__pyx_n_s_cy_fo2Delta, __pyx_k_cy_fo2Delta, sizeof(__pyx_k_cy_fo2Delta), 0, 0, 1, 1},
  {&__pyx_n_s_cy_fo2Path, __pyx_k_cy_fo2Path, sizeof(__pyx_k_cy_fo2Path), 0, 0, 1, 1},
  {&__pyx_n_s_cy_fracLCompAsList, __pyx_k_cy_fracLCompAsList, sizeof(__pyx_k_cy_fracLCompAsList), 0, 0, 1, 1},
  {&__pyx_n_s_cy_fracMass, __pyx_k_cy_fracMass, sizeof(__pyx_k_cy_fracMass), 0, 0, 1, 1},
  {&__pyx_n_s_cy_fracSCompAsListOfLists, __pyx_k_cy_fracSCompAsListOfLists, sizeof(__pyx_k_cy_fracSCompAsListOfLists), 0, 0, 1, 1},
  {&__pyx_n_s_cy_fractionateFlu, __pyx_k_cy_fractionateFlu, sizeof(__pyx_k_cy_fractionateFlu), 0, 0, 1, 1},
  {&__pyx_n_s_cy_fractionateLiq, __pyx_k_cy_fractionateLiq, sizeof(__pyx_k_cy_fractionateLiq), 0, 0, 1, 1},
  {&__pyx_n_s_cy_fractionateSol, __pyx_k_cy_fractionateSol, sizeof(__pyx_k_cy_fractionateSol), 0, 0, 1, 1},
  {&__pyx_n_s_cy_incSolidsAsList, __pyx_k_cy_incSolidsAsList, sizeof(__pyx_k_cy_incSolidsAsList), 0, 0, 1, 1},
  {&__pyx_n_s_cy_incrementEnthalpy, __pyx_k_cy_incrementEnthalpy, sizeof(__pyx_k_cy_incrementEnthalpy), 0, 0, 1, 1},
  {&__pyx_n_s_cy_incrementEntropy, __pyx_k_cy_incrementEntropy, sizeof(__pyx_k_cy_incrementEntropy), 0, 0, 1, 1},
  {&__pyx_n_s_cy_incrementPressure, __pyx_k_cy_incrementPressure, sizeof(__pyx_k_cy_incrementPressure), 0, 0, 1, 1},
  {&__pyx_n_s_cy_incrementTemperature, __pyx_k_cy_incrementTemperature, sizeof(__pyx_k_cy_incrementTemperature), 0, 0, 1, 1},
  {&__pyx_n_s_cy_incrementVolume, __pyx_k_cy_incrementVolume, sizeof(__pyx_k_cy_incrementVolume), 0, 0, 1, 1},
  {&__pyx_n_s_cy_init, __pyx_k_cy_init, sizeof(__pyx_k_cy_init), 0, 0, 1, 1},
  {&__pyx_n_s_cy_initialEnthalpy, __pyx_k_cy_initialEnthalpy, sizeof(__pyx_k_cy_initialEnthalpy), 0, 0, 1, 1},
  {&__pyx_n_s_cy_initialEntropy, __pyx_k_cy_initialEntropy, sizeof(__pyx_k_cy_initialEntropy), 0, 0, 1, 1},
  {&__pyx_n_s_cy_initialPressure, __pyx_k_cy_initialPressure, sizeof(__pyx_k_cy_initialPressure), 0, 0, 1, 1},
  {&__pyx_n_s_cy_initialTemperature, __pyx_k_cy_initialTemperature, sizeof(__pyx_k_cy_initialTemperature), 0, 0, 1, 1},
  {&__pyx_n_s_cy_initialVolume, __pyx_k_cy_initialVolume, sizeof(__pyx_k_cy_initialVolume), 0, 0, 1, 1},
  {&__pyx_n_s_cy_isIsenthalpic, __pyx_k_cy_isIsenthalpic, sizeof(__pyx_k_cy_isIsenthalpic), 0, 0, 1, 1},
  {&__pyx_n_s_cy_isIsochoric, __pyx_k_cy_isIsochoric, sizeof(__pyx_k_cy_isIsochoric), 0, 0, 1, 1},
  {&__pyx_n_s_cy_isIsotropic, __pyx_k_cy_isIsotropic, sizeof(__pyx_k_cy_isIsotropic), 0, 0, 1, 1},
  {&__pyx_n_s_cy_liquidCompAsListOfLists, __pyx_k_cy_liquidCompAsListOfLists, sizeof(__pyx_k_cy_liquidCompAsListOfLists), 0, 0, 1, 1},
  {&__pyx_n_s_cy_liquidMass, __pyx_k_cy_liquidMass, sizeof(__pyx_k_cy_liquidMass), 0, 0, 1, 1},
  {&__pyx_n_s_cy_multipleLiqs, __pyx_k_cy_multipleLiqs, sizeof(__pyx_k_cy_multipleLiqs), 0, 0, 1, 1},
  {&__pyx_n_s_cy_nFracCoexistAsList, __pyx_k_cy_nFracCoexistAsList, sizeof(__pyx_k_cy_nFracCoexistAsList), 0, 0, 1, 1},
  {&__pyx_n_s_cy_nLiquidCoexist, __pyx_k_cy_nLiquidCoexist, sizeof(__pyx_k_cy_nLiquidCoexist), 0, 0, 1, 1},
  {&__pyx_n_s_cy_nSolidCoexistAsList, __pyx_k_cy_nSolidCoexistAsList, sizeof(__pyx_k_cy_nSolidCoexistAsList), 0, 0, 1, 1},
  {&__pyx_n_s_cy_num_oxides, __pyx_k_cy_num_oxides, sizeof(__pyx_k_cy_num_oxides), 0, 0, 1, 1},
  {&__pyx_n_s_cy_num_phase_components, __pyx_k_cy_num_phase_components, sizeof(__pyx_k_cy_num_phase_components), 0, 0, 1, 1},
  {&__pyx_n_s_cy_oxideListAsStrings, __pyx_k_cy_oxideListAsStrings, sizeof(__pyx_k_cy_oxideListAsStrings), 0, 0, 1, 1},
  {&__pyx_n_s_cy_oxygenContent, __pyx_k_cy_oxygenContent, sizeof(__pyx_k_cy_oxygenContent), 0, 0, 1, 1},
  {&__pyx_n_s_cy_pMELTScalculationModeConstant, __pyx_k_cy_pMELTScalculationModeConstant, sizeof(__pyx_k_cy_pMELTScalculationModeConstant), 0, 0, 1, 1},
  {&__pyx_n_s_cy_parseAndLoadDataStructuresFro, __pyx_k_cy_parseAndLoadDataStructuresFro, sizeof(__pyx_k_cy_parseAndLoadDataStructuresFro), 0, 0, 1, 1},
  {&__pyx_n_s_cy_performMELTScalculation, __pyx_k_cy_performMELTScalculation, sizeof(__pyx_k_cy_performMELTScalculation), 0, 0, 1, 1},
  {&__pyx_n_s_cy_phaseListAsStrings, __pyx_k_cy_phaseListAsStrings, sizeof(__pyx_k_cy_phaseListAsStrings), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setBulkCompAsList, __pyx_k_cy_setBulkCompAsList, sizeof(__pyx_k_cy_setBulkCompAsList), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setFinalEnthalpy, __pyx_k_cy_setFinalEnthalpy, sizeof(__pyx_k_cy_setFinalEnthalpy), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setFinalEntropy, __pyx_k_cy_setFinalEntropy, sizeof(__pyx_k_cy_setFinalEntropy), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setFinalPressure, __pyx_k_cy_setFinalPressure, sizeof(__pyx_k_cy_setFinalPressure), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setFinalTemperature, __pyx_k_cy_setFinalTemperature, sizeof(__pyx_k_cy_setFinalTemperature), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setFinalVolume, __pyx_k_cy_setFinalVolume, sizeof(__pyx_k_cy_setFinalVolume), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setFracLCompAsList, __pyx_k_cy_setFracLCompAsList, sizeof(__pyx_k_cy_setFracLCompAsList), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setFracMass, __pyx_k_cy_setFracMass, sizeof(__pyx_k_cy_setFracMass), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setFracSCompAsListOfLists, __pyx_k_cy_setFracSCompAsListOfLists, sizeof(__pyx_k_cy_setFracSCompAsListOfLists), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setFractionateFlu, __pyx_k_cy_setFractionateFlu, sizeof(__pyx_k_cy_setFractionateFlu), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setFractionateLiq, __pyx_k_cy_setFractionateLiq, sizeof(__pyx_k_cy_setFractionateLiq), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setFractionateSol, __pyx_k_cy_setFractionateSol, sizeof(__pyx_k_cy_setFractionateSol), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setIncSolidsAsList, __pyx_k_cy_setIncSolidsAsList, sizeof(__pyx_k_cy_setIncSolidsAsList), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setIncrementEnthalpy, __pyx_k_cy_setIncrementEnthalpy, sizeof(__pyx_k_cy_setIncrementEnthalpy), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setIncrementEntropy, __pyx_k_cy_setIncrementEntropy, sizeof(__pyx_k_cy_setIncrementEntropy), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setIncrementPressure, __pyx_k_cy_setIncrementPressure, sizeof(__pyx_k_cy_setIncrementPressure), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setIncrementTemperature, __pyx_k_cy_setIncrementTemperature, sizeof(__pyx_k_cy_setIncrementTemperature), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setIncrementVolume, __pyx_k_cy_setIncrementVolume, sizeof(__pyx_k_cy_setIncrementVolume), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setInitialEnthalpy, __pyx_k_cy_setInitialEnthalpy, sizeof(__pyx_k_cy_setInitialEnthalpy), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setInitialEntropy, __pyx_k_cy_setInitialEntropy, sizeof(__pyx_k_cy_setInitialEntropy), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setInitialPressure, __pyx_k_cy_setInitialPressure, sizeof(__pyx_k_cy_setInitialPressure), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setInitialTemperature, __pyx_k_cy_setInitialTemperature, sizeof(__pyx_k_cy_setInitialTemperature), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setInitialVolume, __pyx_k_cy_setInitialVolume, sizeof(__pyx_k_cy_setInitialVolume), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setIsenthalpic, __pyx_k_cy_setIsenthalpic, sizeof(__pyx_k_cy_setIsenthalpic), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setIsochoric, __pyx_k_cy_setIsochoric, sizeof(__pyx_k_cy_setIsochoric), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setIsotropic, __pyx_k_cy_setIsotropic, sizeof(__pyx_k_cy_setIsotropic), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setLiquidCompAsListOfLists, __pyx_k_cy_setLiquidCompAsListOfLists, sizeof(__pyx_k_cy_setLiquidCompAsListOfLists), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setLiquidMass, __pyx_k_cy_setLiquidMass, sizeof(__pyx_k_cy_setLiquidMass), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setMultipleLiqs, __pyx_k_cy_setMultipleLiqs, sizeof(__pyx_k_cy_setMultipleLiqs), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setOxygenContent, __pyx_k_cy_setOxygenContent, sizeof(__pyx_k_cy_setOxygenContent), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setSolidCompAsListOfLists, __pyx_k_cy_setSolidCompAsListOfLists, sizeof(__pyx_k_cy_setSolidCompAsListOfLists), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setSolidMass, __pyx_k_cy_setSolidMass, sizeof(__pyx_k_cy_setSolidMass), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setfo2, __pyx_k_cy_setfo2, sizeof(__pyx_k_cy_setfo2), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setfo2Delta, __pyx_k_cy_setfo2Delta, sizeof(__pyx_k_cy_setfo2Delta), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setfo2Path, __pyx_k_cy_setfo2Path, sizeof(__pyx_k_cy_setfo2Path), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setnFracCoexistAsList, __pyx_k_cy_setnFracCoexistAsList, sizeof(__pyx_k_cy_setnFracCoexistAsList), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setnLiquidCoexist, __pyx_k_cy_setnLiquidCoexist, sizeof(__pyx_k_cy_setnLiquidCoexist), 0, 0, 1, 1},
  {&__pyx_n_s_cy_setnSolidCoexistAsList, __pyx_k_cy_setnSolidCoexistAsList, sizeof(__pyx_k_cy_setnSolidCoexistAsList), 0, 0, 1, 1},
  {&__pyx_n_s_cy_solidCompAsListOfLists, __pyx_k_cy_solidCompAsListOfLists, sizeof(__pyx_k_cy_solidCompAsListOfLists), 0, 0, 1, 1},
  {&__pyx_n_s_cy_solidMass, __pyx_k_cy_solidMass, sizeof(__pyx_k_cy_solidMass), 0, 0, 1, 1},
  {&__pyx_n_s_cy_writeDataStructuresToXMLstrin, __pyx_k_cy_writeDataStructuresToXMLstrin, sizeof(__pyx_k_cy_writeDataStructuresToXMLstrin), 0, 0, 1, 1},
  {&__pyx_n_s_cy_xMELTScalculationModeConstant, __pyx_k_cy_xMELTScalculationModeConstant, sizeof(__pyx_k_cy_xMELTScalculationModeConstant), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_entry, __pyx_k_entry, sizeof(__pyx_k_entry), 0, 0, 1, 1},
  {&__pyx_n_s_hValue, __pyx_k_hValue, sizeof(__pyx_k_hValue), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_inc, __pyx_k_inc, sizeof(__pyx_k_inc), 0, 0, 1, 1},
  {&__pyx_n_s_inputXMLstring, __pyx_k_inputXMLstring, sizeof(__pyx_k_inputXMLstring), 0, 0, 1, 1},
  {&__pyx_n_s_j, __pyx_k_j, sizeof(__pyx_k_j), 0, 0, 1, 1},
  {&__pyx_n_s_lc, __pyx_k_lc, sizeof(__pyx_k_lc), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_melts_rMELTSframework, __pyx_k_melts_rMELTSframework, sizeof(__pyx_k_melts_rMELTSframework), 0, 0, 1, 1},
  {&__pyx_n_s_modelSelection, __pyx_k_modelSelection, sizeof(__pyx_k_modelSelection), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_names, __pyx_k_names, sizeof(__pyx_k_names), 0, 0, 1, 1},
  {&__pyx_n_s_nc, __pyx_k_nc, sizeof(__pyx_k_nc), 0, 0, 1, 1},
  {&__pyx_n_s_nl, __pyx_k_nl, sizeof(__pyx_k_nl), 0, 0, 1, 1},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_npc, __pyx_k_npc, sizeof(__pyx_k_npc), 0, 0, 1, 1},
  {&__pyx_n_s_ns, __pyx_k_ns, sizeof(__pyx_k_ns), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_u_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 1, 0, 0},
  {&__pyx_kp_u_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 1, 0, 0},
  {&__pyx_n_s_pValue, __pyx_k_pValue, sizeof(__pyx_k_pValue), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_s_result_b, __pyx_k_result_b, sizeof(__pyx_k_result_b), 0, 0, 1, 1},
  {&__pyx_n_s_sValue, __pyx_k_sValue, sizeof(__pyx_k_sValue), 0, 0, 1, 1},
  {&__pyx_n_s_sc, __pyx_k_sc, sizeof(__pyx_k_sc), 0, 0, 1, 1},
  {&__pyx_n_s_sessionID, __pyx_k_sessionID, sizeof(__pyx_k_sessionID), 0, 0, 1, 1},
  {&__pyx_n_s_setFracMass, __pyx_k_setFracMass, sizeof(__pyx_k_setFracMass), 0, 0, 1, 1},
  {&__pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_k_src_melts_rMELTSframework_pyx, sizeof(__pyx_k_src_melts_rMELTSframework_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_tValue, __pyx_k_tValue, sizeof(__pyx_k_tValue), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_typeOfCalc, __pyx_k_typeOfCalc, sizeof(__pyx_k_typeOfCalc), 0, 0, 1, 1},
  {&__pyx_kp_u_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 1, 0, 0},
  {&__pyx_n_s_vValue, __pyx_k_vValue, sizeof(__pyx_k_vValue), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 135, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 945, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":945
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(1, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":951
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(1, 951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "melts/rMELTSframework.pyx":110
 * # here is the "wrapper" signature
 * 
 * def cy_MELTScalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return MELTScalculationModeConstant()
 * 
 */
  __pyx_codeobj__11 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_MELTScalculationModeConstant, 110, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__11)) __PYX_ERR(0, 110, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":113
 * 	return MELTScalculationModeConstant()
 * 
 * def cy_pMELTScalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return pMELTScalculationModeConstant()
 * 
 */
  __pyx_codeobj__12 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_pMELTScalculationModeConstant, 113, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__12)) __PYX_ERR(0, 113, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":116
 * 	return pMELTScalculationModeConstant()
 * 
 * def cy_xMELTScalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return xMELTScalculationModeConstant()
 * 
 */
  __pyx_codeobj__13 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_xMELTScalculationModeConstant, 116, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__13)) __PYX_ERR(0, 116, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":119
 * 	return xMELTScalculationModeConstant()
 * 
 * def cy_MELTSandCO2calculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return MELTSandCO2calculationModeConstant()
 * 
 */
  __pyx_codeobj__14 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_MELTSandCO2calculationModeCon, 119, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__14)) __PYX_ERR(0, 119, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":122
 * 	return MELTSandCO2calculationModeConstant()
 * 
 * def cy_MELTSandCO2_H2OcalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return MELTSandCO2_H2OcalculationModeConstant()
 * 
 */
  __pyx_codeobj__15 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_MELTSandCO2_H2OcalculationMod, 122, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__15)) __PYX_ERR(0, 122, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":125
 * 	return MELTSandCO2_H2OcalculationModeConstant()
 * 
 * def cy_num_oxides():             # <<<<<<<<<<<<<<
 * 	return num_oxides()
 * 
 */
  __pyx_codeobj__16 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_num_oxides, 125, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__16)) __PYX_ERR(0, 125, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":128
 * 	return num_oxides()
 * 
 * def cy_num_phase_components():             # <<<<<<<<<<<<<<
 * 	return num_phase_components()
 * 
 */
  __pyx_codeobj__17 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_num_phase_components, 128, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__17)) __PYX_ERR(0, 128, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":131
 * 	return num_phase_components()
 * 
 * def cy_oxideListAsStrings():             # <<<<<<<<<<<<<<
 * 	cdef char **names = oxideListAsStrings()
 * 	n = num_oxides()
 */
  __pyx_tuple__18 = PyTuple_Pack(5, __pyx_n_s_names, __pyx_n_s_n, __pyx_n_s_result, __pyx_n_s_i, __pyx_n_s_entry); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);
  __pyx_codeobj__19 = (PyObject*)__Pyx_PyCode_New(0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__18, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_oxideListAsStrings, 131, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__19)) __PYX_ERR(0, 131, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":140
 * 	return result
 * 
 * def cy_phaseListAsStrings(int modelSelection):             # <<<<<<<<<<<<<<
 * 	cdef char **names = phaseListAsStrings(modelSelection)
 * 	n = num_phases(modelSelection)
 */
  __pyx_tuple__20 = PyTuple_Pack(7, __pyx_n_s_modelSelection, __pyx_n_s_modelSelection, __pyx_n_s_names, __pyx_n_s_n, __pyx_n_s_result, __pyx_n_s_i, __pyx_n_s_entry); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);
  __pyx_codeobj__21 = (PyObject*)__Pyx_PyCode_New(1, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__20, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_phaseListAsStrings, 140, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__21)) __PYX_ERR(0, 140, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":149
 * 	return result
 * 
 * def cy_init(int modelSelection):             # <<<<<<<<<<<<<<
 * 	return init(modelSelection)
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(2, __pyx_n_s_modelSelection, __pyx_n_s_modelSelection); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_init, 149, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 149, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":152
 * 	return init(modelSelection)
 * 
 * def cy_parseAndLoadDataStructuresFromXMLstring(inputXMLstring):             # <<<<<<<<<<<<<<
 * 	return parseAndLoadDataStructuresFromXMLstring(inputXMLstring.encode('utf-8'))
 * 
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_n_s_inputXMLstring); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_parseAndLoadDataStructuresFro, 152, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 152, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":155
 * 	return parseAndLoadDataStructuresFromXMLstring(inputXMLstring.encode('utf-8'))
 * 
 * def cy_writeDataStructuresToXMLstring(sessionID):             # <<<<<<<<<<<<<<
 * 	result_b = writeDataStructuresToXMLstring(sessionID.encode('utf-8'))
 * 	result = result_b.decode('utf-8')
 */
  __pyx_tuple__26 = PyTuple_Pack(3, __pyx_n_s_sessionID, __pyx_n_s_result_b, __pyx_n_s_result); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_writeDataStructuresToXMLstrin, 155, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 155, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":161
 * 	return result
 * 
 * def cy_performMELTScalculation(typeOfCalc):             # <<<<<<<<<<<<<<
 * 	return performMELTScalculation(typeOfCalc)
 * 
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_n_s_typeOfCalc); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_performMELTScalculation, 161, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 161, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":164
 * 	return performMELTScalculation(typeOfCalc)
 * 
 * def cy_initialTemperature():             # <<<<<<<<<<<<<<
 * 	return initialTemperature()
 * 
 */
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_initialTemperature, 164, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(0, 164, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":167
 * 	return initialTemperature()
 * 
 * def cy_setInitialTemperature(double tValue):             # <<<<<<<<<<<<<<
 * 	setInitialTemperature(tValue)
 * 
 */
  __pyx_tuple__31 = PyTuple_Pack(2, __pyx_n_s_tValue, __pyx_n_s_tValue); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setInitialTemperature, 167, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 167, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":170
 * 	setInitialTemperature(tValue)
 * 
 * def cy_finalTemperature():             # <<<<<<<<<<<<<<
 * 	return finalTemperature()
 * 
 */
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_finalTemperature, 170, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(0, 170, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":173
 * 	return finalTemperature()
 * 
 * def cy_setFinalTemperature(double tValue):             # <<<<<<<<<<<<<<
 * 	setFinalTemperature(tValue)
 * 
 */
  __pyx_tuple__34 = PyTuple_Pack(2, __pyx_n_s_tValue, __pyx_n_s_tValue); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__34, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setFinalTemperature, 173, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 173, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":176
 * 	setFinalTemperature(tValue)
 * 
 * def cy_incrementTemperature():             # <<<<<<<<<<<<<<
 * 	return incrementTemperature()
 * 
 */
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_incrementTemperature, 176, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 176, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":179
 * 	return incrementTemperature()
 * 
 * def cy_setIncrementTemperature(double tValue):             # <<<<<<<<<<<<<<
 * 	setIncrementTemperature(tValue);
 * 
 */
  __pyx_tuple__37 = PyTuple_Pack(2, __pyx_n_s_tValue, __pyx_n_s_tValue); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setIncrementTemperature, 179, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 179, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":182
 * 	setIncrementTemperature(tValue);
 * 
 * def cy_initialPressure():             # <<<<<<<<<<<<<<
 * 	return initialPressure()
 * 
 */
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_initialPressure, 182, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(0, 182, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":185
 * 	return initialPressure()
 * 
 * def cy_setInitialPressure(double pValue):             # <<<<<<<<<<<<<<
 * 	setInitialPressure(pValue)
 * 
 */
  __pyx_tuple__40 = PyTuple_Pack(2, __pyx_n_s_pValue, __pyx_n_s_pValue); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setInitialPressure, 185, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(0, 185, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":188
 * 	setInitialPressure(pValue)
 * 
 * def cy_finalPressure():             # <<<<<<<<<<<<<<
 * 	return finalPressure()
 * 
 */
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_finalPressure, 188, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 188, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":191
 * 	return finalPressure()
 * 
 * def cy_setFinalPressure(double pValue):             # <<<<<<<<<<<<<<
 * 	setFinalPressure(pValue)
 * 
 */
  __pyx_tuple__43 = PyTuple_Pack(2, __pyx_n_s_pValue, __pyx_n_s_pValue); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setFinalPressure, 191, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 191, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":194
 * 	setFinalPressure(pValue)
 * 
 * def cy_incrementPressure():             # <<<<<<<<<<<<<<
 * 	return incrementPressure()
 * 
 */
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_incrementPressure, 194, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(0, 194, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":197
 * 	return incrementPressure()
 * 
 * def cy_setIncrementPressure(double pValue):             # <<<<<<<<<<<<<<
 * 	setIncrementPressure(pValue)
 * 
 */
  __pyx_tuple__46 = PyTuple_Pack(2, __pyx_n_s_pValue, __pyx_n_s_pValue); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setIncrementPressure, 197, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(0, 197, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":200
 * 	setIncrementPressure(pValue)
 * 
 * def cy_initialVolume():             # <<<<<<<<<<<<<<
 * 	return initialVolume()
 * 
 */
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_initialVolume, 200, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 200, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":203
 * 	return initialVolume()
 * 
 * def cy_setInitialVolume(double vValue):             # <<<<<<<<<<<<<<
 * 	setInitialVolume(vValue)
 * 
 */
  __pyx_tuple__49 = PyTuple_Pack(2, __pyx_n_s_vValue, __pyx_n_s_vValue); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setInitialVolume, 203, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 203, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":206
 * 	setInitialVolume(vValue)
 * 
 * def cy_finalVolume():             # <<<<<<<<<<<<<<
 * 	return finalVolume()
 * 
 */
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_finalVolume, 206, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(0, 206, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":209
 * 	return finalVolume()
 * 
 * def cy_setFinalVolume(double vValue):             # <<<<<<<<<<<<<<
 * 	setFinalVolume(vValue)
 * 
 */
  __pyx_tuple__52 = PyTuple_Pack(2, __pyx_n_s_vValue, __pyx_n_s_vValue); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__52, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setFinalVolume, 209, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(0, 209, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":212
 * 	setFinalVolume(vValue)
 * 
 * def cy_incrementVolume():             # <<<<<<<<<<<<<<
 * 	return incrementVolume()
 * 
 */
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_incrementVolume, 212, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 212, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":215
 * 	return incrementVolume()
 * 
 * def cy_setIncrementVolume(double vValue):             # <<<<<<<<<<<<<<
 * 	setIncrementVolume(vValue)
 * 
 */
  __pyx_tuple__55 = PyTuple_Pack(2, __pyx_n_s_vValue, __pyx_n_s_vValue); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__55, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setIncrementVolume, 215, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(0, 215, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":218
 * 	setIncrementVolume(vValue)
 * 
 * def cy_initialEnthalpy():             # <<<<<<<<<<<<<<
 * 	return initialEnthalpy()
 * 
 */
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_initialEnthalpy, 218, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(0, 218, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":221
 * 	return initialEnthalpy()
 * 
 * def cy_setInitialEnthalpy(double hValue):             # <<<<<<<<<<<<<<
 * 	setInitialEnthalpy(hValue)
 * 
 */
  __pyx_tuple__58 = PyTuple_Pack(2, __pyx_n_s_hValue, __pyx_n_s_hValue); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);
  __pyx_codeobj__59 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__58, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setInitialEnthalpy, 221, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__59)) __PYX_ERR(0, 221, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":224
 * 	setInitialEnthalpy(hValue)
 * 
 * def cy_finalEnthalpy():             # <<<<<<<<<<<<<<
 * 	return finalEnthalpy()
 * 
 */
  __pyx_codeobj__60 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_finalEnthalpy, 224, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__60)) __PYX_ERR(0, 224, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":227
 * 	return finalEnthalpy()
 * 
 * def cy_setFinalEnthalpy(double hValue):             # <<<<<<<<<<<<<<
 * 	setFinalEnthalpy(hValue)
 * 
 */
  __pyx_tuple__61 = PyTuple_Pack(2, __pyx_n_s_hValue, __pyx_n_s_hValue); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__61, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setFinalEnthalpy, 227, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(0, 227, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":230
 * 	setFinalEnthalpy(hValue)
 * 
 * def cy_incrementEnthalpy():             # <<<<<<<<<<<<<<
 * 	return incrementEnthalpy()
 * 
 */
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_incrementEnthalpy, 230, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(0, 230, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":233
 * 	return incrementEnthalpy()
 * 
 * def cy_setIncrementEnthalpy(double hValue):             # <<<<<<<<<<<<<<
 * 	setIncrementEnthalpy(hValue)
 * 
 */
  __pyx_tuple__64 = PyTuple_Pack(2, __pyx_n_s_hValue, __pyx_n_s_hValue); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);
  __pyx_codeobj__65 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__64, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setIncrementEnthalpy, 233, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__65)) __PYX_ERR(0, 233, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":236
 * 	setIncrementEnthalpy(hValue)
 * 
 * def cy_initialEntropy():             # <<<<<<<<<<<<<<
 * 	return initialEntropy()
 * 
 */
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_initialEntropy, 236, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(0, 236, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":239
 * 	return initialEntropy()
 * 
 * def cy_setInitialEntropy(double sValue):             # <<<<<<<<<<<<<<
 * 	setInitialEntropy(sValue)
 * 
 */
  __pyx_tuple__67 = PyTuple_Pack(2, __pyx_n_s_sValue, __pyx_n_s_sValue); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__67, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setInitialEntropy, 239, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) __PYX_ERR(0, 239, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":242
 * 	setInitialEntropy(sValue)
 * 
 * def cy_finalEntropy():             # <<<<<<<<<<<<<<
 * 	return finalEntropy()
 * 
 */
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_finalEntropy, 242, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(0, 242, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":245
 * 	return finalEntropy()
 * 
 * def cy_setFinalEntropy(double sValue):             # <<<<<<<<<<<<<<
 * 	setFinalEntropy(sValue)
 * 
 */
  __pyx_tuple__70 = PyTuple_Pack(2, __pyx_n_s_sValue, __pyx_n_s_sValue); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);
  __pyx_codeobj__71 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__70, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setFinalEntropy, 245, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__71)) __PYX_ERR(0, 245, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":248
 * 	setFinalEntropy(sValue)
 * 
 * def cy_incrementEntropy():             # <<<<<<<<<<<<<<
 * 	return incrementEntropy()
 * 
 */
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_incrementEntropy, 248, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(0, 248, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":251
 * 	return incrementEntropy()
 * 
 * def cy_setIncrementEntropy(double sValue):             # <<<<<<<<<<<<<<
 * 	setIncrementEntropy(sValue)
 * 
 */
  __pyx_tuple__73 = PyTuple_Pack(2, __pyx_n_s_sValue, __pyx_n_s_sValue); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);
  __pyx_codeobj__74 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__73, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setIncrementEntropy, 251, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__74)) __PYX_ERR(0, 251, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":254
 * 	setIncrementEntropy(sValue)
 * 
 * def cy_isIsochoric():             # <<<<<<<<<<<<<<
 * 	return isIsochoric()
 * 
 */
  __pyx_codeobj__75 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_isIsochoric, 254, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__75)) __PYX_ERR(0, 254, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":257
 * 	return isIsochoric()
 * 
 * def cy_setIsochoric(int value):             # <<<<<<<<<<<<<<
 * 	setIsochoric(value)
 * 
 */
  __pyx_tuple__76 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_value); if (unlikely(!__pyx_tuple__76)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__76);
  __Pyx_GIVEREF(__pyx_tuple__76);
  __pyx_codeobj__77 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__76, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setIsochoric, 257, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__77)) __PYX_ERR(0, 257, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":260
 * 	setIsochoric(value)
 * 
 * def cy_isIsenthalpic():             # <<<<<<<<<<<<<<
 * 	return isIsenthalpic()
 * 
 */
  __pyx_codeobj__78 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_isIsenthalpic, 260, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__78)) __PYX_ERR(0, 260, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":263
 * 	return isIsenthalpic()
 * 
 * def cy_setIsenthalpic(int value):             # <<<<<<<<<<<<<<
 * 	setIsenthalpic(value)
 * 
 */
  __pyx_tuple__79 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_value); if (unlikely(!__pyx_tuple__79)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__79);
  __Pyx_GIVEREF(__pyx_tuple__79);
  __pyx_codeobj__80 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__79, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setIsenthalpic, 263, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__80)) __PYX_ERR(0, 263, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":266
 * 	setIsenthalpic(value)
 * 
 * def cy_isIsotropic():             # <<<<<<<<<<<<<<
 * 	return isIsotropic()
 * 
 */
  __pyx_codeobj__81 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_isIsotropic, 266, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__81)) __PYX_ERR(0, 266, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":269
 * 	return isIsotropic()
 * 
 * def cy_setIsotropic(int value):             # <<<<<<<<<<<<<<
 * 	setIsotropic(value)
 * 
 */
  __pyx_tuple__82 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_value); if (unlikely(!__pyx_tuple__82)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);
  __pyx_codeobj__83 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__82, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setIsotropic, 269, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__83)) __PYX_ERR(0, 269, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":272
 * 	setIsotropic(value)
 * 
 * def cy_bulkCompAsList():             # <<<<<<<<<<<<<<
 * 	cdef double *bc = bulkCompAsPtr()
 * 	n = num_oxides()
 */
  __pyx_tuple__84 = PyTuple_Pack(5, __pyx_n_s_bc, __pyx_n_s_n, __pyx_n_s_result, __pyx_n_s_i, __pyx_n_s_entry); if (unlikely(!__pyx_tuple__84)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__84);
  __Pyx_GIVEREF(__pyx_tuple__84);
  __pyx_codeobj__85 = (PyObject*)__Pyx_PyCode_New(0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__84, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_bulkCompAsList, 272, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__85)) __PYX_ERR(0, 272, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":281
 * 	return result
 * 
 * def cy_setBulkCompAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	n = num_oxides()
 * 	cdef double *bc = <double *>malloc(n*sizeof(double))
 */
  __pyx_tuple__86 = PyTuple_Pack(4, __pyx_n_s_values, __pyx_n_s_n, __pyx_n_s_bc, __pyx_n_s_i); if (unlikely(!__pyx_tuple__86)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__86);
  __Pyx_GIVEREF(__pyx_tuple__86);
  __pyx_codeobj__87 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__86, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setBulkCompAsList, 281, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__87)) __PYX_ERR(0, 281, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":289
 * 	# free (bc) memory management in C function
 * 
 * def cy_liquidCompAsListOfLists():             # <<<<<<<<<<<<<<
 * 	cdef double **lc = liquidCompAsPtr()
 * 	nl = nLiquidCoexist()
 */
  __pyx_tuple__88 = PyTuple_Pack(6, __pyx_n_s_lc, __pyx_n_s_nl, __pyx_n_s_nc, __pyx_n_s_result, __pyx_n_s_i, __pyx_n_s_j); if (unlikely(!__pyx_tuple__88)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__88);
  __Pyx_GIVEREF(__pyx_tuple__88);
  __pyx_codeobj__89 = (PyObject*)__Pyx_PyCode_New(0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__88, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_liquidCompAsListOfLists, 289, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__89)) __PYX_ERR(0, 289, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":298
 * 	return result
 * 
 * def cy_setLiquidCompAsListOfLists(list values = [[]]):             # <<<<<<<<<<<<<<
 * 	nl = nLiquidCoexist()
 * 	nc = num_oxides()
 */
  __pyx_tuple__90 = PyTuple_Pack(6, __pyx_n_s_values, __pyx_n_s_nl, __pyx_n_s_nc, __pyx_n_s_lc, __pyx_n_s_i, __pyx_n_s_j); if (unlikely(!__pyx_tuple__90)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__90);
  __Pyx_GIVEREF(__pyx_tuple__90);
  __pyx_codeobj__91 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__90, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setLiquidCompAsListOfLists, 298, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__91)) __PYX_ERR(0, 298, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":311
 * 	#free (lc)  memory management in C function
 * 
 * def cy_nLiquidCoexist():             # <<<<<<<<<<<<<<
 * 	return nLiquidCoexist()
 * 
 */
  __pyx_codeobj__92 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_nLiquidCoexist, 311, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__92)) __PYX_ERR(0, 311, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":314
 * 	return nLiquidCoexist()
 * 
 * def cy_setnLiquidCoexist(int value):             # <<<<<<<<<<<<<<
 * 	setnLiquidCoexist(<int> value)
 * 
 */
  __pyx_tuple__93 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_value); if (unlikely(!__pyx_tuple__93)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);
  __pyx_codeobj__94 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setnLiquidCoexist, 314, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__94)) __PYX_ERR(0, 314, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":317
 * 	setnLiquidCoexist(<int> value)
 * 
 * def cy_liquidMass():             # <<<<<<<<<<<<<<
 * 	return <double> liquidMass()
 * 
 */
  __pyx_codeobj__95 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_liquidMass, 317, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__95)) __PYX_ERR(0, 317, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":320
 * 	return <double> liquidMass()
 * 
 * def cy_setLiquidMass(double value):             # <<<<<<<<<<<<<<
 * 	setLiquidMass(<double> value);
 * 
 */
  __pyx_tuple__96 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_value); if (unlikely(!__pyx_tuple__96)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);
  __pyx_codeobj__97 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__96, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setLiquidMass, 320, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__97)) __PYX_ERR(0, 320, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":323
 * 	setLiquidMass(<double> value);
 * 
 * def cy_multipleLiqs():             # <<<<<<<<<<<<<<
 * 	return False if multipleLiqs() == 0 else True;
 * 
 */
  __pyx_codeobj__98 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_multipleLiqs, 323, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__98)) __PYX_ERR(0, 323, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":326
 * 	return False if multipleLiqs() == 0 else True;
 * 
 * def cy_setMultipleLiqs(bint value):             # <<<<<<<<<<<<<<
 * 	setMultipleLiqs(<int> (1 if value else 0))
 * 
 */
  __pyx_tuple__99 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_value); if (unlikely(!__pyx_tuple__99)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);
  __pyx_codeobj__100 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__99, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setMultipleLiqs, 326, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__100)) __PYX_ERR(0, 326, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":329
 * 	setMultipleLiqs(<int> (1 if value else 0))
 * 
 * def cy_nSolidCoexistAsList():             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()
 */
  __pyx_tuple__101 = PyTuple_Pack(4, __pyx_n_s_ns, __pyx_n_s_npc, __pyx_n_s_result, __pyx_n_s_i); if (unlikely(!__pyx_tuple__101)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);
  __pyx_codeobj__102 = (PyObject*)__Pyx_PyCode_New(0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__101, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_nSolidCoexistAsList, 329, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__102)) __PYX_ERR(0, 329, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":335
 * 	return result
 * 
 * def cy_setnSolidCoexistAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))
 */
  __pyx_tuple__103 = PyTuple_Pack(4, __pyx_n_s_values, __pyx_n_s_npc, __pyx_n_s_ns, __pyx_n_s_i); if (unlikely(!__pyx_tuple__103)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__103);
  __Pyx_GIVEREF(__pyx_tuple__103);
  __pyx_codeobj__104 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__103, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setnSolidCoexistAsList, 335, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__104)) __PYX_ERR(0, 335, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":343
 * 	# free (ns)  memory management in C function
 * 
 * def cy_solidCompAsListOfLists():             # <<<<<<<<<<<<<<
 * 	cdef double **sc = solidCompAsPtr()
 * 	cdef int *ns = nSolidCoexist()
 */
  __pyx_tuple__105 = PyTuple_Pack(6, __pyx_n_s_sc, __pyx_n_s_ns, __pyx_n_s_npc, __pyx_n_s_result, __pyx_n_s_i, __pyx_n_s_j); if (unlikely(!__pyx_tuple__105)) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__105);
  __Pyx_GIVEREF(__pyx_tuple__105);
  __pyx_codeobj__106 = (PyObject*)__Pyx_PyCode_New(0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__105, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_solidCompAsListOfLists, 343, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__106)) __PYX_ERR(0, 343, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":352
 * 	return result
 * 
 * def cy_setSolidCompAsListOfLists(list values = [[]]):             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()
 */
  __pyx_tuple__107 = PyTuple_Pack(6, __pyx_n_s_values, __pyx_n_s_ns, __pyx_n_s_npc, __pyx_n_s_sc, __pyx_n_s_i, __pyx_n_s_j); if (unlikely(!__pyx_tuple__107)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__107);
  __Pyx_GIVEREF(__pyx_tuple__107);
  __pyx_codeobj__108 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__107, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setSolidCompAsListOfLists, 352, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__108)) __PYX_ERR(0, 352, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":365
 * 	#free (sc)  memory management in C function
 * 
 * def cy_solidMass():             # <<<<<<<<<<<<<<
 * 	return <double> solidMass()
 * 
 */
  __pyx_codeobj__109 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_solidMass, 365, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__109)) __PYX_ERR(0, 365, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":368
 * 	return <double> solidMass()
 * 
 * def cy_setSolidMass(double value):             # <<<<<<<<<<<<<<
 * 	setSolidMass(<double> value);
 * 
 */
  __pyx_tuple__110 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_value); if (unlikely(!__pyx_tuple__110)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__110);
  __Pyx_GIVEREF(__pyx_tuple__110);
  __pyx_codeobj__111 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__110, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setSolidMass, 368, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__111)) __PYX_ERR(0, 368, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":371
 * 	setSolidMass(<double> value);
 * 
 * def cy_incSolidsAsList():             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef int *inc = incSolids();
 */
  __pyx_tuple__112 = PyTuple_Pack(4, __pyx_n_s_npc, __pyx_n_s_inc, __pyx_n_s_result, __pyx_n_s_i); if (unlikely(!__pyx_tuple__112)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__112);
  __Pyx_GIVEREF(__pyx_tuple__112);
  __pyx_codeobj__113 = (PyObject*)__Pyx_PyCode_New(0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__112, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_incSolidsAsList, 371, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__113)) __PYX_ERR(0, 371, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":377
 * 	return result
 * 
 * def cy_setIncSolidsAsList(list values = []): # bint[:]             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef int *inc = <int *>malloc(npc*sizeof(int))
 */
  __pyx_tuple__114 = PyTuple_Pack(4, __pyx_n_s_values, __pyx_n_s_npc, __pyx_n_s_inc, __pyx_n_s_i); if (unlikely(!__pyx_tuple__114)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__114);
  __Pyx_GIVEREF(__pyx_tuple__114);
  __pyx_codeobj__115 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__114, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setIncSolidsAsList, 377, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__115)) __PYX_ERR(0, 377, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":385
 * 	# free (inc)  memory management in C function
 * 
 * def cy_fo2():             # <<<<<<<<<<<<<<
 * 	return fo2();
 * 
 */
  __pyx_codeobj__116 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_fo2, 385, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__116)) __PYX_ERR(0, 385, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":388
 * 	return fo2();
 * 
 * def cy_setfo2(double value):             # <<<<<<<<<<<<<<
 * 	setfo2(<double> value)
 * 
 */
  __pyx_tuple__117 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_value); if (unlikely(!__pyx_tuple__117)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__117);
  __Pyx_GIVEREF(__pyx_tuple__117);
  __pyx_codeobj__118 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__117, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setfo2, 388, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__118)) __PYX_ERR(0, 388, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":391
 * 	setfo2(<double> value)
 * 
 * def cy_fo2Path():             # <<<<<<<<<<<<<<
 * 	return fo2Path();
 * 
 */
  __pyx_codeobj__119 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_fo2Path, 391, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__119)) __PYX_ERR(0, 391, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":394
 * 	return fo2Path();
 * 
 * def cy_setfo2Path(int value):             # <<<<<<<<<<<<<<
 * 	setfo2Path(<int> value)
 * 
 */
  __pyx_tuple__120 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_value); if (unlikely(!__pyx_tuple__120)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__120);
  __Pyx_GIVEREF(__pyx_tuple__120);
  __pyx_codeobj__121 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__120, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setfo2Path, 394, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__121)) __PYX_ERR(0, 394, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":397
 * 	setfo2Path(<int> value)
 * 
 * def cy_fo2Delta():             # <<<<<<<<<<<<<<
 * 	return fo2Delta()
 * 
 */
  __pyx_codeobj__122 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_fo2Delta, 397, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__122)) __PYX_ERR(0, 397, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":400
 * 	return fo2Delta()
 * 
 * def cy_setfo2Delta(double value):             # <<<<<<<<<<<<<<
 * 	setfo2Delta(<double> value)
 * 
 */
  __pyx_tuple__123 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_value); if (unlikely(!__pyx_tuple__123)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__123);
  __Pyx_GIVEREF(__pyx_tuple__123);
  __pyx_codeobj__124 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__123, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setfo2Delta, 400, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__124)) __PYX_ERR(0, 400, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":403
 * 	setfo2Delta(<double> value)
 * 
 * def cy_oxygenContent():             # <<<<<<<<<<<<<<
 * 	return oxygenContent()
 * 
 */
  __pyx_codeobj__125 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_oxygenContent, 403, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__125)) __PYX_ERR(0, 403, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":406
 * 	return oxygenContent()
 * 
 * def cy_setOxygenContent(double value):             # <<<<<<<<<<<<<<
 * 	setOxygenContent(<double> value)
 * 
 */
  __pyx_tuple__126 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_value); if (unlikely(!__pyx_tuple__126)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__126);
  __Pyx_GIVEREF(__pyx_tuple__126);
  __pyx_codeobj__127 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__126, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setOxygenContent, 406, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__127)) __PYX_ERR(0, 406, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":409
 * 	setOxygenContent(<double> value)
 * 
 * def cy_fractionateSol():             # <<<<<<<<<<<<<<
 * 	return True if fractionateSol() == 1 else False
 * 
 */
  __pyx_codeobj__128 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_fractionateSol, 409, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__128)) __PYX_ERR(0, 409, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":412
 * 	return True if fractionateSol() == 1 else False
 * 
 * def cy_setFractionateSol(bint value):             # <<<<<<<<<<<<<<
 * 	setFractionateSol(<int>(1 if value else 0));
 * 
 */
  __pyx_tuple__129 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_value); if (unlikely(!__pyx_tuple__129)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);
  __pyx_codeobj__130 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__129, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setFractionateSol, 412, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__130)) __PYX_ERR(0, 412, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":415
 * 	setFractionateSol(<int>(1 if value else 0));
 * 
 * def cy_fractionateLiq():             # <<<<<<<<<<<<<<
 * 	return True if fractionateLiq() == 1 else False
 * 
 */
  __pyx_codeobj__131 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_fractionateLiq, 415, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__131)) __PYX_ERR(0, 415, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":418
 * 	return True if fractionateLiq() == 1 else False
 * 
 * def cy_setFractionateLiq(bint value):             # <<<<<<<<<<<<<<
 * 	setFractionateLiq(<int>(1 if value else 0))
 * 
 */
  __pyx_tuple__132 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_value); if (unlikely(!__pyx_tuple__132)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__132);
  __Pyx_GIVEREF(__pyx_tuple__132);
  __pyx_codeobj__133 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__132, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setFractionateLiq, 418, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__133)) __PYX_ERR(0, 418, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":421
 * 	setFractionateLiq(<int>(1 if value else 0))
 * 
 * def cy_fractionateFlu():             # <<<<<<<<<<<<<<
 * 	return True if fractionateFlu() == 1 else False
 * 
 */
  __pyx_codeobj__134 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_fractionateFlu, 421, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__134)) __PYX_ERR(0, 421, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":424
 * 	return True if fractionateFlu() == 1 else False
 * 
 * def cy_setFractionateFlu(bint value):             # <<<<<<<<<<<<<<
 * 	setFractionateFlu(<int>(1 if value else 0));
 * 
 */
  __pyx_tuple__135 = PyTuple_Pack(2, __pyx_n_s_value, __pyx_n_s_value); if (unlikely(!__pyx_tuple__135)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__135);
  __Pyx_GIVEREF(__pyx_tuple__135);
  __pyx_codeobj__136 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setFractionateFlu, 424, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__136)) __PYX_ERR(0, 424, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":427
 * 	setFractionateFlu(<int>(1 if value else 0));
 * 
 * def cy_nFracCoexistAsList():             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nFracCoexist()
 * 	if ns == NULL:
 */
  __pyx_tuple__137 = PyTuple_Pack(4, __pyx_n_s_ns, __pyx_n_s_npc, __pyx_n_s_result, __pyx_n_s_i); if (unlikely(!__pyx_tuple__137)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);
  __pyx_codeobj__138 = (PyObject*)__Pyx_PyCode_New(0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_nFracCoexistAsList, 427, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__138)) __PYX_ERR(0, 427, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":435
 * 	return result
 * 
 * def cy_setnFracCoexistAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	if values is None:
 * 		return
 */
  __pyx_tuple__139 = PyTuple_Pack(4, __pyx_n_s_values, __pyx_n_s_npc, __pyx_n_s_ns, __pyx_n_s_i); if (unlikely(!__pyx_tuple__139)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__139);
  __Pyx_GIVEREF(__pyx_tuple__139);
  __pyx_codeobj__140 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__139, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setnFracCoexistAsList, 435, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__140)) __PYX_ERR(0, 435, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":445
 * 	# free (ns)  memory management in C function
 * 
 * def cy_fracSCompAsListOfLists():             # <<<<<<<<<<<<<<
 * 	cdef double **sc = fracSComp()
 * 	cdef int *ns = nSolidCoexist()
 */
  __pyx_tuple__141 = PyTuple_Pack(6, __pyx_n_s_sc, __pyx_n_s_ns, __pyx_n_s_npc, __pyx_n_s_result, __pyx_n_s_i, __pyx_n_s_j); if (unlikely(!__pyx_tuple__141)) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__141);
  __Pyx_GIVEREF(__pyx_tuple__141);
  __pyx_codeobj__142 = (PyObject*)__Pyx_PyCode_New(0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__141, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_fracSCompAsListOfLists, 445, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__142)) __PYX_ERR(0, 445, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":456
 * 	return result
 * 
 * def cy_setFracSCompAsListOfLists(list values = [[]]):             # <<<<<<<<<<<<<<
 * 	if values is None:
 * 		return
 */
  __pyx_tuple__143 = PyTuple_Pack(6, __pyx_n_s_values, __pyx_n_s_ns, __pyx_n_s_npc, __pyx_n_s_sc, __pyx_n_s_i, __pyx_n_s_j); if (unlikely(!__pyx_tuple__143)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);
  __pyx_codeobj__144 = (PyObject*)__Pyx_PyCode_New(1, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__143, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setFracSCompAsListOfLists, 456, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__144)) __PYX_ERR(0, 456, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":471
 * 	#free (sc)  memory management in C function
 * 
 * def cy_fracLCompAsList():             # <<<<<<<<<<<<<<
 * 	nc = num_oxides()
 * 	cdef double *nl = fracLComp()
 */
  __pyx_tuple__145 = PyTuple_Pack(4, __pyx_n_s_nc, __pyx_n_s_nl, __pyx_n_s_result, __pyx_n_s_i); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);
  __pyx_codeobj__146 = (PyObject*)__Pyx_PyCode_New(0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__145, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_fracLCompAsList, 471, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__146)) __PYX_ERR(0, 471, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":479
 * 	return result
 * 
 * def cy_setFracLCompAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	if values is None:
 * 		return
 */
  __pyx_tuple__147 = PyTuple_Pack(4, __pyx_n_s_values, __pyx_n_s_nc, __pyx_n_s_nl, __pyx_n_s_i); if (unlikely(!__pyx_tuple__147)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);
  __pyx_codeobj__148 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__147, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setFracLCompAsList, 479, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__148)) __PYX_ERR(0, 479, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":489
 * 	#free (nl)  memory management in C function
 * 
 * def cy_fracMass():             # <<<<<<<<<<<<<<
 * 	return <double> fracMass()
 * 
 */
  __pyx_codeobj__149 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_fracMass, 489, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__149)) __PYX_ERR(0, 489, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":492
 * 	return <double> fracMass()
 * 
 * def cy_setFracMass(double value):             # <<<<<<<<<<<<<<
 * 	setFracMass = <double> value
 * 
 */
  __pyx_tuple__150 = PyTuple_Pack(3, __pyx_n_s_value, __pyx_n_s_value, __pyx_n_s_setFracMass); if (unlikely(!__pyx_tuple__150)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__150);
  __Pyx_GIVEREF(__pyx_tuple__150);
  __pyx_codeobj__151 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__150, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_melts_rMELTSframework_pyx, __pyx_n_s_cy_setFracMass, 492, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__151)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(2, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(3, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(4, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(1, 200, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(1, 223, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(1, 227, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(1, 239, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_generic) __PYX_ERR(1, 771, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType(__pyx_t_1, "numpy", "number", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_number) __PYX_ERR(1, 773, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_integer) __PYX_ERR(1, 775, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(1, 777, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(1, 779, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(1, 781, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_floating) __PYX_ERR(1, 783, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(1, 785, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(1, 787, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType(__pyx_t_1, "numpy", "character", sizeof(PyObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5numpy_character) __PYX_ERR(1, 789, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __Pyx_ImportType_CheckSize_Ignore);
   if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(1, 827, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initrMELTSframework(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initrMELTSframework(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_rMELTSframework(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_rMELTSframework(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_rMELTSframework(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'rMELTSframework' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_rMELTSframework(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("rMELTSframework", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_melts__rMELTSframework) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "melts.rMELTSframework")) {
      if (unlikely(PyDict_SetItemString(modules, "melts.rMELTSframework", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  (void)__Pyx_modinit_type_init_code();
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "melts/rMELTSframework.pyx":96
 * from libc.stdlib cimport malloc, free
 * from cpython cimport PyObject, Py_INCREF
 * import ctypes             # <<<<<<<<<<<<<<
 * 
 * # Import the Python-level symbols of numpy
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_ctypes, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ctypes, __pyx_t_1) < 0) __PYX_ERR(0, 96, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":99
 * 
 * # Import the Python-level symbols of numpy
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * # Import the C-level symbols of numpy
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_numpy, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_1) < 0) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":106
 * # Numpy must be initialized. When using numpy from C or Cython you must
 * # _always_ do that, or you will have segfaults
 * np.import_array()             # <<<<<<<<<<<<<<
 * 
 * # here is the "wrapper" signature
 */
  __pyx_t_2 = __pyx_f_5numpy_import_array(); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 106, __pyx_L1_error)

  /* "melts/rMELTSframework.pyx":110
 * # here is the "wrapper" signature
 * 
 * def cy_MELTScalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return MELTScalculationModeConstant()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_1cy_MELTScalculationModeConstant, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_MELTScalculationModeConstant, __pyx_t_1) < 0) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":113
 * 	return MELTScalculationModeConstant()
 * 
 * def cy_pMELTScalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return pMELTScalculationModeConstant()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_3cy_pMELTScalculationModeConstant, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_pMELTScalculationModeConstant, __pyx_t_1) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":116
 * 	return pMELTScalculationModeConstant()
 * 
 * def cy_xMELTScalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return xMELTScalculationModeConstant()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_5cy_xMELTScalculationModeConstant, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_xMELTScalculationModeConstant, __pyx_t_1) < 0) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":119
 * 	return xMELTScalculationModeConstant()
 * 
 * def cy_MELTSandCO2calculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return MELTSandCO2calculationModeConstant()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_7cy_MELTSandCO2calculationModeConstant, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_MELTSandCO2calculationModeCon, __pyx_t_1) < 0) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":122
 * 	return MELTSandCO2calculationModeConstant()
 * 
 * def cy_MELTSandCO2_H2OcalculationModeConstant():             # <<<<<<<<<<<<<<
 * 	return MELTSandCO2_H2OcalculationModeConstant()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_9cy_MELTSandCO2_H2OcalculationModeConstant, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_MELTSandCO2_H2OcalculationMod, __pyx_t_1) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":125
 * 	return MELTSandCO2_H2OcalculationModeConstant()
 * 
 * def cy_num_oxides():             # <<<<<<<<<<<<<<
 * 	return num_oxides()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_11cy_num_oxides, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_num_oxides, __pyx_t_1) < 0) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":128
 * 	return num_oxides()
 * 
 * def cy_num_phase_components():             # <<<<<<<<<<<<<<
 * 	return num_phase_components()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_13cy_num_phase_components, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_num_phase_components, __pyx_t_1) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":131
 * 	return num_phase_components()
 * 
 * def cy_oxideListAsStrings():             # <<<<<<<<<<<<<<
 * 	cdef char **names = oxideListAsStrings()
 * 	n = num_oxides()
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_15cy_oxideListAsStrings, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_oxideListAsStrings, __pyx_t_1) < 0) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":140
 * 	return result
 * 
 * def cy_phaseListAsStrings(int modelSelection):             # <<<<<<<<<<<<<<
 * 	cdef char **names = phaseListAsStrings(modelSelection)
 * 	n = num_phases(modelSelection)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_17cy_phaseListAsStrings, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_phaseListAsStrings, __pyx_t_1) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":149
 * 	return result
 * 
 * def cy_init(int modelSelection):             # <<<<<<<<<<<<<<
 * 	return init(modelSelection)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_19cy_init, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_init, __pyx_t_1) < 0) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":152
 * 	return init(modelSelection)
 * 
 * def cy_parseAndLoadDataStructuresFromXMLstring(inputXMLstring):             # <<<<<<<<<<<<<<
 * 	return parseAndLoadDataStructuresFromXMLstring(inputXMLstring.encode('utf-8'))
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_21cy_parseAndLoadDataStructuresFromXMLstring, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_parseAndLoadDataStructuresFro, __pyx_t_1) < 0) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":155
 * 	return parseAndLoadDataStructuresFromXMLstring(inputXMLstring.encode('utf-8'))
 * 
 * def cy_writeDataStructuresToXMLstring(sessionID):             # <<<<<<<<<<<<<<
 * 	result_b = writeDataStructuresToXMLstring(sessionID.encode('utf-8'))
 * 	result = result_b.decode('utf-8')
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_23cy_writeDataStructuresToXMLstring, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_writeDataStructuresToXMLstrin, __pyx_t_1) < 0) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":161
 * 	return result
 * 
 * def cy_performMELTScalculation(typeOfCalc):             # <<<<<<<<<<<<<<
 * 	return performMELTScalculation(typeOfCalc)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_25cy_performMELTScalculation, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_performMELTScalculation, __pyx_t_1) < 0) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":164
 * 	return performMELTScalculation(typeOfCalc)
 * 
 * def cy_initialTemperature():             # <<<<<<<<<<<<<<
 * 	return initialTemperature()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_27cy_initialTemperature, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_initialTemperature, __pyx_t_1) < 0) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":167
 * 	return initialTemperature()
 * 
 * def cy_setInitialTemperature(double tValue):             # <<<<<<<<<<<<<<
 * 	setInitialTemperature(tValue)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_29cy_setInitialTemperature, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setInitialTemperature, __pyx_t_1) < 0) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":170
 * 	setInitialTemperature(tValue)
 * 
 * def cy_finalTemperature():             # <<<<<<<<<<<<<<
 * 	return finalTemperature()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_31cy_finalTemperature, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_finalTemperature, __pyx_t_1) < 0) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":173
 * 	return finalTemperature()
 * 
 * def cy_setFinalTemperature(double tValue):             # <<<<<<<<<<<<<<
 * 	setFinalTemperature(tValue)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_33cy_setFinalTemperature, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setFinalTemperature, __pyx_t_1) < 0) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":176
 * 	setFinalTemperature(tValue)
 * 
 * def cy_incrementTemperature():             # <<<<<<<<<<<<<<
 * 	return incrementTemperature()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_35cy_incrementTemperature, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_incrementTemperature, __pyx_t_1) < 0) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":179
 * 	return incrementTemperature()
 * 
 * def cy_setIncrementTemperature(double tValue):             # <<<<<<<<<<<<<<
 * 	setIncrementTemperature(tValue);
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_37cy_setIncrementTemperature, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setIncrementTemperature, __pyx_t_1) < 0) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":182
 * 	setIncrementTemperature(tValue);
 * 
 * def cy_initialPressure():             # <<<<<<<<<<<<<<
 * 	return initialPressure()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_39cy_initialPressure, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_initialPressure, __pyx_t_1) < 0) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":185
 * 	return initialPressure()
 * 
 * def cy_setInitialPressure(double pValue):             # <<<<<<<<<<<<<<
 * 	setInitialPressure(pValue)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_41cy_setInitialPressure, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setInitialPressure, __pyx_t_1) < 0) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":188
 * 	setInitialPressure(pValue)
 * 
 * def cy_finalPressure():             # <<<<<<<<<<<<<<
 * 	return finalPressure()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_43cy_finalPressure, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_finalPressure, __pyx_t_1) < 0) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":191
 * 	return finalPressure()
 * 
 * def cy_setFinalPressure(double pValue):             # <<<<<<<<<<<<<<
 * 	setFinalPressure(pValue)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_45cy_setFinalPressure, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setFinalPressure, __pyx_t_1) < 0) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":194
 * 	setFinalPressure(pValue)
 * 
 * def cy_incrementPressure():             # <<<<<<<<<<<<<<
 * 	return incrementPressure()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_47cy_incrementPressure, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_incrementPressure, __pyx_t_1) < 0) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":197
 * 	return incrementPressure()
 * 
 * def cy_setIncrementPressure(double pValue):             # <<<<<<<<<<<<<<
 * 	setIncrementPressure(pValue)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_49cy_setIncrementPressure, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setIncrementPressure, __pyx_t_1) < 0) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":200
 * 	setIncrementPressure(pValue)
 * 
 * def cy_initialVolume():             # <<<<<<<<<<<<<<
 * 	return initialVolume()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_51cy_initialVolume, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_initialVolume, __pyx_t_1) < 0) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":203
 * 	return initialVolume()
 * 
 * def cy_setInitialVolume(double vValue):             # <<<<<<<<<<<<<<
 * 	setInitialVolume(vValue)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_53cy_setInitialVolume, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setInitialVolume, __pyx_t_1) < 0) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":206
 * 	setInitialVolume(vValue)
 * 
 * def cy_finalVolume():             # <<<<<<<<<<<<<<
 * 	return finalVolume()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_55cy_finalVolume, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_finalVolume, __pyx_t_1) < 0) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":209
 * 	return finalVolume()
 * 
 * def cy_setFinalVolume(double vValue):             # <<<<<<<<<<<<<<
 * 	setFinalVolume(vValue)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_57cy_setFinalVolume, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setFinalVolume, __pyx_t_1) < 0) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":212
 * 	setFinalVolume(vValue)
 * 
 * def cy_incrementVolume():             # <<<<<<<<<<<<<<
 * 	return incrementVolume()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_59cy_incrementVolume, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_incrementVolume, __pyx_t_1) < 0) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":215
 * 	return incrementVolume()
 * 
 * def cy_setIncrementVolume(double vValue):             # <<<<<<<<<<<<<<
 * 	setIncrementVolume(vValue)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_61cy_setIncrementVolume, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setIncrementVolume, __pyx_t_1) < 0) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":218
 * 	setIncrementVolume(vValue)
 * 
 * def cy_initialEnthalpy():             # <<<<<<<<<<<<<<
 * 	return initialEnthalpy()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_63cy_initialEnthalpy, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_initialEnthalpy, __pyx_t_1) < 0) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":221
 * 	return initialEnthalpy()
 * 
 * def cy_setInitialEnthalpy(double hValue):             # <<<<<<<<<<<<<<
 * 	setInitialEnthalpy(hValue)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_65cy_setInitialEnthalpy, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setInitialEnthalpy, __pyx_t_1) < 0) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":224
 * 	setInitialEnthalpy(hValue)
 * 
 * def cy_finalEnthalpy():             # <<<<<<<<<<<<<<
 * 	return finalEnthalpy()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_67cy_finalEnthalpy, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_finalEnthalpy, __pyx_t_1) < 0) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":227
 * 	return finalEnthalpy()
 * 
 * def cy_setFinalEnthalpy(double hValue):             # <<<<<<<<<<<<<<
 * 	setFinalEnthalpy(hValue)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_69cy_setFinalEnthalpy, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setFinalEnthalpy, __pyx_t_1) < 0) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":230
 * 	setFinalEnthalpy(hValue)
 * 
 * def cy_incrementEnthalpy():             # <<<<<<<<<<<<<<
 * 	return incrementEnthalpy()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_71cy_incrementEnthalpy, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_incrementEnthalpy, __pyx_t_1) < 0) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":233
 * 	return incrementEnthalpy()
 * 
 * def cy_setIncrementEnthalpy(double hValue):             # <<<<<<<<<<<<<<
 * 	setIncrementEnthalpy(hValue)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_73cy_setIncrementEnthalpy, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setIncrementEnthalpy, __pyx_t_1) < 0) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":236
 * 	setIncrementEnthalpy(hValue)
 * 
 * def cy_initialEntropy():             # <<<<<<<<<<<<<<
 * 	return initialEntropy()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_75cy_initialEntropy, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_initialEntropy, __pyx_t_1) < 0) __PYX_ERR(0, 236, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":239
 * 	return initialEntropy()
 * 
 * def cy_setInitialEntropy(double sValue):             # <<<<<<<<<<<<<<
 * 	setInitialEntropy(sValue)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_77cy_setInitialEntropy, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setInitialEntropy, __pyx_t_1) < 0) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":242
 * 	setInitialEntropy(sValue)
 * 
 * def cy_finalEntropy():             # <<<<<<<<<<<<<<
 * 	return finalEntropy()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_79cy_finalEntropy, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_finalEntropy, __pyx_t_1) < 0) __PYX_ERR(0, 242, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":245
 * 	return finalEntropy()
 * 
 * def cy_setFinalEntropy(double sValue):             # <<<<<<<<<<<<<<
 * 	setFinalEntropy(sValue)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_81cy_setFinalEntropy, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setFinalEntropy, __pyx_t_1) < 0) __PYX_ERR(0, 245, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":248
 * 	setFinalEntropy(sValue)
 * 
 * def cy_incrementEntropy():             # <<<<<<<<<<<<<<
 * 	return incrementEntropy()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_83cy_incrementEntropy, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_incrementEntropy, __pyx_t_1) < 0) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":251
 * 	return incrementEntropy()
 * 
 * def cy_setIncrementEntropy(double sValue):             # <<<<<<<<<<<<<<
 * 	setIncrementEntropy(sValue)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_85cy_setIncrementEntropy, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setIncrementEntropy, __pyx_t_1) < 0) __PYX_ERR(0, 251, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":254
 * 	setIncrementEntropy(sValue)
 * 
 * def cy_isIsochoric():             # <<<<<<<<<<<<<<
 * 	return isIsochoric()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_87cy_isIsochoric, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_isIsochoric, __pyx_t_1) < 0) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":257
 * 	return isIsochoric()
 * 
 * def cy_setIsochoric(int value):             # <<<<<<<<<<<<<<
 * 	setIsochoric(value)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_89cy_setIsochoric, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setIsochoric, __pyx_t_1) < 0) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":260
 * 	setIsochoric(value)
 * 
 * def cy_isIsenthalpic():             # <<<<<<<<<<<<<<
 * 	return isIsenthalpic()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_91cy_isIsenthalpic, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_isIsenthalpic, __pyx_t_1) < 0) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":263
 * 	return isIsenthalpic()
 * 
 * def cy_setIsenthalpic(int value):             # <<<<<<<<<<<<<<
 * 	setIsenthalpic(value)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_93cy_setIsenthalpic, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setIsenthalpic, __pyx_t_1) < 0) __PYX_ERR(0, 263, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":266
 * 	setIsenthalpic(value)
 * 
 * def cy_isIsotropic():             # <<<<<<<<<<<<<<
 * 	return isIsotropic()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_95cy_isIsotropic, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_isIsotropic, __pyx_t_1) < 0) __PYX_ERR(0, 266, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":269
 * 	return isIsotropic()
 * 
 * def cy_setIsotropic(int value):             # <<<<<<<<<<<<<<
 * 	setIsotropic(value)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_97cy_setIsotropic, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setIsotropic, __pyx_t_1) < 0) __PYX_ERR(0, 269, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":272
 * 	setIsotropic(value)
 * 
 * def cy_bulkCompAsList():             # <<<<<<<<<<<<<<
 * 	cdef double *bc = bulkCompAsPtr()
 * 	n = num_oxides()
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_99cy_bulkCompAsList, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_bulkCompAsList, __pyx_t_1) < 0) __PYX_ERR(0, 272, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":281
 * 	return result
 * 
 * def cy_setBulkCompAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	n = num_oxides()
 * 	cdef double *bc = <double *>malloc(n*sizeof(double))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_k_ = ((PyObject*)__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_101cy_setBulkCompAsList, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setBulkCompAsList, __pyx_t_1) < 0) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":289
 * 	# free (bc) memory management in C function
 * 
 * def cy_liquidCompAsListOfLists():             # <<<<<<<<<<<<<<
 * 	cdef double **lc = liquidCompAsPtr()
 * 	nl = nLiquidCoexist()
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_103cy_liquidCompAsListOfLists, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_liquidCompAsListOfLists, __pyx_t_1) < 0) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":298
 * 	return result
 * 
 * def cy_setLiquidCompAsListOfLists(list values = [[]]):             # <<<<<<<<<<<<<<
 * 	nl = nLiquidCoexist()
 * 	nc = num_oxides()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_k__2 = ((PyObject*)__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_105cy_setLiquidCompAsListOfLists, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setLiquidCompAsListOfLists, __pyx_t_3) < 0) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":311
 * 	#free (lc)  memory management in C function
 * 
 * def cy_nLiquidCoexist():             # <<<<<<<<<<<<<<
 * 	return nLiquidCoexist()
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_107cy_nLiquidCoexist, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_nLiquidCoexist, __pyx_t_3) < 0) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":314
 * 	return nLiquidCoexist()
 * 
 * def cy_setnLiquidCoexist(int value):             # <<<<<<<<<<<<<<
 * 	setnLiquidCoexist(<int> value)
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_109cy_setnLiquidCoexist, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setnLiquidCoexist, __pyx_t_3) < 0) __PYX_ERR(0, 314, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":317
 * 	setnLiquidCoexist(<int> value)
 * 
 * def cy_liquidMass():             # <<<<<<<<<<<<<<
 * 	return <double> liquidMass()
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_111cy_liquidMass, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_liquidMass, __pyx_t_3) < 0) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":320
 * 	return <double> liquidMass()
 * 
 * def cy_setLiquidMass(double value):             # <<<<<<<<<<<<<<
 * 	setLiquidMass(<double> value);
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_113cy_setLiquidMass, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setLiquidMass, __pyx_t_3) < 0) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":323
 * 	setLiquidMass(<double> value);
 * 
 * def cy_multipleLiqs():             # <<<<<<<<<<<<<<
 * 	return False if multipleLiqs() == 0 else True;
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_115cy_multipleLiqs, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_multipleLiqs, __pyx_t_3) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":326
 * 	return False if multipleLiqs() == 0 else True;
 * 
 * def cy_setMultipleLiqs(bint value):             # <<<<<<<<<<<<<<
 * 	setMultipleLiqs(<int> (1 if value else 0))
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_117cy_setMultipleLiqs, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setMultipleLiqs, __pyx_t_3) < 0) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":329
 * 	setMultipleLiqs(<int> (1 if value else 0))
 * 
 * def cy_nSolidCoexistAsList():             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_119cy_nSolidCoexistAsList, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_nSolidCoexistAsList, __pyx_t_3) < 0) __PYX_ERR(0, 329, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":335
 * 	return result
 * 
 * def cy_setnSolidCoexistAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef int *ns = <int *>malloc(npc*sizeof(int))
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_k__3 = ((PyObject*)__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_121cy_setnSolidCoexistAsList, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setnSolidCoexistAsList, __pyx_t_3) < 0) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":343
 * 	# free (ns)  memory management in C function
 * 
 * def cy_solidCompAsListOfLists():             # <<<<<<<<<<<<<<
 * 	cdef double **sc = solidCompAsPtr()
 * 	cdef int *ns = nSolidCoexist()
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_123cy_solidCompAsListOfLists, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_solidCompAsListOfLists, __pyx_t_3) < 0) __PYX_ERR(0, 343, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":352
 * 	return result
 * 
 * def cy_setSolidCompAsListOfLists(list values = [[]]):             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nSolidCoexist()
 * 	npc = num_phase_components()
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_k__4 = ((PyObject*)__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_125cy_setSolidCompAsListOfLists, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setSolidCompAsListOfLists, __pyx_t_1) < 0) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":365
 * 	#free (sc)  memory management in C function
 * 
 * def cy_solidMass():             # <<<<<<<<<<<<<<
 * 	return <double> solidMass()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_127cy_solidMass, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 365, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_solidMass, __pyx_t_1) < 0) __PYX_ERR(0, 365, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":368
 * 	return <double> solidMass()
 * 
 * def cy_setSolidMass(double value):             # <<<<<<<<<<<<<<
 * 	setSolidMass(<double> value);
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_129cy_setSolidMass, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setSolidMass, __pyx_t_1) < 0) __PYX_ERR(0, 368, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":371
 * 	setSolidMass(<double> value);
 * 
 * def cy_incSolidsAsList():             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef int *inc = incSolids();
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_131cy_incSolidsAsList, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_incSolidsAsList, __pyx_t_1) < 0) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":377
 * 	return result
 * 
 * def cy_setIncSolidsAsList(list values = []): # bint[:]             # <<<<<<<<<<<<<<
 * 	npc = num_phase_components()
 * 	cdef int *inc = <int *>malloc(npc*sizeof(int))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_k__5 = ((PyObject*)__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_133cy_setIncSolidsAsList, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setIncSolidsAsList, __pyx_t_1) < 0) __PYX_ERR(0, 377, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":385
 * 	# free (inc)  memory management in C function
 * 
 * def cy_fo2():             # <<<<<<<<<<<<<<
 * 	return fo2();
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_135cy_fo2, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_fo2, __pyx_t_1) < 0) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":388
 * 	return fo2();
 * 
 * def cy_setfo2(double value):             # <<<<<<<<<<<<<<
 * 	setfo2(<double> value)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_137cy_setfo2, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setfo2, __pyx_t_1) < 0) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":391
 * 	setfo2(<double> value)
 * 
 * def cy_fo2Path():             # <<<<<<<<<<<<<<
 * 	return fo2Path();
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_139cy_fo2Path, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_fo2Path, __pyx_t_1) < 0) __PYX_ERR(0, 391, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":394
 * 	return fo2Path();
 * 
 * def cy_setfo2Path(int value):             # <<<<<<<<<<<<<<
 * 	setfo2Path(<int> value)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_141cy_setfo2Path, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setfo2Path, __pyx_t_1) < 0) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":397
 * 	setfo2Path(<int> value)
 * 
 * def cy_fo2Delta():             # <<<<<<<<<<<<<<
 * 	return fo2Delta()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_143cy_fo2Delta, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_fo2Delta, __pyx_t_1) < 0) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":400
 * 	return fo2Delta()
 * 
 * def cy_setfo2Delta(double value):             # <<<<<<<<<<<<<<
 * 	setfo2Delta(<double> value)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_145cy_setfo2Delta, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setfo2Delta, __pyx_t_1) < 0) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":403
 * 	setfo2Delta(<double> value)
 * 
 * def cy_oxygenContent():             # <<<<<<<<<<<<<<
 * 	return oxygenContent()
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_147cy_oxygenContent, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_oxygenContent, __pyx_t_1) < 0) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":406
 * 	return oxygenContent()
 * 
 * def cy_setOxygenContent(double value):             # <<<<<<<<<<<<<<
 * 	setOxygenContent(<double> value)
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_149cy_setOxygenContent, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setOxygenContent, __pyx_t_1) < 0) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":409
 * 	setOxygenContent(<double> value)
 * 
 * def cy_fractionateSol():             # <<<<<<<<<<<<<<
 * 	return True if fractionateSol() == 1 else False
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_151cy_fractionateSol, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_fractionateSol, __pyx_t_1) < 0) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":412
 * 	return True if fractionateSol() == 1 else False
 * 
 * def cy_setFractionateSol(bint value):             # <<<<<<<<<<<<<<
 * 	setFractionateSol(<int>(1 if value else 0));
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_153cy_setFractionateSol, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setFractionateSol, __pyx_t_1) < 0) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":415
 * 	setFractionateSol(<int>(1 if value else 0));
 * 
 * def cy_fractionateLiq():             # <<<<<<<<<<<<<<
 * 	return True if fractionateLiq() == 1 else False
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_155cy_fractionateLiq, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_fractionateLiq, __pyx_t_1) < 0) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":418
 * 	return True if fractionateLiq() == 1 else False
 * 
 * def cy_setFractionateLiq(bint value):             # <<<<<<<<<<<<<<
 * 	setFractionateLiq(<int>(1 if value else 0))
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_157cy_setFractionateLiq, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setFractionateLiq, __pyx_t_1) < 0) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":421
 * 	setFractionateLiq(<int>(1 if value else 0))
 * 
 * def cy_fractionateFlu():             # <<<<<<<<<<<<<<
 * 	return True if fractionateFlu() == 1 else False
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_159cy_fractionateFlu, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_fractionateFlu, __pyx_t_1) < 0) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":424
 * 	return True if fractionateFlu() == 1 else False
 * 
 * def cy_setFractionateFlu(bint value):             # <<<<<<<<<<<<<<
 * 	setFractionateFlu(<int>(1 if value else 0));
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_161cy_setFractionateFlu, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setFractionateFlu, __pyx_t_1) < 0) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":427
 * 	setFractionateFlu(<int>(1 if value else 0));
 * 
 * def cy_nFracCoexistAsList():             # <<<<<<<<<<<<<<
 * 	cdef int *ns = nFracCoexist()
 * 	if ns == NULL:
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_163cy_nFracCoexistAsList, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_nFracCoexistAsList, __pyx_t_1) < 0) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":435
 * 	return result
 * 
 * def cy_setnFracCoexistAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	if values is None:
 * 		return
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_k__6 = ((PyObject*)__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_165cy_setnFracCoexistAsList, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setnFracCoexistAsList, __pyx_t_1) < 0) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":445
 * 	# free (ns)  memory management in C function
 * 
 * def cy_fracSCompAsListOfLists():             # <<<<<<<<<<<<<<
 * 	cdef double **sc = fracSComp()
 * 	cdef int *ns = nSolidCoexist()
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_167cy_fracSCompAsListOfLists, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_fracSCompAsListOfLists, __pyx_t_1) < 0) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "melts/rMELTSframework.pyx":456
 * 	return result
 * 
 * def cy_setFracSCompAsListOfLists(list values = [[]]):             # <<<<<<<<<<<<<<
 * 	if values is None:
 * 		return
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_k__7 = ((PyObject*)__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_169cy_setFracSCompAsListOfLists, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setFracSCompAsListOfLists, __pyx_t_3) < 0) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":471
 * 	#free (sc)  memory management in C function
 * 
 * def cy_fracLCompAsList():             # <<<<<<<<<<<<<<
 * 	nc = num_oxides()
 * 	cdef double *nl = fracLComp()
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_171cy_fracLCompAsList, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_fracLCompAsList, __pyx_t_3) < 0) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":479
 * 	return result
 * 
 * def cy_setFracLCompAsList(list values = []):             # <<<<<<<<<<<<<<
 * 	if values is None:
 * 		return
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_k__8 = ((PyObject*)__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_173cy_setFracLCompAsList, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setFracLCompAsList, __pyx_t_3) < 0) __PYX_ERR(0, 479, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":489
 * 	#free (nl)  memory management in C function
 * 
 * def cy_fracMass():             # <<<<<<<<<<<<<<
 * 	return <double> fracMass()
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_175cy_fracMass, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_fracMass, __pyx_t_3) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":492
 * 	return <double> fracMass()
 * 
 * def cy_setFracMass(double value):             # <<<<<<<<<<<<<<
 * 	setFracMass = <double> value
 * 
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5melts_15rMELTSframework_177cy_setFracMass, NULL, __pyx_n_s_melts_rMELTSframework); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cy_setFracMass, __pyx_t_3) < 0) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "melts/rMELTSframework.pyx":1
 * cdef extern from "cframework.h":             # <<<<<<<<<<<<<<
 * 	int MELTScalculationModeConstant();
 * 	int pMELTScalculationModeConstant();
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "../../../usr/local/lib/python3.7/site-packages/numpy/__init__.pxd":1014
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init melts.rMELTSframework", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init melts.rMELTSframework");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* decode_c_bytes */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
